\ cf2018.f 2018 Jun 06b  convert NASM file to Forth 

decimal

4 constant INDENT_SPACES

variable delayedStop
: KK ( -- )   -1 delayedStop ! ;

variable markAsDone
: MM ( -- )   -1 markAsDone ! ;

10000 constant MAX_NUMBER_OF_LINES

: MyFileName ( -- $ )   s" cf2018.nasm" ;

variable NASMfileHandle
0 NASMfileHandle !

: OpenReadFile 
   MyFileName r/o open-file 
   if ." Couldn't open read file "  MyFileName type space  #2001 throw  then
   NASMfileHandle !
;

$200 constant |ReadLineBuffer| 
|ReadLineBuffer| Buffer: ReadLineBuffer         \ the original complete line

|ReadLineBuffer| Buffer: ReadLineBufferLabel    \ the label name from words that end in a ':' - do not change this
|ReadLineBuffer| Buffer: ReadLineBufferBody     \ the actual code - process this
|ReadLineBuffer| Buffer: ReadLineBufferComment  \ the comment - do not change at all, apart from aligning columns

: ReadOneLine ( -- f )    \ into the buffer
   NASMfileHandle @ 0= if  OpenReadFile  then
   ReadLineBuffer 1+ |ReadLineBuffer| NASMfileHandle @ Read-line abort" could not ReadOneLine from file"
   ( flag ) >r
   ReadLineBuffer c! \ count
   r> ( flag ) 
;

: CloseReadFile ( -- )  
   NASMfileHandle @ CLOSE-FILE drop 
;

variable line#

variable StartAddress

$200 constant |MyLineBuffer|
|MyLineBuffer| Buffer: MyLineBuffer

\ move to the given column number
: ToColumn ( u -- )   get-xy drop - 1 max 200 min  spaces ;

: -Leading ( a n - a' n' )
   2dup 
   over + swap ?do
      i c@ $20 > if  leave  then
      1 /string
   loop  
   0 max
;

\ type without the leadingand trailing spaces
: -LTtype ( $ -- )   -Leading -Trailing type ;

: %define ( -- )   
   SOURCE
   -Leading  BL split  2drop           \ remove the "%define"
   -Leading  BL split                  \ split the name and the argument string         
   2swap -Leading -Trailing 
   type    \ the argument comes first
   $03 ToColumn
   ."  constant "                      \ create a constant
   -LTtype             \ add its name
   postpone \                          \ skip over this line
;

: %assign ( -- )   
   SOURCE
   -Leading  BL split  2drop           \ remove the "%assign"
   -Leading  BL split                  \ split the name and the argument string         
   2swap -Leading -Trailing 
   type    \ the argument comes first
   $03 ToColumn
   ."  constant "                      \ create a constant
   type                                \ add its name
   postpone \                          \ skip over this line
;

: CPU ( -- )
   SOURCE 
   -Leading  BL split  2drop
   -Leading  BL split  2drop
   -Leading type
   postpone \                          \ skip over this line
;

: LineContains ( $ -- f )   \ return true is the line contains the given string   
   ReadLineBuffer count -Leading 2swap SEARCH >r 2drop r> 
;

: LineStartsWith ( $ -- f )   \ return true is the line starts with the given string, ignoring leading blanks    
   ReadLineBuffer count -Leading drop over
   compare 0=
;

: NumberOfLeadingSpaces ( -- c )    \ number of leading spaces in the given string
   0 ReadLineBuffer count over + swap ?do
      i c@ BL = not if  leave  then
      1+
   loop  
;

: Process_equ ( $ -- )  
   -Leading  BL split  ( name ) 2>r       \ save the name for later       
   -Leading  BL split 2drop               \ remove the "equ"
   -Leading -Trailing type           \ the argument comes first
   $03 ToColumn  ."  constant "           \ create a constant
   2r> -Leading -Trailing type            \ add its name
   MM
;

: Process_equ_As: ( $ -- )  
   ." : " 
   -Leading  BL split  ( name ) -LTtype   \ type the name       
   -Leading  BL split 2drop               \ remove the "equ"
   -LTtype                                \ the argument comes first
   \ $03 ToColumn  ."  constant "           \ create a constant
   MM    \ mark this line as processed
;

: Process_%if ( $ -- )
   s" " 2swap s" (" REPLACE     \ remove '('
   s" " 2swap s" )" REPLACE     \ remove ')'
   s" " 2swap s" %if" REPLACE     \ remove '%if'

   s" == " LineContains if
      -Leading  BL split -Trailing type  space    
      -Leading  BL split 2drop      \ remove the "=="
      -LTtype
      ."  ="
   else
      -LTtype
   then
   ."  [IF] "                       \ create an IF
   \ KK \ stop at this point to inspect the conversion
   MM
;

: commentOnly ( -- f )    
   ReadLineBuffer count -Leading drop c@ [char] ; = 
;

: commentStartPosition ( -- n )    
   ReadLineBuffer count [char] ; split 2swap 2drop swap drop 
;

: commentIsAtStartOfLine ( -- f )    
   ReadLineBuffer count drop c@ [char] ; = 
;

: crOnly ( -- f )    
   ReadLineBuffer count -Leading -Trailing  swap drop 0=
;

: LineBodyContains ( $ -- f )   \ return true is the line body text contains the given string   
   ReadLineBufferBody count -Leading 2swap SEARCH >r 2drop r> 
;

\ *****************************************************************************

15 constant dd_POSITION

: process_dd ( -- )
\   cr ." process_dw "  ReadLineBuffer count 7 attribute type  0 attribute
   ReadLineBufferBody count -Leading -Trailing
   s" " 2swap s" dd " REPLACE 
   INDENT_SPACES spaces
   begin 
      [char] , split 
      s"  $" 2swap s"  0x" REPLACE     \ replace "0x" by "$"   
      type 
      s" ," LineBodyContains if
         space
      else
         dd_POSITION ToColumn  
      then
      ." ,  "  
      dup
   while
      1 /string
   repeat
   2drop 
;

: process_dw ( -- )
\   cr ." process_dw "  ReadLineBuffer count 7 attribute type  0 attribute
   ReadLineBufferBody count -Leading -Trailing
   s" " 2swap s" dw " REPLACE 
   INDENT_SPACES spaces
   begin 
      [char] , split 
      s"  $" 2swap s"  0x" REPLACE     \ replace "0x" by "$"   
      type 
      s" ," LineBodyContains if
         space
      else
         dd_POSITION ToColumn  
      then
      ." w, "  
      dup
   while
      1 /string
   repeat
   2drop 
;

: process_db ( -- )
\   cr ." process_dw "  ReadLineBuffer count 7 attribute type  0 attribute
   ReadLineBufferBody count -Leading -Trailing
   s" " 2swap s" db " REPLACE 
   INDENT_SPACES spaces
   begin 
      [char] , split 
      s"  $" 2swap s"  0x" REPLACE     \ replace "0x" by "$"   
      type 
      s" ," LineBodyContains if
         space
      else
         dd_POSITION ToColumn  
      then
      ." c, "  
      dup
   while
      1 /string
   repeat
   2drop 
;

: LineUpComments ( -- )
   0     \ default
   commentStartPosition 0= if  exit  then
   commentStartPosition 20 < if  commentStartPosition +  exit  then
   $40 +
;

\ *****************************************************************************
\ Process opcodes
\ *****************************************************************************

: LineBodyNumberOfCommas ( -- n )
   0
   ReadLineBufferBody count over + swap ?do     
      i c@ [char] , = if  1+  then
   loop
;

: Process_opcodes ( -- )
   s" mov" LineStartsWith if
      ReadLineBufferBody count type
      MM    \ mark this line as done
   then
;

\ *****************************************************************************
\ Process the whole line of NASM code
\ *****************************************************************************

: ShowLineFiltered ( -- )
   NumberOfLeadingSpaces spaces
   ReadLineBufferLabel count ?dup if  1 attribute type  0 attribute  space  else  drop  then    \ label
   ReadLineBufferBody count -Leading ( dup  2 u.r  ."  | " ) 5 attribute type                    \ body
   \ 4 attribute ReadLineBufferComment count type                                   \ comment
   
   $60 ToColumn 5 attribute ." <<<<<<<<<<<<< " line# @ 4 u.r  space                   \ line info   
   0 attribute
   ekey #27 = if  CloseReadFile cr ." ok " cr  abort  then
;

: ProcessNASMline ( -- )  
   0 delayedStop !
   0 markAsDone !

   ReadLineBuffer count 
   s" \ " 2swap  s" ; " REPLACE     \ first check for a ';' with a space after it
   s" \ " 2swap  s" ;" REPLACE      \ then check for a ';' with no following space
   s"  $" 2swap s"  0x" REPLACE     \ replace "0x" by "$"

   [char] \ split  
   2swap  -Trailing ReadLineBufferComment place     \ save the part including and after the "\"    

   \ ********************************* labels **********************************
   2dup s" :" SEARCH >r 2drop r> if
      \ 3 attribute  2dup cr type  0 attribute
      [char] : split 1+ -Trailing ReadLineBufferLabel place  \ Note the 1+ to include the ':' 
      1 /string -Trailing                                    \ Note the 1 /string to remove the ':'
      \ KK
   else
      0 ReadLineBufferLabel c!
      -Trailing 
   then
   \ make any "global" changes to the body text here :
   s" HERE gdt0 -" 2swap s" $ - gdt0" REPLACE    
   ReadLineBufferBody place        

   \ ********************************* body ***********************************
   cr    \ start outputting on a new line
 
   \ ." N= " commentStartPosition . space 
   
   s" CPU 386" LineContains if
      ReadLineBufferBody count evaluate 
      MM    \ mark this line as done
   then

   s" %define" LineContains if  
      ReadLineBufferBody count evaluate 
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" %assign" LineContains if  
      ReadLineBufferBody count evaluate 
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" equ" LineContains if  
      ReadLineBufferBody count Process_equ 
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" %if " LineContains if  
      ReadLineBufferBody count Process_%if 
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" %else" LineContains if  
      ReadLineBufferBody count 
      s" [ELSE] " 2swap s" %else " REPLACE   \ look for a following space first, 
      s" [ELSE] " 2swap s" %else" REPLACE    \ then, if no following space, add one
      type
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" %endif" LineContains if  
      ReadLineBufferBody count 
      s" [THEN] " 2swap s" %endif " REPLACE   \ look for a following space first, 
      s" [THEN] " 2swap s" %endif" REPLACE    \ then, if no following space, add one
      type
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" dd " LineStartsWith if
      process_dd
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" dw " LineStartsWith if
      process_dw
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" db " LineStartsWith if
      process_db
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   Process_opcodes


   \ ********************************* debug ***********************************
      
   markAsDone @ 0=  if
      ReadLineBufferBody count swap drop 0= not if
         \  show the filtered input line, split into three parts : label, body ( and comment added later )
         ShowLineFiltered
      then
   then

   delayedStop @ if
      ekey #27 = if  CloseReadFile cr ." ok " cr  abort  then
      0 delayedStop !
   then   

   \ ******************************** comments ********************************
   LineUpComments ToColumn
   ReadLineBufferComment count type                                     \ comment
   \ crOnly if  cr  then
;

: ProcessNASMfile  ( -- )
   OpenReadFile   
   MAX_NUMBER_OF_LINES 0 do
      1 line# +!
      ['] ReadOneLine catch if  cr ." caught!!!"  then  0= if  leave  then
      ProcessNASMline
   loop  
   CloseReadFile
;

: cf2018 ProcessNASMfile ;

\\

cr 
CPU 386 ; Assemble instructions for the 386 instruction set

%define FORCE_800x600_VESA  0   ; true to force 800 x 600 x 16 bits for testing in  bochs

%define START_BLOCK_NUMBER      20  ; must be an even number
%define SIZE_OF_FONT_IN_BLOCKS   6
%define OFFSET_OF_FONT          ( ( START_BLOCK_NUMBER - SIZE_OF_FONT_IN_BLOCKS ) * 0x400 )

