\ cf2018.f 2018 Jun 10zzz  convert NASM file to Forth 

decimal 

4 constant INDENT_SPACES

: 2**   ( c -- u )   \ takes 2 to the power c
   1 swap lshift
;

: HostVersion  s" cf2018 V1.0 2018 Jun 10" ;

\ The strings below are used to by MSwindow.f
: Acronym  HostVersion  drop 6 ;   \ must not have any spaces in the acronym
create AboutHeader  ," About cf2018"
create AboutLine0   ," NASM to Forth assembler conversion program "
create AboutLine1   HostVersion STRING,
create AboutLine2   ,"  2018 Inventio Software Ltd"
create AboutLine3   ," www.inventio.co.uk"

create PictureFilename1  ," StartScreen.gif"
create PictureFilename2  ," StartScreen2.gif"
create HelpFilename      ," cf2018_help.chm"

#1024 constant cMainWindowSizeX
#512  constant cMainWindowSizeY

\ we define a drawing area area in memory - we write to it then display it fitted to the current Window size
#1024 constant DRAWING_SIZE_X    \ the fixed horizontal size in pixels of the drawing area
#768  constant DRAWING_SIZE_Y    \ the fixed vertical size in pixels of the drawing area
#32   constant BITS_PER_PIXEL    \ the fixed colour depth of the drawing area in pixels
BITS_PER_PIXEL 8 / constant OCTETS_PER_PIXEL    \ the fixed colour depth of the drawing area in octets

37 constant Col1           \ display column tabs
Col1 16 +  constant Col2
Col2 27 +  constant Col3
Col3 20 +  constant Col4

: DefaultSaveFilename   s" cf2018SaveFile" ;

$200 constant |SaveFilename|
|SaveFilename| Buffer: SaveFilename

DefaultSaveFilename SaveFilename place

\ *****************************************************************************

include LogToFile.f

\ *****************************************************************************

\ for .stats, count up how many of each type of action is in the file
variable #args=0
variable #args=1
variable #args=2
variable #args=3
variable #args=X
variable #%macros
variable #%defines
variable #equs
variable #%assigns

: tally ( a -- )  1 swap +! ;   

decimal

variable delayedStop
: KK ( -- )   -1 delayedStop ! ;

variable markAsDone
: MM ( -- )   -1 markAsDone ! ;

10000 constant MAX_NUMBER_OF_LINES

: MyFileName ( -- $ )   s" cf2018.nasm" ;

variable NASMfileHandle
0 NASMfileHandle !

: OpenReadFile 
   MyFileName r/o open-file 
   if ." Couldn't open read file "  MyFileName type space  #2001 throw  then
   NASMfileHandle !
;

$200 constant |ReadLineBuffer| 
|ReadLineBuffer| Buffer: ReadLineBuffer         \ the original complete line

|ReadLineBuffer| Buffer: ReadLineBufferLabel    \ the label name from words that end in a ':' - do not change this
|ReadLineBuffer| Buffer: ReadLineBufferBody     \ the actual code - process this
|ReadLineBuffer| Buffer: ReadLineBufferComment  \ the comment - do not change at all, apart from aligning columns

: CloseReadFile ( -- )  
   NASMfileHandle @ CLOSE-FILE drop 
   0 NASMfileHandle !
;

: ReadOneLine ( -- f )    \ into the buffer
   \ NASMfileHandle @ 0= if  OpenReadFile  then
   ReadLineBuffer 1+ |ReadLineBuffer| NASMfileHandle @ Read-line if  
      CloseReadFile  cr ." could not ReadOneLine from file"  #12090 throw
   then
   ( flag ) >r
   dup $FF > if 
      -Log
      cr 2 attribute  ." ReadOneLine ReadLineBuffer count > $FF, count =  " dup .  0 attribute
      KK    \ Note : KK does not work as expected while logging!
      +Log

   then
   $0000000FF and       \ this is needed to avoid an ACCESS_VIOLATION when cf2018 is run for the first time after loading (!?)
   ReadLineBuffer c! \ count
   r> ( flag ) 
;

variable line#

variable StartAddress

$200 constant |MyLineBuffer|
|MyLineBuffer| Buffer: MyLineBuffer


: AskForKey ( -- )
   ekey #27 = if  CloseReadFile cr ." ok " cr  abort  then
;

\ move to the given column number
: ToColumn ( u -- )   get-xy drop - 1 max 200 min  spaces ;

: -Leading ( a n - a' n' )
   2dup 
   over + swap ?do
      i c@ $20 > if  leave  then
      1 /string
   loop  
   0 max
;

\ type without the leadingand trailing spaces
: -LTtype ( $ -- )   -Leading -Trailing type ;

: LineBodyNumberOfCommas ( -- n )
   0
   ReadLineBufferBody count over + swap ?do     
      i c@ [char] , = if  1+  then
   loop
;

: NumberOfOpenBrackets ( $ -- n )
   0 -rot
   over + swap ?do     
      i c@ [char] ( ( ) = if  1+  then
   loop
;

$200 Buffer: StringPad0[]
$200 Buffer: StringPad1[]
$200 Buffer: StringPad2[]
$200 Buffer: StringPad3[]

9 2** constant |StringPadOut|             \ the size of our output buffer - must be a power of 2
|StringPadOut| Buffer: StringPadOut[]     \ our send buffer
variable #StringPadOut                    \ how many bytes sent so far

: bufferEmit ( c -- )
   StringPadOut[] #StringPadOut @ |StringPadOut| 1- and + c!    \ store the character in the send buffer
   1 #StringPadOut +!             \ increment the count
;

: bufferType ( $ -- )   over + swap ?do  i c@ bufferEmit  loop ;
: -LTbufferType ( $ -- )   -Leading -Trailing bufferType ;

: bufferSpace ( -- )  BL bufferEmit ;

: InitStringPadOut ( -- )   0 #StringPadOut ! ;

: NumberOf+ ( -- n )  0 StringPad0[] count over + swap ?do  i c@ [char] + = if  1+  then  loop ;
: NumberOf- ( -- n )  0 StringPad0[] count over + swap ?do  i c@ [char] - = if  1+  then  loop ;
: NumberOf* ( -- n )  0 StringPad0[] count over + swap ?do  i c@ [char] * = if  1+  then  loop ;
: NumberOf/ ( -- n )  0 StringPad0[] count over + swap ?do  i c@ [char] / = if  1+  then  loop ;

: $NumberOf, ( $-- n )  0 -rot over + swap ?do  i c@ [char] , = if  1+  then  loop ;

: NumberOfOperators ( $ -- n )
   0  -rot
    over + swap ?do     
      i c@ >r
      r@ [char] + = if  1+  then
      r@ [char] - = if  1+  then
      r@ [char] * = if  1+  then
      r@ [char] / = if  1+  then
      r> drop
   loop
;

: ReOrderOnePair ( $ -- $ )
   InitStringPadOut
   StringPad0[] place
   0 
   NumberOf+ if  drop [char] +  then
   NumberOf- if  drop [char] -  then
   NumberOf* if  drop [char] *  then
   NumberOf/ if  drop [char] /  then
   >r                                     \ save the character for later
   StringPad0[] count r@ split
   -LTbufferType 
   bufferSpace 
   -Leading 1 /string   \ remove the 'slit' character from the remaining string
   -LTbufferType bufferSpace r@ bufferEmit 
   r> drop
   StringPadOut[] #StringPadOut @
   -Leading -Trailing
;

\ convert the ASCII string a , n to a double number, using the current base
: Dnumber ( a n -- d )
   2dup 2>r NUMBER? case
   0 of 
\     RED.  ."  Invalid number string : " 2r@ Dump  NORMAL.  0.   ( key drop ) 
      #10023 throw
   endof
   1 of  s>d  endof
   2 of  endof
   cr ." what?"
   endcase
   2r> 2drop
;

: IsNumber? ( $ -- f )
\   dup 0= if  2drop  -2  exit  then     \ ignore an empty string
\   cr ."       IsNumber? '" 2dup type ." '" space 
   -Leading -Trailing
   ['] Dnumber catch if  0  else  -1  then
   >r 2drop  r>  
\   ."   = "    dup . space 
;

variable vOnlyAnumber

: NumberOfNonNumbers ( $ -- f )   \ returns 0 if the string contains only numbers
   0 vOnlyAnumber !
   20 0 do
      -Leading -Trailing  BL split  \ 1 /string 
      dup 0= if  leave  then
      IsNumber? \ dup -2 = if  leave  then
      0= if  1 vOnlyAnumber +! then
   loop
   2drop
   2drop
   vOnlyAnumber @
;

: ttNumberOfNonNumbers ( -- )
   cr s" 1 2 3 4" 2dup type space NumberOfNonNumbers ."  = " .
   cr s" 1 2 X 4" 2dup type space NumberOfNonNumbers ."  = " .
   cr s" 1 Hello1234 X 4" 2dup type space NumberOfNonNumbers ."  = " .
;

: ReOrderParams ( $ -- $' )
   2dup NumberOfOpenBrackets 
\   1 atitribute dup . 0 attribute space
   case
   0 of    endof  \ no brackets, so do nothing
   1 of  \ 1 pair, remove them and put the operator at the end
      2dup NumberOfOperators 0= if
         s" " 2swap s" (" REPLACE     \ remove '('
         s" " 2swap s" )" REPLACE     \ remove ')' 
      else
         -Leading -Trailing 
         [char] ( ( ) split  2drop
         1 /string  1-  -Leading -Trailing 
         ReOrderOnePair
      then
      \ 5 attribute cr 2dup type 0 attribute cr
     \  AskForKey
   endof  
   2 of  
      2dup NumberOfOperators 0= if
         s" " 2swap s" (" REPLACE     \ remove '('
         s" " 2swap s" )" REPLACE     \ remove ')' 
      else
         -Leading -Trailing 
         [char] ( ( ) split  2drop
         1 /string  1-  -Leading -Trailing 
         ReOrderOnePair StringPad1[] place
         StringPad1[] count [char] ) split 2swap -Leading -Trailing 1 /string StringPad2[] place  \ the last part
         ReOrderOnePair  -Leading -Trailing 1 /string StringPad3[] place
         \ cr 2dup type cr 
         StringPad3[] count StringPad1[] place
         s"  " StringPad1[] append
         StringPad2[] count StringPad1[] append
         StringPad1[] count
      then
     \ 5 attribute cr 2dup type 0 attribute cr
     \ AskForKey
   endof
   AskForKey
   endcase
;

: %assign ( -- )   
   #%assigns tally
   SOURCE
   -Leading  BL split  2drop           \ remove the "%assign"
   -Leading  BL split                  \ split the name and the argument string         
   2swap -Leading -Trailing 
   type    \ the argument comes first
   $03 ToColumn
   ."  constant "                      \ create a constant
   type                                \ add its name
\   postpone \                          \ skip over this line
;

: DIGIT> ( c -- c )   [char] 0 - 0 max  dup 9 > if  #223 and 7 -  then  $0F and ;   \ convert the hex ASCII digit 0 to 9 or A to F or a to z to a 4 bit number
: HexC@ ( a -- c )   dup c@ DIGIT> $10 * swap 1+ c@ DIGIT> or ;                     \ convert the BigEndian hex value byte pair at a to an 8 bit number

: %macro ( f -- )    \ remove the macro argument count if f is true
   #%macros tally
   >r
\   ." ALSO ASSEMBLER DEFINITIONS"   
   ." : " 
   ReadLineBufferBody count -Leading  BL split   ( %macro ) 2drop
    BL split
    2drop \ -Leading -Trailing drop c@ DIGIT> space .
   \ BL split
   \ -Leading -Trailing 3 attribute type  0 attribute  \ name
   -Leading -Trailing 2dup 
   r> if
      1- 0 max    \ remove the last character, the number of parameters
   then
   BL split -LTtype  \ the macro's name
   2>r    
   + 1- c@ DIGIT>    \ convert the number of parameters digit to a value
   \ dup space . 
   space
   case
   0 of   ." ( -- )"  endof
   1 of   ." ( n -- )"  endof
   2 of   ." ( n n -- )"  endof
   3 of   ." ( n n n -- )"  endof
   endcase
   3 spaces  2r> 
   ReOrderParams
   -LTtype   \ the macro's content
   MM    \ mark this line as processed
;

: %endmacro ( -- )
   ."  ; " 
\   ." PREVIOUS FORTH DEFINITIONS" 
   MM    \ mark this line as processed
;

: %define ( -- )
   #%defines tally   
   ReadLineBufferBody count 
   -Leading  BL split  2drop           \ remove the "%define"
   -Leading  BL split                  \ split the name and the argument string         
   2swap -Leading -Trailing 
   ReOrderParams
\   cr ." After ReOrderParams = '" 2dup type ." '" space
   2dup NumberOfNonNumbers 0= if
      -LTtype    \ the argument comes first
      $03 ToColumn
      ." constant "                      \ create a constant
      -LTtype             \ add its name
   else
      2drop
      2drop 
      0 %macro
      %endmacro
   then
;


: %ifndef ( -- )
   s" [UNDEFINED]" ReadLineBufferBody count s" %ifndef" REPLACE -LTtype    \ remove '%if' 
   ."  [IF]"
   MM    \ mark this line as processed
;

: process_CPU ( -- )
   SOURCE 
   -Leading  BL split  2drop
   -Leading  BL split  2drop
   ." \ " -Leading type
\   postpone \                          \ skip over this line
;

: LineContains ( $ -- f )   \ return true is the line contains the given string   
   ReadLineBuffer count -Leading 2swap SEARCH >r 2drop r> 
;

: LineStartsWith ( $ -- f )   \ return true is the line starts with the given string, ignoring leading blanks    
   ReadLineBuffer count -Leading drop over
   compare 0=
;

: NumberOfLeadingSpaces ( -- c )    \ number of leading spaces in the given string
   0 ReadLineBuffer count over + swap ?do
      i c@ BL = not if  leave  then
      1+
   loop  
;

: Process_equ ( $ -- )
   #equs tally
   -Leading  BL split  ( name ) 2>r       \ save the name for later       
   -Leading  BL split 2drop               \ remove the "equ"
   -LTtype           \ the argument comes first
   $03 ToColumn  ."  constant "           \ create a constant
   2r> -LTtype            \ add its name
   MM
;

: xxxProcess_equ_As: ( $ -- )  
   ." : " 
   -Leading  BL split  ( name ) -LTtype   \ type the name       
   -Leading  BL split 2drop               \ remove the "equ"
   -LTtype                                \ the argument comes first
   \ $03 ToColumn  ."  constant "           \ create a constant
   MM    \ mark this line as processed
;

: Process_%if ( $ -- )
   s" " 2swap s" (" REPLACE     \ remove '('
   s" " 2swap s" )" REPLACE     \ remove ')'
   s" " 2swap s" %if" REPLACE     \ remove '%if'

   s" == " LineContains if
      -Leading  BL split -Trailing type  space    
      -Leading  BL split 2drop      \ remove the "=="
      -LTtype
      ."  ="
   else
      -LTtype
   then
   ."  [IF] "                       \ create an IF
   \ KK \ stop at this point to inspect the conversion
   MM
;

: commentAtStartOfLine ( -- f )    
   ReadLineBuffer count -Leading drop c@ [char] ; = 
;

: commentStartPosition ( -- n )    
   ReadLineBuffer count [char] ; split 2swap 2drop swap drop 
;

: commentOnly ( -- f ) 
   ReadLineBuffer count drop c@ [char] ; = if  -1  exit  then
   ReadLineBuffer count -Leading [char] ; split  2swap 2drop 
   -Leading swap drop 0=
;

: commentIsAtStartOfLine ( -- f )    
   ReadLineBuffer count drop c@ [char] ; = 
;

: crOnly ( -- f )    
   ReadLineBuffer count -Leading -Trailing  swap drop 0=
;

: LineBodyContains ( $ -- f )   \ return true is the line body text contains the given string   
   ReadLineBufferBody count -Leading 2swap SEARCH >r 2drop r> 
;

\ *****************************************************************************

15 constant dd_POSITION

: process_dd ( -- )
\   cr ." process_dw "  ReadLineBuffer count 7 attribute type  0 attribute
   ReadLineBufferBody count -Leading -Trailing
   s" " 2swap s" dd " REPLACE 
   INDENT_SPACES spaces
   begin 
      [char] , split 
      s"  $" 2swap s"  0x" REPLACE     \ replace "0x" by "$"   
      type 
      s" ," LineBodyContains if
         space
      else
         dd_POSITION ToColumn  
      then
      ." ,  "  
      dup
   while
      1 /string
   repeat
   2drop 
;

: process_dw ( -- )
\   cr ." process_dw "  ReadLineBuffer count 7 attribute type  0 attribute
   ReadLineBufferBody count -Leading -Trailing
   s" " 2swap s" dw " REPLACE 
   INDENT_SPACES spaces
   begin 
      [char] , split 
      s"  $" 2swap s"  0x" REPLACE     \ replace "0x" by "$"   
      type 
      s" ," LineBodyContains if
         space
      else
         dd_POSITION ToColumn  
      then
      ." w, "  
      dup
   while
      1 /string
   repeat
   2drop 
;

: process_db ( -- )
\   cr ." process_dw "  ReadLineBuffer count 7 attribute type  0 attribute
   ReadLineBufferBody count -Leading -Trailing
   s" " 2swap s" db " REPLACE 
   INDENT_SPACES spaces
   begin 
      [char] , split 
      s"  $" 2swap s"  0x" REPLACE     \ replace "0x" by "$"   
      type 
      s" ," LineBodyContains if
         space
      else
         dd_POSITION ToColumn  
      then
      ." c, "  
      dup
   while
      1 /string
   repeat
   2drop 
;

: LineUpComments ( -- )
   commentAtStartOfLine if  exit  then    \ because ToColumn alway leaves at least one space
   commentStartPosition 0= if  exit  then
   commentStartPosition 20 < if  commentStartPosition ToColumn  exit  then
   $40 ToColumn
;

\ *****************************************************************************
\ Process opcodes
\ *****************************************************************************

: ShowReadLineBufferBody ( -- )   ReadLineBufferBody count -LTtype ;

: ProcessOpCodeArgs ( opCodeLength -- )
   drop
   INDENT_SPACES spaces \ ToColumn 
\   5 attribute ." | " ReadLineBufferBody count type 0 attribute
   
  \ ReadLineBufferBody count $NumberOf, 0= not if 
  \    ReadLineBufferBody count [char] , split 2drop 
  \ else
   ReadLineBufferBody count -Leading -Trailing  BL split 2drop 
  \ then
   -Leading -Trailing swap drop 0= 
   if    \ no args at all
      0
   else
      ReadLineBufferBody count $NumberOf, 1+ 
   then 
   \ cr ." ProcessOpCodeArgs = " dup . 

   case
   0 of  ShowReadLineBufferBody  #args=0 tally  endof   \ 0 args, so just output the text
   1 of     \ 1 arg, no commas
      #args=1 tally
      ReadLineBufferBody count -Leading -Trailing
      \ cr 1 attribute 2dup type 0 attribute cr
      BL split 2>r  
      -LTtype space  \ the one argument
      2r> -LTtype    \ opcode   
   endof   
   2 of     \ 2 args, 1 comma 
      #args=2 tally  
      ReadLineBufferBody count -Leading
     \ cr 1 attribute 2dup type 0 attribute cr
      BL split 2>r 
      [char] , split  2swap 1 /string  -LTtype  \ the source arg Note : the 2swap reverses the order of the two args  
      space -LTtype space                       \ the destination arg
      2r> -LTtype    \ opcode    
   endof   
   3 of     \ 3 args, 2 commas 
      #args=3 tally  
      ReadLineBufferBody count -Leading
      \ cr 1 attribute 2dup type 0 attribute cr
      BL split 2>r 
      [char] , split  2swap 1 /string  2>r   \ the source arg Note : the 2swap reverses the order of the two args  
      [char] , split  2swap 1 /string  -LTtype  \ the source arg Note : the 2swap reverses the order of the two args  
      2r> -LTtype
      space -LTtype space                       \ the destination arg
      2r> -LTtype    \ opcode    
   endof   
      ReadLineBufferBody count type    \ not known, so just output the text
      #args=X tally
      KK
   endcase
   0 attribute
\   .s  KK
;

8 2** constant |OpCodeName| 
|OpCodeName| Buffer: OpCodeName[]

: xxx$MatchesOpCode ( $ -- f )
   s"        " OpCodeName[]  place  \ add a trailing space
   6 min  OpCodeName[]  place
    ReadLineBufferBody count -Leading -Trailing  BL split 2swap 2drop \ 1+    \ do not include one space after the opcode
    OpCodeName[] count drop over compare 0=
;

\ a linked list of counted strings of opcode names. Note : ," adds a trailing null
create OpCodeList[] 
   ," mov" ," lea" ," push"  ," pop" ," xchg" ," cli" ," call" ," jmp" 
   ," times" ," align" ," org" ," nop" ," xor" ," lgdt" ," out" ," in"
   ," and" ," bts" ," or" ," jnz" ," jz" ," int" ," inc" ," dec" 
   ," cmp" ," add" ," sub" ," jc" ," jne" ," jna" ," je" ," cli"
   ," ret" ," pusha" ," popa" ," pushf" ," popf" ," lodsw" ," loop" ," lodsb"
   ," stosw" ," iret" ," pushad" ," mul" ," div" ," rol" ," shl" ," popad"
   ," retf" ," repne" ," std" ," shr" ," rep" ," cld" ," test" ," sar"
   ," ror"  ," js" ," jns" ," imul" ," jnc" ," jng" ," neg" ," not"
   ," stosd" ," pushfd" ," incbin" ," lodsd"
   ," eax" ," ebx"   \ not really opcodes...
   ," _DUP_" ," _DROP_" ," EMIT_IMM" ," EMIT_IMM(" ," _SWAP_" ," LOAD_RELATIVE_ADDRESS"
   0 c,  \ terminating null 

2variable matchString
variable lastDepth
variable lastOpCodeLengthFound

: MatchOpCode ( $ -- n f )  \ return true if the given string is recognised, with opCode name length underneath
\   depth 2- lastDepth !
   dup >r dup lastOpCodeLengthFound !
   \ s"        " OpCodeName[]  place  \ add a trailing space
   |OpCodeName| 1- min  OpCodeName[]  place
   \ cr OpCodeName[] count dump
   0 
   OpCodeList[]
   $100 0 do 
 \     cr dup count type space 
      dup count OpCodeName[] count compare 0= if  
         swap 1+ swap  
         leave  
      then
      count + 
      1+    \ Note : ," adds a trailing null, so we must skip over it in the linked list
      dup c@ 0= if  leave  then 
   loop
   drop
\   depth 5 > if  cr ." WWWWWWWWWW         " depth . KK then
   r> swap
;

: ttMatchOpCode ( -- )  
   cr s" lea" MatchOpCode . .
   cr s" leX" MatchOpCode . .   
;

variable #blankLines

: Process_opcodes ( -- )
   ReadLineBufferBody count -Leading -Trailing swap drop 0= if  1 #blankLines +!  exit  then    \ blank line, so ignore it

   s" o16" LineContains if  ShowReadLineBufferBody  MM  exit  then   \ do not reverse the params
   s" o32" LineContains if  ShowReadLineBufferBody  MM  exit  then   \ do not reverse the params   
   
   \ cr ." Process_opcodes" 
   ReadLineBufferBody count -Leading -Trailing  BL split 2swap 2drop  \ 1+   \ the 1+ adds a trailing space
   MatchOpCode if
      ProcessOpCodeArgs
      MM    \ mark this line as done
   else
      drop
      cr ." Process_opcodes not found!!!" 
      cr 2 attribute ReadLineBufferBody count type 0 attribute
      \ KK
   then
;

\ *****************************************************************************
\ Process the whole line of NASM code
\ *****************************************************************************

: ShowLineFiltered ( -- )
   NumberOfLeadingSpaces spaces
   ReadLineBufferLabel count ?dup if  1 attribute type  0 attribute  space  else  drop  then    \ label
   ReadLineBufferBody count -Leading ( dup  2 u.r  ."  | " ) 5 attribute type                    \ body
   \ 4 attribute ReadLineBufferComment count type                                   \ comment
   
   $60 ToColumn 5 attribute ." <<<<<<<<<<<<< " line# @ 4 u.r  space                   \ line info   
   0 attribute
   AskForKey
;

: QQ  .s depth lastDepth @ = not if ." QQQQQQQQQQ" KK  then ;

: ProcessNASMlineNoComments ( $ -- )  
   \ cr ." ProcessNASMlineNoComments "
   \ ********************************* labels **********************************
   2dup s" :" SEARCH >r 2drop                   \ the line contains a ':'
   2dup s" fs:" SEARCH >r 2drop r> 0=  r> and   \ but not an 'fs:'
   if
      \ 3 attribute  2dup cr type  0 attribute
      [char] : split -Trailing ReadLineBufferLabel place    \ 
      1 /string -Trailing                                   \ Note the 1 /string to remove the ':'
      \ KK
   else
      0 ReadLineBufferLabel c!
      -Trailing 
   then

   ReadLineBufferLabel count swap drop if
      ." LABEL " ReadLineBufferLabel count type space 
   then

   \ make any "global" changes to the body text here :
   \  to                    from      replace : Note the order of the parameters
   s" HERE gdt0 -" 2swap s" $ - gdt0" REPLACE    
   s" 1 RELOC_BIT lshift" 2swap s" 1 << RELOC_BIT" REPLACE    
   s" 0 [ESI]" 2swap s" [ esi ]" REPLACE    
   s" [ECX] [EDX]" 2swap s" [ ecx + edx ]" REPLACE    
   s" RELOCATED []" 2swap s" ( ( ( %1 - $$ ) + RELOCATED ) )" REPLACE
  \ s" 32BIT: " 2swap s" o32 " REPLACE  \ ToDo : fix this!!!  
   
   ReadLineBufferBody place        

   s" o16" LineContains if
      s" 16BIT: " ReadLineBufferBody count s" o16 " REPLACE  
      ReadLineBufferBody place 
   then

   s" o32" LineContains if
      s" 32BIT: " ReadLineBufferBody count s" o32 " REPLACE  
      ReadLineBufferBody place 
   then

   s" [BITS" LineContains if
      s" #16 BitsMode !" ReadLineBufferBody count s" [BITS 16]" REPLACE  ReadLineBufferBody place
      s" #32 BitsMode !" ReadLineBufferBody count s" [BITS 32]" REPLACE  ReadLineBufferBody place
      ShowReadLineBufferBody
      MM    \ mark this line as done      
   then

   \ ********************************* body ***********************************
   \ ." N= " commentStartPosition . space 

   s" CPU 386" LineContains if
      process_CPU 
      MM    \ mark this line as done
   then

   s" %define" LineContains if  
      %define 
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" %assign" LineContains if  
      %define \ %assign
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then
   
   s" %macro" LineContains if  
      -1 %macro
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" %endmacro" LineContains if  
      %endmacro
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" equ" LineContains if  
      ReadLineBufferBody count Process_equ 
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" %if " LineContains if  
      ReadLineBufferBody count Process_%if 
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" %else" LineContains if  
      ReadLineBufferBody count 
      s" [ELSE] " 2swap s" %else " REPLACE   \ look for a following space first, 
      s" [ELSE] " 2swap s" %else" REPLACE    \ then, if no following space, add one
      type
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" %endif" LineContains if  
      ReadLineBufferBody count 
      s" [THEN] " 2swap s" %endif " REPLACE   \ look for a following space first, 
      s" [THEN] " 2swap s" %endif" REPLACE    \ then, if no following space, add one
      type
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" dd " LineStartsWith if
      process_dd
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" dw " LineStartsWith if
      process_dw
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" db " LineStartsWith if
      process_db
      \ KK \ stop at this point to inspect the conversion
      MM    \ mark this line as done
   then

   s" %ifndef" LineStartsWith if
      %ifndef
   then

   markAsDone @ 0= if   \ only do this if we have not already processed the line
      Process_opcodes
   then

   \ ********************************* debug ***********************************
      
   markAsDone @ 0=  if
      ReadLineBufferBody count swap drop 0= not if
         \  show the filtered input line, split into three parts : label, body ( and comment added later )
         ShowLineFiltered
      then
   then

   delayedStop @ if
      AskForKey
      0 delayedStop !
   then   
;

: ProcessNASMline ( -- ) 
   0 delayedStop !
   0 markAsDone !

   depth lastDepth !

   cr    \ start outputting on a new line
   ReadLineBuffer count 
\    cr 2dup 1 attribute type 0 attribute
   commentOnly not if
      s" \ " 2swap  s" ; " REPLACE     \ first check for a ';' with a space after it
      s" \ " 2swap  s" ;" REPLACE      \ then check for a ';' with no following space
      s"  $" 2swap s"  0x" REPLACE     \ replace " 0x" by " $"
      s" $" 2swap s" 0x" REPLACE     \ replace "0x" by "$"
      [char] \ split  
      2swap  -Trailing ReadLineBufferComment place     \ save the part including and after the "\"    
      ProcessNASMlineNoComments
   else
      s" \ " 2swap  s" ; " REPLACE     \ first check for a ';' with a space after it
      s" \ " 2swap  s" ;" REPLACE      \ then check for a ';' with no following space
      s"  $" 2swap s"  0x" REPLACE     \ replace "0x" by "$"
      ReadLineBufferComment place
   then
   
   \ ******************************** comments ********************************
   LineUpComments
   ReadLineBufferComment count type                                     \ comment
 
   \ crOnly if  cr  then
;

: .stats ( -- )
  cr ." #blankLines = "  #blankLines @ .  
  cr ." #args=0     = "  #args=0     @ .  
  cr ." #args=1     = "  #args=1     @ .   
  cr ." #args=2     = "  #args=2     @ .   
  cr ." #args=3     = "  #args=3     @ .   
  cr ." #args=X     = "  #args=X     @ .  
  cr ." #%macros    = "  #%macros    @ .  
  cr ." #%defines   = "  #%defines   @ .  
  cr ." #equs       = "  #equs       @ .  
  cr ." #%assigns   = "  #%assigns   @ .  
;                                   

: ProcessNASMfile  ( -- )
   0 #blankLines !
   0 #args=0 !    
   0 #args=1 !   
   0 #args=2 !   
   0 #args=3 !   
   0 #args=X !   
   0 #%macros !   
   0 #%defines !   
   0 #equs ! 
   0 #%assigns !

   InitLogBuffer
   cr    \ new line on screen, but not in Log file
   +Log
   ." \ ProcessNASMfile  text output from NASM file : " MyFileName type   \ title on screen and in Log file
   cr
   cr ." ONLY FORTH ALSO ASSEMBLER DEFINITIONS"
   cr
   OpenReadFile   
   MAX_NUMBER_OF_LINES 0 do
      1 line# +!
      ['] ReadOneLine catch if  cr ." caught!!!"  then  0= if  leave  then
\      cr ." >>>> line# = "  line# @ 4 u.r
      ProcessNASMline
\      cr ." <<<<<" 
   loop  
   CloseReadFile
   cr
   cr ." PREVIOUS FORTH DEFINITIONS"
   cr
   -Log
    MyFileName SaveFilename place
   s" .f" SaveFilename append
   LogToFile
   cr ." Done." cr
   .stats
;

: cf2018 ProcessNASMfile ;

: cf2018less  ( -- )
   InitLogBuffer
   cr    \ new line on screen, but not in Log file
   +Log
   ." \ ProcessNASMfile  text output from NASM file : " MyFileName type   \ title on screen and in Log file
   cr
   cr ." ONLY FORTH ALSO ASSEMBLER DEFINITIONS"
   cr
   OpenReadFile   
   MAX_NUMBER_OF_LINES 0 do
      1 line# +!
      ['] ReadOneLine catch if  cr ." caught!!!"  then  0= if  leave  then
\      cr ." >>>> line# = "  line# @ 4 u.r
\     ProcessNASMline
\      cr ." <<<<<" 
   loop  
   CloseReadFile
   cr
   cr ." PREVIOUS FORTH DEFINITIONS"
   cr
   -Log
    MyFileName SaveFilename place
   s" .f" SaveFilename append
   LogToFile
   cr ." Done." cr
;


\\

cr 
CPU 386 ; Assemble instructions for the 386 instruction set

%define FORCE_800x600_VESA  0   ; true to force 800 x 600 x 16 bits for testing in  bochs

%define START_BLOCK_NUMBER      20  ; must be an even number
%define SIZE_OF_FONT_IN_BLOCKS   6
%define OFFSET_OF_FONT          ( ( START_BLOCK_NUMBER - SIZE_OF_FONT_IN_BLOCKS ) * 0x400 )

