\ ProcessNASMfile  text output from NASM file : cf2018.nasm

ONLY FORTH ALSO ASSEMBLER DEFINITIONS

\ cf2018.nasm 2018 Jun 14  "mischief-ages"  with arrow, Enter and Escape keys
\ colorForth for 80x86 PC for NASM , with 1024x768 and 800x600 graphics options
\ Adapted by Howerd Oakford from code by :
\ Chuck Moore : inventor, MASM
\ Mark Slicker : ported to GNU Assembler
\ Peter Appelman : ported to NASM with qwerty keyboard
\ John Comeau : BIOS boot from ClusterFix
\ and others... Thanks to all!!!
\ Feedback welcome : howerd@inventio.co.uk www.inventio.co.uk

\       \ Assemble instructions for the 386 instruction set

0  constant FORCE_80$600_VESA                                   \ true to force 800 x 600 x 16 bits for testing in  bochs

20 constant START_BLOCK_NUMBER                                  \ must be an even number
6  constant SIZE_OF_FONT_IN_BLOCKS                              
: OFFSET_OF_FONT ( -- )   START_BLOCK_NUMBER SIZE_OF_FONT_IN_BLOCKS -  $400 * ;  

576 constant SECTORS_TO_LOAD                                    \ number of 512 octet sectors 576 => 288 Kbytes

16 constant BITS_PER_PIXEL                                      \ MUST BE 16 !!! display pixel sizes, colour depth = 16 bit ( 2 bytes )

\ for the maximum supported screen : 1024 x 768 pixels :
1024 constant MAX_SCREEN_WIDTH                                  \ maximum screen width in pixels
768 constant MAX_SCREEN_HEIGHT                                  \ maximum screen height in pixels

: BYTES_PER_PIXEL ( -- )   BITS_PER_PIXEL 8 / ;                 

1   constant PIXEL_SHIFT                                        \ how many bits to shift to scale by BYTES_PER_PIXEL

\ Memory Map
\ start   length
\ $100000 ....   RAM
\ $C0000 0xFFFFF BIOS video ROM - its not RAM!
\ $B8000 0x08000 BIOS video RAM
\ $10000 0xA8000 COLORFTH.CFX file is copied here
\ $0F000 0x01000 BIOS shadow RAM - its OK to use this if we do not call the video BIOS
\ $0A000 0x05000 BIOS video RAM - do not use until we have changed video mode
\ $07c00 0x00200 BPB Boot sector after loading by BIOS
\ $07c0b <----- di points here, the BPB ( + offset )  and variables ( - offset ) are accessed via [di]
\ $07b8c 0x00080 variables referenced via [di], followed by BPB variables referenced via [di]
\ $07800         Stacks, size = 0x0200 each , growing downwards
\ $02000 0x06000 SECTOR_BUFFER
\ $00000 0x02000 BIOS RAM

: SECTOR_BUFFER ( -- )   $00002000 ;                            \ buffer for disk reads and writes
: SECTOR_BUFFER_SIZE ( -- )   $4000 ;                           
: VESA_BUFFER ( -- )   SECTOR_BUFFER $0400 - ;                  \ for the VESA mode information
: DAP_BUFFER ( -- )   VESA_BUFFER $0020 - ;                     \ for the Int $13 Disk Address Packet (DAP)
: DISK_INFO ( -- )   DAP_BUFFER $0020 - ;                       \ for the Int $13 AH=08h get info
: V_REGS ( -- )   DISK_INFO $0020 - ;                           \ test only - registers before and after thunk call

: TRASH_BUFFER ( -- )   $00094C00 ;                             \ saves words deleted while editing

: vesa_BytesPerScanLine ( -- )   VESA_BUFFER $0E + ;            \ screen width ( number of horizontal pixels )
: vesa_XResolution ( -- )   VESA_BUFFER $12 + ;                 \ screen width ( number of horizontal pixels )
: vesa_YResolution ( -- )   VESA_BUFFER $14 + ;                 \ screen height ( number of vertical pixels )
: vesa_BitsPerPixel ( -- )   VESA_BUFFER $19 + ;                \ bits per pixel
: vesa_PhysBasePtr ( -- )   VESA_BUFFER $28 + ;                 \ address of linear frame buffer

: BOOTOFFSET ( -- )   $7C00 ;                                   

16 constant RELOC_BIT                                           \ the relocation address must be a power of 2
: RELOCATED    1 RELOC_BIT lshift ;                             \ $10000

\ stack allocation, three pairs of data and return stacks
\ Note : the return stack must be in the lowest 64K byte segment, for the BIOS calls to work.
: RETURN_STACK_0 ( -- )   $7800 ;                               \ top of stack memory area \  $a0000 in CM's code, 0x10000 in JC's code
: DATA_STACK_SIZE ( -- )   $0200 ;                              
: RETURN_STACK_SIZE ( -- )   $0200 ;                            
\ combined stack sizes
: STACK_SIZE ( -- )   DATA_STACK_SIZE RETURN_STACK_SIZE + ;     
: TWOxSTACK_SIZE ( -- )   STACK_SIZE 2 * ;                      
: TOTAL_STACK_SIZE ( -- )   STACK_SIZE 3 * ;                    \ three pairs of stacks, one for each task
: STACK_MEMORY_START ( -- )   RETURN_STACK_0 TOTAL_STACK_SIZE - ;  
\ data stacks
: DATA_STACK_0 ( -- )   RETURN_STACK_0 RETURN_STACK_SIZE - ;    \ $9f400 in CM's code
: DATA_STACK_1 ( -- )   DATA_STACK_0 STACK_SIZE - ;             
: DATA_STACK_2 ( -- )   DATA_STACK_0 TWOxSTACK_SIZE - ;         
\ return stacks
: RETURN_STACK_1 ( -- )   RETURN_STACK_0 STACK_SIZE - ;         
: RETURN_STACK_2 ( -- )   RETURN_STACK_0 TWOxSTACK_SIZE - ;     

: _TOS_ ( n -- )   eax ;  
: _SCRATCH_ ( n -- )   ebx ;                                    
: _MOV_TOS_LIT_ ( -- )   $B8 ;                                  \ the opcode for mov eax, 32_bit_literal  (in next 32 bit cell)

: _DUP_ ( -- )                                                  \ Top Of Stack is in the  _TOS_  register
    byte $04 esi sub                                            \ lea esi, [ esi - $04 ]     \  pre-decrement the stack pointer
    _TOS_ 0 [ESI] mov                                           \ copy the Top Of Stack ( TOS ) register to Second On Stack ( on the real stack )
 ;       

: _SWAP_ ( -- )    
    0 [ESI] _TOS_ xchg                                          
 ;       

: _DROP_ ( -- )    
    lodsd 
 ;       

: START_OF_RAM ( -- )   $00468000 ;                             
: ForthNames    START_OF_RAM ;                                  \ copied to RAM here from ROM ( i.e. boot program ) version
: ForthJumpTable ( -- )   ForthNames $2800 + ;                  \ copied to RAM here from ROM ( i.e. boot program ) version
: MacroNames ( -- )   ForthJumpTable $2800 + ;                  \ copied to RAM here from ROM ( i.e. boot program ) version
: MacroJumpTable ( -- )   MacroNames $2800 + ;                  \ copied to RAM here from ROM ( i.e. boot program ) version

: H0 ( -- )   MacroJumpTable $2800 + ;                          \ initial value of the dictionary pointer

512 constant SECTOR                                             \ bytes per floppy sector
2  constant HEADS                                               \ heads on 1.44M floppy drive
18 constant SECTORS                                             \ floppy sectors per track
: CYLINDER ( -- )   SECTOR SECTORS * HEADS * ;                  
4  constant CELL                                                \ bytes per cell
: DEBUGGER ( n -- )   $e1 ;                                     \ port to hardware debugger?

\ return the cell address of the given variable - obsolete
\ %macro LOAD_RELATIVE_ADDRESSdiv4  1
\    mov _TOS_, ( ( ( %1 - $$ ) + RELOCATED ) >> 2 )
\ %endmacro

: LOAD_RELATIVE_ADDRESS ( n -- )                                
    RELOCATED [] _TOS_ mov                                      
 ;       

\ emit the given following character
: EMIT_IMM ( n -- )    
\    push esi
    _DUP_ 
    %1 _TOS_ mov 
    emit_ call 
\    pop esi
 ;       

\ *****************************************************************************
\ Registers used
\ *****************************************************************************
\ _TOS_ is the top stack item ( eax --> ebx )
\ esp the call ... ret  return stack pointer
\ edi  dictionary pointer ( H --> : HERE ( -- a )   H @ \  )
\ esi is the stack pointer, also needed by lods and movs
\ e.g. lodsd  loads a 32 bit dword from [ds:esi] into _TOS_, increments  esi  by 4
\ ebx  scratch register
\ ecx  counter and scratch register
\ edx  run-time pointer (?), "a register" used by  a!  , otherwise scratch register
\ ebp  variable pointer register, set to VARIABLES_START_ADDRESS
\ "ds" = selector $10 ==> 0x0000:0000
\ "es" = selector $10 ==> 0x0000:0000
\ "ss" = selector $10 ==> 0x0000:0000

[UNDEFINED] NOT_NORMAL [IF] 
\ colours
$1082  constant colour_background                               
$1082FFE0  constant colour_yellow                               
$10820000  constant colour_black                                
$1082F800  constant colour_red                                  
$10820600  constant colour_green                                
$108207FF  constant colour_cyan                                 
$1082FFFF  constant colour_white                                
$1082841F  constant colour_light_blue                           
$1082C618  constant colour_silver                               
$1082F81F  constant colour_magenta                              
$1082001F  constant colour_blue                                 
$1082E200  constant colour_orange                               
$1082FFE0  constant colour_dark_yellow                          
$108207C0  constant colour_dark_green                           
$FFE0E200  constant colour_PacMan                               
$C618E200  constant colour_blockNumber                          
[ELSE]  
\ colours
$7CCF  constant colour_background                               
$7CCFFFE0  constant colour_yellow                               
$7CCF0000  constant colour_black                                
$7CCFF800  constant colour_red                                  
$7CCF0600  constant colour_green                                
$7CCF07FF  constant colour_cyan                                 
$7CCFFFFF  constant colour_white                                
$7CCF841F  constant colour_light_blue                           
$7CCFC618  constant colour_silver                               
$7CCFF81F  constant colour_magenta                              
$7CCF001F  constant colour_blue                                 
$7CCFE200  constant colour_orange                               
$7CCFFFE0  constant colour_dark_yellow                          
$7CCF07C0  constant colour_dark_green                           
$FFE0E200  constant colour_PacMan                               
$C618E200  constant colour_blockNumber                          
[THEN]  

#16 BitsMode ! 

    RELOCATED org 

LABEL start  
LABEL codeStart  
    start0 jmp     \ $03 bytes |  EB 58 90  00 Jump to boot code
    3 - ($ - $$) nop times                                      \ fill with 1 or 0 no-ops to address 3
    \ BIOS boot parameter table = $25 bytes
    'cf2018 0' c,                                               \ 03 Eight byte OEM name
    $0200      w,                                               \ 11 Number of Bytes Per Sector
    $08        c,                                               \ 13 Number of Sectors Per Cluster
    $05E0      w,                                               \ 14 Number of Reserved Sectors until the FAT
    $02        c,                                               \ 16 Number of Copies of FAT : always = 2
    $0000      w,                                               \ 17 Maximum number of Root Directory Entries
    $0000      w,                                               \ 19 Not used for FAT32
    $F8        c,                                               \ 21 Media type F0 = 1.44M 3.5 inch floppy disk, F8 = hard disk
    $0000      w,                                               \ 22 Sectors Per FAT for FAT12 and FAT16 - not used for FAT32
    $003F      w,                                               \ 24 Sectors per Track
    $00FF      w,                                               \ 26 Number of heads
    $00000038  ,                                                \ 28 Hidden sectors preceding the partition that contains this FAT volume
    $007477C8  ,                                                \ 32
    $00001D10  ,                                                \ 36 Sectors Per FAT for FAT32
    $0000      w,                                               \ 40
    $0000      w,                                               \ 42
    $00000002  ,                                                \ 44 Start of all directories, including root.
    $0001      w,                                               \ 48
    $0006      w,                                               \ 50 Offset in sectors from this sector to the backup BPB sector
\    times 12 db 0      \  $0C bytes |  00 00 00 00 00 00 00 00 00 00 00 00  52
\    db $00            \  64
\    db $00            \  65
\    db $29            \  66 Extended Boot Signature
\    dd $44444444      \  67 serial number
\    db 'colorForth '   \  71 Eleven byte Volume Label
\    db 'cFblocks'      \  82 Eight byte File System name

\ ******************************************************************************
\ ******************************************************************************

    nop 8 align \ has to be aligned to 8 for GDT
    \ Note : we are NOT using null descriptor as GDT descriptor, see: http://wiki.osdev.org/GDT_Tutorial
    \ "The null descriptor which is never referenced by the processor. Certain emulators, like Bochs, will complain about limit exceptions if you do not have one present.
    \ Some use this descriptor to store a pointer to the GDT itself (to use with the LGDT instruction).
    \ The null descriptor is 8 bytes wide and the pointer is 6 bytes wide so it might just be the perfect place for this."

LABEL gdt                                                       \ the GDT descriptor
    gdt_end - gdt - 1 w,                                        \ GDT limit
    gdt0 + BOOTOFFSET w,                                        \ pointer to start of table, low 16 bits
    0  w,  0 w,                                                 \ the high bits of the longword pointer to gdt

LABEL gdt0                                                      \ null descriptor
    0          w,  \ 0,1 limit 15:0
    0          w,  \ 2,3 base  15:0
    0          c,  \ 4   base  23:16
    0          c,  \ 5   type
    0          c,  \ 6   limit 19:16, flags
    0          c,  \ 7   base  31:24
HERE gdt0 -  constant code32p_SELECTOR_$08                      
\ bytes   1 0     3 2     5 4     7 6
    $FFFF w,  $0000 w,  $9A00 w,  $00CF w,                      \ 32-bit protected-mode code, limit 0xFFFFF
HERE gdt0 -  constant data32p_SELECTOR_$10                      
    $FFFF w,  $0000 w,  $9200 w,  $00CF w,                      \ 32-bit protected-mode data, limit 0xFFFFF
HERE gdt0 -  constant code16r_SELECTOR_$18                      
    $FFFF w,  $0000 w,  $9A00 w,  $0000 w,                      \ 16-bit real-mode code, limit 0xFFFFF
HERE gdt0 -  constant data16r_SELECTOR_$20                      
    $FFFF w,  $0000 w,  $9200 w,  $0000 w,                      \ 16-bit real-mode data, limit 0xFFFFF
LABEL gdt_end  

\ ******************************************************************************
\ ******************************************************************************

\ align to 4 so we can access variables from high-level Forth
    nop 4 align 

LABEL data_area  \ data area begins here

LABEL bootsector                                                \ LBA of boot sector
    0          ,   

\ save disk information, cylinder, sector, head and drive from BIOS call
LABEL driveinfo_Drive_DX                                        \ use low byte to store boot Drive into from BIOS DL
    0          w,  

LABEL driveinfo_CX                                              \ [7:6] [15:8][7] logical last index of cylinders = number_of - 1 (because index starts with 0)
                      \ [5:0][7] logical last index of sectors per track = number_of (because index starts with 1)
    0          w,  

\ cylinders, sectors, heads of boot drive
\ low word: high byte is head
\ high word: cylinder and sector: C76543210 C98S543210
LABEL driveinfo_Cylinder  
    0          c,  
LABEL driveinfo_Head  
    0          c,  
LABEL driveinfo_SectorsPertrack                                 
    0          w,  

    nop 4 align 

LABEL vesamode                                                  \ holds selected vesa mode number for 1024x768x32 display
    $0000      ,                                                \ Qemu and Bochs use $4144 , we scan all modes from 0x4100 to 0x41FE

LABEL destination  
    RELOCATED  ,   

LABEL dispPtr  
    $00000140  ,   

LABEL v_bytesPerLine  
    $00        ,   

LABEL v_scanCode  
    $00        ,   

    4 align 

\ ******************************************************************************
\ ******************************************************************************

LABEL start0  

    cli                                                         \ clear interrupts
                                \ turns out we don't need interrupts at all, even when using BIOS routines
                                \ but we need to turn them off after disk calls because BIOS leaves them on

    si push                                                     \ need to transfer SI to unused register BX later

\ note: cannot touch DX or BP registers until we've checked for partition boot
\ (SI could be used as well as BP but we use SI for relocation)

 \  ee mbrboot.nasm
                                \ Note : relocate the bootblock before we do anything else
    bx pop                                                      \ we cannot use the current stack after changing SS or SP
                                \ ... because mbrboot.nasm places stack at $7c00, in SECTOR_BUFFER
                                \ and we cannot use BP because its default segment is SS
    ax ax xor 
    ax ds mov 
    ax es mov 

    BOOTOFFSET si mov                                           
    SECTOR_BUFFER di mov                                        
    di sp mov 
    $100 cx mov  
    movsw rep                                                   \ note that this instruction doesn't change AX , it moves DS:SI to ES:DI and increments SI and DI

    ax ss mov                                                   \ stack segment also zero
    $b8 ah mov                                                  \ video RAM
    ax gs mov                                                   \ store in unused segment register

    [gdt - $$ + BOOTOFFSET] lgdt                                

    SetupUnrealMode call                                        \ gs and ss must be initialized before going to Unreal Mode

\ *****************************************************************************
\ Enable the A20 address line, otherwise all odd 1 MByte pages are disabled
\ Using the "PS/2 Controller" or 8042 "Keyboard controller"
\ *****************************************************************************
    \ from  http://wiki.osdev.org/%228042%22_PS/2_Controller#Step_1:_Initialise_USB_Controllers
    \ Write a command to the on-board 8042 "Keyboard controller" port $64 :
    \ $20     Read "byte 0" from internal RAM     Controller Configuration Byte
    \ $21 to 0x3F    Read "byte N" from internal RAM (where 'N' is the command byte & 0x1F)
    \ $60    Write next byte to "byte 0" of internal RAM (Controller Configuration Byte)
    \ $61 to 0x7F    Write next byte to "byte N" of internal RAM (where 'N' is the command byte & 0x1F)
    \ $A7    Disable second PS/2 port
    \ $A8    Enable second PS/2 port
    \ $A9    Test second PS/2 port
    \     $00 test passed
    \     $01 clock line stuck low
    \     $02 clock line stuck high
    \     $03 data line stuck low
    \     $04 data line stuck high
    \ $AA    Test PS/2 Controller
    \     $55 test passed
    \     $FC test failed
    \ $AB    Test first PS/2 port
    \     $00 test passed
    \     $01 clock line stuck low
    \     $02 clock line stuck high
    \     $03 data line stuck low
    \     $04 data line stuck high
    \ $AC    Diagnostic dump (real all bytes of internal RAM)    Unknown
    \ $AD    Disable first PS/2 port     None
    \ $AE    Enable first PS/2 port  None
    \ $C0    Read controller input port  Unknown (none of these bits have a standard/defined purpose)
    \ $C1    Copy bits 0 to 3 of input port to status bits 4 to 7    None
    \ $C2    Copy bits 4 to 7 of input port to status bits 4 to 7    None
    \ $D0    Read Controller Output Port     Controller Output Port (see below)
    \ $D1    Write next byte to Keyboard Controller Output Port Note: Check if output buffer is empty first
    \ $D2    Write next byte to first PS/2 port output buffer
    \ $D3    Write next byte to second PS/2 port output buffer
    \ $D4    Write next byte to second PS/2 port input buffer
    \ $F0 to 0xFF  Pulse output line low for 6 ms.
    \     Bits 0 to 3 are used as a mask (0 = pulse line, 1 = do not pulse line) and correspond to 4 different output lines.
    \     Bit 0 is the "reset" line, active low.
    $D1 al mov                                                  \ $D1 = Write next byte to Keyboard Controller Output Port
    al $64 out                                                  \ On-board controller Command Write
LABEL .back  
    $64 al in  
    $02 al and  
    .back jnz 
    $4B al mov  
    al $60 out  

\ *****************************************************************************
\ Get disk drive parameters from the BIOS
\ *****************************************************************************

    (data_area - $$ + BOOTOFFSET) di mov                        \ setup the data index pointer
    eax eax xor 
    16 eax bts                                                  \ in case NOT booted from partition: sector 1, head 0, cylinder 0
    dh dh or                                                    \ booted from partition?
    .forward3 jz 
    [ bx + 8 ] eax mov                                          \ SI (now BX) contains pointer to partition record
    eax [ byte di + (bootsector - data_area) ] mov              \ offset 8 was LBA of first absolute sector
    [bx] eax mov                                                \ CHS of first sector in partition
LABEL .forward3  
    dl al mov                                                   \ bootdrive into AL
    eax [ byte di + ( driveinfo_Drive_DX - data_area) ] mov     \ save the Drive info from BIOS
    8 ah mov                                                    \ get drive parameters
    es push                                                     \ this operation messes with ES
    di push                                                     \ and DI
    DISK_INFO di mov                                            \ point di at the table returned by this software interrupt
    $13 int 
    $ jc                                                        \ stop here on error

    ReSetupUnrealMode call                                      
    di pop 
    es pop 

\ ******************************************************************************
\ load the bootdisk into both low and high RAM
\ ******************************************************************************

    dx [ byte di + ( driveinfo_Cylinder - data_area) ] mov      \ heads in high byte
    $3F cl and                                                  \ we don't care about two high bits of cylinder count
    cx [ byte di + ( driveinfo_SectorsPertrack - data_area) ] mov \ cylinders and sectors/track
    [ di + ( driveinfo_Drive_DX - data_area) ] dx mov           \ restore dl Drive value from BIOS, dh = 0
    [ di + ( driveinfo_CX - data_area) ] cx mov                 \ restore cl value, ch = 0
    SECTORS_TO_LOAD si mov                                      

    SECTOR_BUFFER bx mov                                        \ relocate the sector we are running from
    relocate call 

    BOOTOFFSET bx mov                                           \ we will fix this below by adding $200
                                    \ remember the sector is 1-based, head and cylinder both 0-based

LABEL .nextsector  
    cl inc 
    si dec 
    setVideoMode jz                                             \ success, so setup the video now...

LABEL .bootload  
    $201 ax mov                                                 \ read 1 sector
    $02 bh add                                                  \ into next available slot in RAM
    .forward jnz 
    $02 bh sub                                                  \ at $10000 we go back to 0xfe00
LABEL .forward  
    $13 int 
    ReSetupUnrealMode call                                      
    $ jc                                                        \ stop here on error
    relocate call 
    cl al mov 
    $3F al and                                                  \ low 6 bits
    [ byte di + ( driveinfo_SectorsPertrack - data_area) ] al cmp 
    .nextsector jnz 
    dh inc                                                      \ next head
    [ byte di + ( driveinfo_Head - data_area) ] dh cmp          
    .forward2 jna                                               \ not JNZ, the head index is 1 less than head count
    dh dh xor 
    ch inc                                                      \ next cylinder
    .forward2 jnz 
    $40 cl add                                                  \ bit 8 of cylinder count
LABEL .forward2  
    $C0 cl and                                                  \ clear sector count, low 6 bits of cl
    short .nextsector jmp                                       

\ ******************************************************************************
\ ******************************************************************************
\ Start here after loading the program
\ ******************************************************************************
\ ******************************************************************************

\ From : VESA BIOS EXTENSION (VBE) Core Functions Standard Version: 3.0 Date: September 16, 1998
\ Mandatory information for all VBE revisions
\ dw ModeAttributes         \  $00 mode attributes
\ db WinAAttributes         \  $02 window A attributes
\ db WinBAttributes         \  $03 window B attributes
\ dw WinGranularity         \  $04 window granularity
\ dw WinSize                \  $06 window size
\ dw WinASegment            \  $08 window A start segment
\ dw WinBSegment            \  $0A window B start segment
\ dd WinFuncPtr             \  $0C real mode pointer to window function
\ dw BytesPerScanLine       \  $10 bytes per scan line               <--------------
\ Mandatory information for VBE 1.2 and above
\ dw XResolution            \  $12 horizontal resolution in pixels   <-------------- scrnw
\ dw YResolution            \  $14 vertical resolution in pixels     <-------------- scrnh
\ db XCharSize              \  $16 character cell width in pixels
\ db YCharSize              \  $17 character cell height in pixels
\ db NumberOfPlanes         \  $18 number of memory planes
\ db BitsPerPixel           \  $19 bits per pixel                    <-------------- bpp
\ db NumberOfBanks          \  $1A number of banks
\ db MemoryModel            \  $1B memory model type
\ db BankSize               \  $1C bank size in KB
\ db NumberOfImagePages     \  $1D number of images
\ db Reserved               \  $1E reserved for page function
\ Direct Color fields (required for direct/6 and YUV/7 memory models)
\ db RedMaskSize            \  $1F size of direct color red mask in bits
\ db RedFieldPosition       \  $20 bit position of lsb of red mask
\ db GreenMaskSize          \  $21 size of direct color green mask in bits
\ db GreenFieldPosition     \  $22 bit position of lsb of green mask
\ db BlueMaskSize           \  $23 size of direct color blue mask in bits
\ db BlueFieldPosition      \  $24 bit position of lsb of blue mask
\ db RsvdMaskSize           \  $25 size of direct color reserved mask in bits
\ db RsvdFieldPosition      \  $26 bit position of lsb of reserved mask
\ db DirectColorModeInfo    \  $27 direct color mode attributes
\ Mandatory information for VBE 2.0 and above
\ dd PhysBasePtr            \  $28 physical address for flat memory frame buffer  <-------------- frame
\ dd Reserved               \  $2C Reserved - always set to 0
\ dw Reserved               \  $30 Reserved - always set to 0
\ Mandatory information for VBE 3.0 and above
\ dw LinBytesPerScanLine    \  $32 bytes per scan line for linear modes
\ db BnkNumberOfImagePages  \  $34 number of images for banked modes
\ db LinNumberOfImagePages  \  $35 number of images for linear modes
\ db LinRedMaskSize         \  $36 size of direct color red mask (linear modes)
\ db LinRedFieldPosition    \  $37 bit position of lsb of red mask (linear modes)
\ db LinGreenMaskSize       \  $38 size of direct color green mask  (linear modes)
\ db LinGreenFieldPosition  \  $39 bit position of lsb of green mask (linear modes)
\ db LinBlueMaskSize        \  $3A size of direct color blue mask  (linear modes)
\ db LinBlueFieldPosition   \  $3B bit position of lsb of blue mask (linear modes)
\ db LinRsvdMaskSize        \  $3C size of direct color reserved mask (linear modes)
\ db LinRsvdFieldPosition   \  $3D bit position of lsb of reserved mask (linear modes)
\ dd MaxPixelClock          \  $3E maximum pixel clock (in Hz) for graphics mode
\ times 189 db 0            \  $42 remainder of ModeInfoBlock
\ End                       \  $FF

LABEL scanVESA  \ ( w+h+b -- )  in ax
    ax bx mov 
    di push                                                     \ save di
    ( $4100 - 1 ) cx mov                                        \ start scanning from VESA mode $4100 ( the -1 is because of the  inc cx  below )
LABEL .back  
    cx inc 
    $FF cl cmp                                                  \ scanned from 0 to FE, not found, so show error
    .failure jz 
    VESA_BUFFER di mov                                          \ buffer for the VESA mode information block
    $4F01 ax mov                                                \ INT $10, AX=0x4F01, CX=mode Get Mode Info
    $10 int 
    $4F al cmp                                                  \ success code = $4F
    .back jne                                                   \ try the next VESA mode
    [di + $12] ax mov                                           \ width
    [di + $14] ax add                                           \ height
    [di + $19] al add                                           \ bits per pixel
\    adc ah, 0                       \  should not be necessary for the expected result, $400+0x300+0x10
    bx ax cmp                                                   \ width + height + bits per pixel
    .success je 
    .back jne                                                   \ try the next VESA mode
LABEL .failure                                                  \ VESA mode not found, so continue
    di pop                                                      \ restore di
    0 ax mov                                                    \ return flag false
    0 ax add                                                    \ set the zero flag
    ret 
LABEL .success  
    di pop                                                      \ restore di
    1 ax mov                                                    \ return flag true
    0 ax add                                                    \ set the zero flag
    ret 

LABEL setVESA  \ we found a valid VESA mode

    cx [ di + ( vesamode - data_area) ] mov                     

    ds push                                                     \ clear all flags including Interrupt using DS, known to be zero
    popf                                                        \ this is necessary to clear T flag also, end register display

    greet call                                                  \ show greeting message

    cx bx mov 
    $4F02 ax mov                                                \ INT $10, AX=0x4F02, BX=mode, ES:DI=CRTCInfoBlock Set Video Mode
    $10 int 

    graphicsmode jmp                                            

LABEL setVideoMode  
FORCE_80$600_VESA 0 = [IF]                                      \ test the 800x600 mode in bochs, which supports 1024x768
    ( 1024 + 768 + BITS_PER_PIXEL ) ax mov                      \ try the highest resolution first
    scanVESA call                                               \ if VESA mode is found, jump to setVESA
    constant jnz                                                \ success - we found the requested VESA mode
[THEN]  
    ( 800 + 600 + BITS_PER_PIXEL ) ax mov                       \ then try a lower resolution
    scanVESA call                                               \ if VESA mode is found, jump to setVESA
    constant jnz                                                \ success - we found the requested VESA mode

\    mov ax, 640 + 480 + BITS_PER_PIXEL  \  then try an even lower resolution
\    call scanVESA                   \  if VESA mode is found, jump to setVESA
\    jnz setVESA                     \  success - we found the requested VESA mode
    showVESAerror jmp                                           \ we have tried all VESA modes without success, so report an error

\ ******************************************************************************
\ ******************************************************************************

LABEL relocate                                                  \ copy 512 bytes from  [bx]  to FS:[destination]
    pusha 
    $200 / 2 cx mov                                             
    bx si mov 
    [ byte di + ( destination - data_area) ] ebx mov            
LABEL .back  
    lodsw                                                       \ load the 16 bit value pointed to by SI into  ax
    ax [fs:ebx] mov                                             \ Note : the  fs:  uses the 32 bit FS value setup in Unreal Mode to move the data outside of the 1 Mbyte Real Mode address range
    byte +2 ebx add                                             
    .back loop 

    ebx [ byte di + ( destination - data_area) ] mov            
    popa 
    ret 

    \ not used because it is very slow :
\ now set up for trap displaying registers on screen during bootup
\    push cs
\    push showstate - $$ + BOOTOFFSET
\    pop dword [word +4]

\ ******************************************************************************
\ ******************************************************************************
\  . MasterBoot Record - MBR   at Sector     0 (decimal 0)        MBR
\ Partition at offset 1BE
\   BootSignature                0
\   Start Head|Sector|Cylinder   1   1     0
\   Partition Type               B  DOS 7.1+
\   End   Head|Sector|Cylinder  FE  3F   3E5
\   BPBsectorNumber                   00  \ was 3F
\   Size of partition (decimal)  16035777 sectors,  8210317824 bytes,  8017889 Ki bytes,  7830 Mi bytes,  8 Gi bytes
\ Partition at offset 1CE
\   BootSignature                0
\   Start Head|Sector|Cylinder   0   0     0
\   Partition Type               0  Empty partition
\   End   Head|Sector|Cylinder   0   0     0
\   BPBsectorNumber                    0
\   Size of partition (decimal)         0 sectors,           0 bytes,  0 Ki bytes,  0 Mi bytes,

\ pretend to be a Master Boot Record so that the BIOS will load us
    ( $000001BE - ( $ - $$ ) ) db $77 times                     
    $80 c,  $01 c,  $01 c,  $00 c,  $0B c,  $FE c,  $FF c,  $E5 c,  $00 c,  $00 c,  $00 c,  $00 c,  $C1 c,  $AF c,  $F4 c,  $00 c,  \ 0x1BE DOS partition 0  working on PC
    00 c,  00 c,  00 c,  00 c,  00 c,  00 c,  00 c,  00 c,      \ $1CE first 8 bytes of empty partition 1

LABEL SetupUnrealMode  
    \ set the FS segment in "unreal" mode, must be done before the Trap Flag is set in EFLAGS register
    cr0 eax mov 
    1 al or     \ set the "protected mode enable" bit => "unreal mode"
    eax cr0 mov 
    word data32p_SELECTOR_$10 push                              \ set the FS segment
    fs pop 
    al dec      \ clear the "protected mode enable" bit
    eax cr0 mov 
    ds push                                                     \ now set FS to 0
    fs pop 

LABEL ReSetupUnrealMode  
    cs push                                                     \ for iret
    pushf                                                       \ for iret
    pusha 
    sp bp mov 
    [bp + 16] ax mov                                            \ get flags
\    or ah, $01                    \  set Trap Flag, bit 8 in the EFLAGS register ; debug only - very slow!
    ~$02 ah and                                                 \ reset interrupt flag
    [ bp + 20 ] ax xchg                                         \ swap flags with return address
    ax [ bp + 16 ] mov                                          \ return address at top of stack after popa
    popa 
    iret 

\ ******************************************************************************
\ ******************************************************************************

    512 - 2 - ($ - $$) nop times                                \ fill with no-ops to 55AA at end of boot sector
    $55  c,  $AA c,                                             \ boot sector terminating bytes

\ ******************************************************************************
\ End of Boot Sector
\ ******************************************************************************

\ ******************************************************************************
\ Show the user a null terminated string - writes directly into video RAM
\ ******************************************************************************

LABEL displayString  

    \ restore the pointer to screen memory into di
    (data_area - $$ + BOOTOFFSET) di mov                        
    [ di + ( dispPtr - data_area) ] ax mov                      
    ax di mov 

    es push                                                     \ save es
    $b800 ax mov                                                \ video RAM segment
    ax es mov 

LABEL backhere2  
    lodsb                                                       \ loads a byte from [ds:si] into al, then increments  si
    0 al cmp 
    forward1 jz                                                 \ If al = 0 then leave the loop
    $0D ah mov                                                  \ text colour, magenta on black background
    stosw                                                       \ stores  ax  into  [es:di]  then increments  di
    backhere2 jmp 
LABEL forward1  
    \ save the pointer to screen memory from di
    di ax mov 
    (data_area - $$ + BOOTOFFSET) di mov                        
    ax [ di + ( dispPtr - data_area) ] mov                      
    es pop                                                      \ restore es
    ret 

\ display a string then Wait for a key press
LABEL displayStringW  

    pusha 
    displayString call                                          

    ax ax xor                                                   \ wait for and get a key press ( AX = 0 )
    $16 int                                                     \ BIOS interrupt Read a Key From the Keyboard
    popa 
    ret 

\ msg_greeting2:
\     db ' Press any key : ' , $00

LABEL msg_VESAerror  
    'No valid VESA mode found!! '  c,  $02 c,  $00 c,           
\    db ' No VESA mode ' , $02, 0x00

#16 BitsMode ! 

LABEL showVESAerror  
    greet call 
    si push 
    $000001E0 word [ di + ( dispPtr - data_area) ] mov          \ line 3 $50 x 2 x 3 = 0x1E0
    ( msg_VESAerror - $$ + BOOTOFFSET ) si mov                  \ string to display
    displayStringW call                                         
    si pop 
    ret 

LABEL greet  \ jump here to show 16 bit version text
    si push 
    $00000140 word [ di + ( dispPtr - data_area) ] mov          \ line 2 $50 x 2 x 2 = 0x140
    ( version - $$ + BOOTOFFSET ) si mov                        \ string to display
    displayString call                                          
\    mov si, ( msg_greeting2 - $$ + BOOTOFFSET )  \  string to display
\    call displayStringW
    si pop 
    ret 

LABEL graphicsmode  

\ *****************************************************************************
\ *****************************************************************************

    setProtectedModeAPI call                                    \ called from 16 bit code, returns to 32 bit code

#32 BitsMode ! 

    RETURN_STACK_0 esp mov                                      \ setup the return stack pointer
    ( DATA_STACK_0 + 4 ) esi mov                                \ setup our data stack pointer

    dword warm jmp 

\ *****************************************************************************
\ calculate Cylinder, Head and Sector from zero-based sector number
\ see http://teaching.idallen.com/dat2343/00f/calculating_cylinder.htm
\ Note : uses pushad to copy registers onto the ESP stack, stores the
\ calculated values onto the stack at the correct offsets, then restores the
\ stack back to the registers.
\ *****************************************************************************

LABEL sector_chs  \ calculate CHS from a sector number in eax,
    \ returns with DX = HHDD, CX = CCSS where HH=head, DD=drive, CC=cylinder, SS=sector
    \ Note that the input sector number is zero based, and that the high 16 bits of EAX must be 0
    pushad  \ Pushes all general purpose registers onto the stack in the following order:
        \ EAX, ECX, EDX, EBX, ESP, EBP, ESI, EDI. The value of ESP is the value before the actual push of ESP
        \  7    6    5    4    3    2    1    0   offset in cells from ESP
    esp ebp mov                                                 \ copy the original ESP stack pointer to EBP so we can access items on the stack easily

    [ bootsector - $$ + BOOTOFFSET] eax add                     
    eax push                                                    \ save it while we calculate heads*sectors-per-track
    [ driveinfo_Head - $$ + BOOTOFFSET] al mov                  \ index of highest-numbered head
    al inc                                                      \ 1-base the number to make count of heads
    byte [ driveinfo_SectorsPertrack - $$ + BOOTOFFSET] mul     \ sectors per track
    eax ebx mov 
    eax pop 
    edx edx xor                                                 \ clear high 32 bits
    ebx div                                                     \ leaves cylinder number in eax, remainder in edx
    eax ecx mov                                                 \ store cylinder number in another register
    edx eax mov                                                 \ get remainder into AX
    [ driveinfo_SectorsPertrack - $$ + BOOTOFFSET] bl mov       \ number of sectors per track
    bl div                                                      \ head number into AX, remainder into DX
    al bl mov                                                   \ result must be one byte, so store it in BL
    8 ecx rol                                                   \ high 2 bits of cylinder number into high 2 bits of CL
    6 cl shl                                                    \ makes room for sector number
    ah cl or                                                    \ merge cylinder number with sector number
    cl inc                                                      \ one-base sector number
    ecx [ ebp + ( 6 * 4 ) ] mov                                 \ store the result in ECX position on esp stack
    bl ch mov                                                   \ place head number in high bits
    [ driveinfo_Drive_DX - $$ + BOOTOFFSET] cl mov              \ drive number in low 8 bits
    ecx [ ebp + ( 5 * 4 ) ] mov                                 \ store the result in EDX position on esp stack
    popad                                                       \ restore registers from esp stack
    ret 

\ *****************************************************************************
\ enter Protected Mode (32 bit) and Real Mode (16 bit)
\ from http://ringzero.free.fr/os/protected%20mode/Pm/PM1.ASM
\ *****************************************************************************

#16 BitsMode ! \ Real Mode code (16 bit)

LABEL enterProtectedMode                                        \ must come from a 'call' , can not be inlined
    ax pop 
    code32p_SELECTOR_$08 push                                   
    ax push 
    retf 

LABEL setProtectedModeAPI                                       \ set protected mode from 'Real' mode. Called from 16 bit code, returns to 32 bit code
    pushad                                                      \ save all registers as doublewords
    cr0 eax mov 
    1 al or 
    eax cr0 mov                                                 \ set the Protected Mode bit in the Control Register
    eax eax xor                                                 \ clear high bits of eax
    enterProtectedMode call                                     

#32 BitsMode ! \ Protected Mode code (32 bit)

    data32p_SELECTOR_$10 eax mov                                \ Protected Mode data segment
    ax es mov 
    ax ds mov 
    ax ss mov                                                   \ this makes stack segment 32 bits
    popad 
16BIT: ret 

LABEL enter16bitProtectedMode                                   \ 32 bit code. Must come from a 'call' , can not be inlined
    eax pop                                                     \ return address
    dword code16r_SELECTOR_$18 push                             \ select 16-bit Protected Mode AKA 'Real' Mode
    eax push 
    retf 

LABEL setRealModeAPI                                            \ set 'Real' mode from protected mode. Called from 32 bit code, returns to 16 bit code
    \ assumed that protected-mode stack is based at 0
    \ and that bits 16 through 19 will not change during time in realmode
    pushad                                                      \ save 32-bit values of registers
    esp ecx mov                                                 \ do all possible 32-bit ops before going to 16 bits
    cr0 edx mov 
    enter16bitProtectedMode call                                

#16 BitsMode ! \ Real Mode code (16 bit)

    data16r_SELECTOR_$20 ax mov                                 
    ax ds mov 
    ax es mov 
    ax ss mov                                                   \ here the stack becomes 16 bits based at 0, and SP used not ESP
                        \ *** consider stack invalid from here until we reach real mode ***
    cx cx xor                                                   \ clear low 16 bits
    4 ecx shr                                                   \ move high 4 bits into cl
    dl dec                                                      \ leave protected mode, only works if we KNOW bit 0 is set
    edx cr0 mov 
    enterRealMode call                                          
    ax ax xor 
    ax ds mov 
    ax es mov 
    cx ss mov 
    \ note we don't need to set SP to 8xxx if ESP is b8xxx, since
    \ the b000 is now in SS, and the b of b8xxx is ignored in real mode
    popad 
32BIT: ret 

LABEL enterRealMode                                             \ 16 bit code. Must come from a 'call' , can not be inlined
    ax pop 
    fs push                                                     \ real-mode code segment
    ax push 
    retf 

#32 BitsMode ! \ Protected Mode code (32 bit)

\ *****************************************************************************
\ *****************************************************************************

\  include "JCreadwrite.nasm"
\ JCreadwrite.nasm 2012 Oct 23   read and write the disk using 16 bit BIOS calls
\ BIOS read and write routines for colorForth

#32 BitsMode ! \ Protected Mode code (32 bit)

LABEL bios_read  \ ( a c -- a' c' )   \  read cylinder c into address a , leave next address and cylinder
                                \ c is cylinder, we will use 1.44Mb floppy's idea of cylinder regardless
                                \ a is byte address
                                \ leave updated c and a on stack as c' and a'
                                \ a cylinder is 36 tracks of 512 bytes each, $4800 bytes, 0x1200 cells (words)

    cli                                                         \ disable interrupts
    pushad                                                      \ push all registers ( except esp ) and flags onto the stack
    esp ebp mov                                                 \ copy of stack pointer for use below ( * ), points to registers copied by pushad , above

    HEADS * SECTORS ecx mov                                     \ sectors per track (both heads)
    cl mul                                                      \ sector number goes into AX
                                \ note that resultant sector number is zero-based going into sector_chs!
                                \ set up loop to read one floppy cylinder's worth

    eax push                                                    \ absolute sector number to start
LABEL .back  
    ecx push 
    sector_chs call                                             \ convert to Cylinder-Head-Sector in CX-DX
    .readsector call                                            

    [ ebp + ( 1 * 4 ) ] ebx mov                                 \ ( * ) get ESI stored on stack, via stack pointer saved in ebp
    [ebx] edi mov                                               \ destination index address for movsd
    ( 512 >> 2 ) ecx mov                                        \ number of 32-bit words to move, 512 bytes
    SECTOR_BUFFER esi mov                                       \ source index for movsd
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    edi [ebx] mov 
    ecx pop 
    eax pop 
    eax inc 
    eax push 
    .back loop 
    eax pop 
    dword [ebp + 7 * 4] inc                                     \ for updated cylinder number after return
    popad 
    ret 

LABEL .readsector                                               \ no need to save registers because we take care of them in calling routine
    setRealModeAPI call                                         
#16 BitsMode ! 
    SECTOR_BUFFER bx mov                                        
    $0201 ax mov                                                \ read 1 sector
    $13 int 
    cli                                                         \ BIOS might have left interrupts enabled
    setProtectedModeAPI call                                    \ called from 16 bit code, returns to 32 bit code
#32 BitsMode ! 
    ret 

LABEL bios_write  \ ( a c -- a' c' )   \  write cylinder c from address a , leave next address and cylinder
    cli                                                         \ disable interrupts
    pushad 
    esp ebp mov 
                                    \ eax contains cylinder to start, the 'c' parameter
    HEADS * SECTORS ecx mov                                     \ sectors per track (both heads)
    cl mul                                                      \ absolute sector number goes into AX

    [ebp + ( 1 * 4 ) ] ebx mov                                  \ stored ESI on stack
    [ebx] esi mov                                               \ word address, 'a' parameter
\        shl esi, 2                 \  change word address into byte address
                                    \ set up loop to write one floppy cylinder's worth
    eax push                                                    \ absolute sector number to start

LABEL .back  
    ecx push 
                                    \ load sector data into buffer
                                    \ take advantage of knowing ECX only has byte value
    128 cl mov  \ ( 512 >> 2 )      \  number of 32-bit words to move
    SECTOR_BUFFER edi mov                                       
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    sector_chs call                                             \ convert to Cylinder-Head-Sector in CX-DX
    .writesector call                                           
    ecx pop 
    eax pop 
    eax inc 
    eax push 
    .back loop 
    eax pop 
    dword [ ebp + ( 7 * 4 ) ] inc                               \ for updated cylinder after return (EAX)
    [ ebp + ( 1 * 4 ) ] ebx mov                                 \ stored ESI on stack
    esi [ebx] mov                                               \ updated address
    popad 
    ret 

LABEL .writesector                                              \ no need to save registers because we take care of them in calling routine
    setRealModeAPI call                                         
#16 BitsMode ! 
    SECTOR_BUFFER bx mov                                        
    $0301 ax mov                                                \ read 1 sector
    $13 int 
    cli                                                         \ BIOS might have left interrupts enabled
    setProtectedModeAPI call                                    \ called from 16 bit code, returns to 32 bit code
#32 BitsMode ! 
    ret 

    ($400 - ($ - $$)) nop times                                 

\ *****************************************************************************
\ *****************************************************************************
\ After Two Sectors
\ *****************************************************************************
\ *****************************************************************************

LABEL nul  
    ret 

\ *****************************************************************************
\ cooperative multi-tasker
\ *****************************************************************************

LABEL me  
    God        ,   
LABEL x_screenTask  
    nul        ,   
LABEL x_serverTask  
    nul        ,   

LABEL pause_  
    _DUP_ 
    esi push 
    [ me ] _TOS_ mov                                            \ points to God at startup
    esp [_TOS_] mov                                             
    byte $04 _TOS_ add                                          
    _TOS_ jmp 

LABEL unpause  
    _TOS_ pop 
    [_TOS_] esp mov                                             
    _TOS_ [ me ] mov                                            
    esi pop 
    _DROP_ 
    ret 

LABEL round  
    unpause call                                                
LABEL God                                                       \ graphics update task
    0          ,                                                \ new stack location
    unpause call 
LABEL main                                                      \ main program task
    0          ,                                                \ new stack location
\    call unpause
LABEL otherTask  
\    dd RETURN_STACK_2 - 8          \  new stack location
    short round jmp                                             \ loop forever between 3 stacks

LABEL activate  \ ( a -- )    \ activate the main task to execute colorForth code at the given address
    DATA_STACK_1 - 4 edx mov                                    
    ecx [edx] mov 
    RETURN_STACK_1 - 4 ecx mov                                  
    dword [ecx] pop 
    [ ecx - $04 ] ecx lea                                       
    edx [ecx] mov 
    ecx dword [ main ] mov                                      
    _DROP_ 
    ret 

LABEL show  \ ( -- )    \ set the screen task to execute the code following  show
    dword [ x_screenTask ] pop                                  \ copy the return address of the calling word into the screenTask variable
    _DUP_ 
    _TOS_ _TOS_ xor                                             
    activate call 
LABEL .show  
    graphAction call                                            \ perform a graphical update
    [ x_screenTask ] call                                       \ execute the code that called show, saved on entry
    switch call                                                 \ copy the screen image to the VESA buffer
    _TOS_ _TOS_ xor                                             
    pause_ call 
    _TOS_ inc 
    short .show jmp 

LABEL initshow                                                  \ called by warm
    show call 
    \ <--- this address ( on the return stack from the preceding call ) goes into  x_screenTask
    ret                                                         \ makes this a no-op "show"

\ *****************************************************************************
\ "other" task execution
\ *****************************************************************************

LABEL activate2  \ ( a -- )    \ activate the  other  task to execute colorForth code at the given address
    DATA_STACK_2 - 4 edx mov                                    
    ecx [edx] mov 
    RETURN_STACK_2 - 4 ecx mov                                  
    dword [ecx] pop 
    [ ecx - $04 ] ecx lea                                       
    edx [ecx] mov 
    ecx [ otherTask ] mov                                       
    _DROP_ 
    ret 

LABEL freeze  
    dword [ x_screenTask ] pop                                  
    activate call 
LABEL .back  
    [ x_screenTask ] call                                       
    short .back jmp 

LABEL serve  
    dword [ x_serverTask ] pop                                  
    activate2 call 
LABEL .back  
    _TOS_ _TOS_ xor                                             
    pause_ call 
    [ x_serverTask ] call                                       
    short .back jmp 

LABEL initserve  
    serve call 
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL c_  \ ( -- )   \ clear the data stack for keyboard task
    ( DATA_STACK_0 + 4 ) esi mov                                
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL mark  
    [ v_MacroWordCount] ecx mov                                 
    ecx [ mark_MacroWordCount] mov                              
    [ v_ForthWordCount ] ecx mov                                
    ecx [ mark_v_ForthWordCount] mov                            
    [ v_H ] ecx mov                                             
    ecx [ mark_H ] mov                                          
    ret 

LABEL empty  
    [ mark_H ] ecx mov                                          
    ecx [ v_H ] mov                                             
    [ mark_v_ForthWordCount] ecx mov                            
    ecx [ v_ForthWordCount ] mov                                
    [ mark_MacroWordCount] ecx mov                              
    ecx [ v_MacroWordCount ] mov                                
    $0 dword [ class] mov                                       
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL mfind  \ ( sf -- )   \ ecx = index \  find the Shannon-Fano word sf in the Macro wordlist, return its index in ecx
    [ v_MacroWordCount ] ecx mov                                \ count of Macro wordlist words
    edi push 
    [ ( ecx * 4 ) + MacroNames - 4 ] edi lea                    
    short ffind jmp 

LABEL find_  \ ( sf -- )   \ ecx = index \  find the Shannon-Fano word sf in the Forth wordlist, return its index in ecx
    [ v_ForthWordCount ] ecx mov                                \ count of Forth wordlist words
    edi push 
    [ ( ecx * 4 ) + ForthNames - 4 ] edi lea                    \ set edi to the top of the Forth name table
LABEL ffind  
    std                                                         \ scan backwards
    scasd repne                                                 \ locate the 32 bit Shanon-Fano encoded name, compare eax with doubleword at es:edi and set status flags.
    cld                                                         \ reset the direction flag
    edi pop 
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL abort  
    dword [ x_abort ] jmp                                       

\ *****************************************************************************
\ *****************************************************************************

LABEL cdrop  
    [ v_H ] edx mov                                             
    edx [ list ] mov                                            
    $AD byte [edx] mov                                          \ $AD is the opcode for 'lodsd'
    dword [ v_H ] inc                                           
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL qdup  
    [ v_H ] edx mov                                             
    edx dec 
    edx dword [ list ] cmp                                      
    cdup jnz 
    $AD byte [edx] cmp                                          \ $AD is the opcode for 'lodsd'
    cdup jnz 
    edx [ v_H ] mov                                             
    ret 

LABEL cdup  \ compile action of  dup  macro
    [ v_H ] edx mov                                             
[edx], $89FC768D  constant mov                                  \ assemble the instruction sequence for  DUP  "lea esi, [ esi - 4 ]" , "mov [esi], eax"
    $006 byte [ edx + 4 ] mov                                   \  "8d 76 fc" , "89 06" ( the first 4 are expressed in little endian format above )
    byte $05 dword [ v_H ] add                                  
    ret 

LABEL adup  
    _DUP_   \ interpret action of  dup  macro
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL sdefine  
    dword [ adefine ] pop                                       
    ret 

LABEL macro  \ select the Macro wordlist
    sdefine call 
LABEL macrod  
    _TOS_ push 
    [ v_MacroWordCount] ecx mov                                 
    dword [ v_MacroWordCount] inc                               
    [ ( ecx * 4 ) + MacroNames ] ecx lea                        
    ( MacroJumpTable - MacroNames ) _TOS_ mov                   \ mov _TOS_, $218
    short forthdd jmp                                           

LABEL forth  \ select the Forth wordlist
    sdefine call 
LABEL forthd  
    _TOS_ push 
    [ v_ForthWordCount ] ecx mov                                
    dword [ v_ForthWordCount ] inc                              
    [ ( ecx * 4 ) + ForthNames ] ecx lea                        
    ( ForthJumpTable - ForthNames ) _TOS_ mov                   
LABEL forthdd  
    [ ( edi * 4 ) - $04 ] edx mov                               
    byte -$10 edx and                                           
    edx [ecx] mov 
    [ v_H ] edx mov                                             
    edx [ecx+_TOS_] mov                                         
    [ecx+_TOS_] edx lea                                         
    $02 edx shr  
    edx [ v_last ] mov                                          
    _TOS_ pop 
    esp [ list ] mov                                            
    adup dword [ lit ] mov                                      
    -1 dword [ class ] test                                     
    .fthd jz 
    dword [ class ] jmp                                         
LABEL .fthd  
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL var1  \ interpret time code for magenta variable
    _DUP_ 
    [ 4 + ForthNames + ( ecx * 4 ) ] _TOS_ mov                  
    2 _TOS_ shl 
    ret 

LABEL m_variable  \ create a magenta variable
    forthd call 
    var1 dword [ ForthJumpTable - ForthNames + ecx ] mov        
    dword [ v_ForthWordCount ] inc                              \ dummy entry for source address
    edi [ 4 + ecx ] mov                                         
    macrod call 
    .var dword [ MacroJumpTable - MacroNames + ecx ] mov        
    dword [ v_MacroWordCount ] inc                              
    edi [ 4 + ecx ] mov                                         
    edi inc 
    ret 

LABEL .var  \ compile time code for magenta variable in Macro dictionary
    [ lit ] call 
    [ 4 + MacroNames + ( ecx * 4 ) ] _TOS_ mov                  
    2 _TOS_ shl 
    short cshrt jmp 

\ *****************************************************************************
\ *****************************************************************************

LABEL alit  
    adup dword [ lit ] mov                                      

LABEL literal  
    qdup call 
    [ list ] edx mov                                            \ select the wordlist to add the literal to
    edx [ list + 4 ] mov                                        
    [ v_H ] edx mov                                             
    edx [ list ] mov                                            
    _MOV_TOS_LIT_ byte [edx] mov                                \ the opcode for mov eax, 32_bit_literal  (in next 32 bit cell)
    _TOS_ [ edx + $01 ] mov                                     \ the literal value follows in the next 4 bytes in the dictionary
    byte $05 dword [ v_H ] add                                  \ move the dictionary pointer forward 5 bytes
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL cnum  
    [ lit ] call 
    [ ( edi * 4 ) + $00 ] _TOS_ mov                             
    edi inc 
    short cshrt jmp 

LABEL cshort  
    [ lit] call 
    [ ( edi * 4 ) - $04 ] _TOS_ mov                             
    $05 _TOS_ sar  
LABEL cshrt  
    literal call 
    _DROP_ 
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL ex1  
    edi edi xor 
LABEL .back  
    dword [ v_words ] dec                                       
    ex2 jz 
    _DROP_ 
    short .back jmp 

LABEL execute_lit  \ ( -- )
    alit dword [ lit ] mov                                      
    _DUP_ 
    [ ( edi * 4 ) - $04 ] _TOS_ mov                             
LABEL execute  \ ( name -- )
    byte -$10 _TOS_ and                                         
LABEL ex2  
    find_ call 
    abort jnz 
    _DROP_ 
    dword [ ( ecx * 4 ) + ForthJumpTable ] jmp                  

\ *****************************************************************************
\ *****************************************************************************

LABEL qcompile  
    [ lit ] call 
    [ ( edi * 4 ) - $04 ] _TOS_ mov                             
    byte -$10 _TOS_ and                                         
    mfind call 
    .forward jnz 
    _DROP_ 
    dword [ ( ecx * 4 ) + MacroJumpTable ] jmp                  
LABEL .forward  
    find_ call 
    [ ( ecx * 4 ) + ForthJumpTable ] _TOS_ mov                  

LABEL qcom1  
    abort jnz 
LABEL call_  
    [ v_H ] edx mov                                             
    edx [ list ] mov                                            
    $E8 byte [edx] mov                                          \ $E8 is the opcode for 'call immediate'
    byte $05 edx add                                            
    edx _TOS_ sub 
    _TOS_ [ edx - $04 ] mov                                     
    edx [ v_H ] mov                                             
    _DROP_ 
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL compile  
    [ lit] call 
    [ ( edi * 4 ) - $04 ] _TOS_ mov                             
    byte -$10 _TOS_ and                                         
    mfind call 
    [ ( ecx * 4 ) + MacroJumpTable ] _TOS_ mov                  
    short qcom1 jmp 

\ *****************************************************************************
\ *****************************************************************************

LABEL short_  
    alit dword [ lit] mov                                       
    _DUP_ 
    [ ( edi * 4 ) - $04 ] _TOS_ mov                             
    $05 _TOS_ sar  
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL num  
    alit dword [ lit] mov                                       
    _DUP_ 
    [ ( edi * 4 ) + $00 ] _TOS_ mov                             
    edi inc 
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL comma     \ 4 byte  ,
    $04 ecx mov  
LABEL dcomma  \ c,  performed n times ( n in ecx )
    [ v_H ] edx mov                                             
    _TOS_ [edx] mov                                             
    0 [ESI] _TOS_ mov                                           
    [ECX] [EDX] edx lea                                         
    [ esi + $04 ] esi lea                                       
    edx [ v_H ] mov                                             
    ret 

LABEL comma1  \ 1 byte  c,
    $01 ecx mov  
    short dcomma jmp                                            

LABEL comma2  \ 2 byte  w,
    $02 ecx mov  
    short dcomma jmp                                            

LABEL comma3  \ 3 byte  c, c, c,
    $03 ecx mov  
    short dcomma jmp                                            

\ *****************************************************************************
\ *****************************************************************************

LABEL semicolon  
    [ v_H ] edx mov                                             
    byte $05 edx sub                                            
    edx [ list ] cmp                                            
    .forward jnz 
    $E8 byte [edx] cmp                                          \ $E8 is the opcode for 'call immediate'
    .forward jnz 
    byte [edx] inc 
    ret 
LABEL .forward  
    $C3 byte [ edx + $05 ] mov                                  \ 0xC3 is the opcode for 'ret'
    dword [ v_H ] inc                                           
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL then  
    esp [ list ] mov                                            
    [ v_H ] edx mov                                             
    _TOS_ edx sub 
    dl [ _TOS_ - $01 ] mov                                      
    _DROP_ 
    ret 

LABEL begin  
    esp [ list ] mov                                            
LABEL here  
    _DUP_ 
    [v_H] _TOS_ mov                                             
    ret 

\ *****************************************************************************
\ *****************************************************************************

LABEL qlit  \ ?lit
    [ v_H ] edx mov                                             
    [ edx - $05 ] edx lea                                       
    edx [ list ] cmp                                            
    .forward jnz 
    _MOV_TOS_LIT_ byte [edx] cmp                                \ the opcode for mov eax, 32_bit_literal  (in next 32 bit cell)
    .forward jnz 
    _DUP_ 
    [ list + 4 ] _TOS_ mov                                      
    _TOS_ [ list ] mov                                          
    [ edx + $01 ] _TOS_ mov                                     
    $89FC768D dword [ edx - 5 ] cmp                             \ assemble code 8D 76 FC 89 rr => lea esi, [ esi - $04 ] \   mov [ esi ], register
    \ like dup but with the register value still to follow in the next byte
    .forward2 jz 
    edx [ v_H ] mov                                             
    dword cdrop jmp 
LABEL .forward2  
    byte -$0A dword [ v_H ] add                                 
    ret 
LABEL .forward  
    edx edx xor 
    ret 

LABEL less  
    _TOS_ 0 [ESI] cmp                                           
    .forward js 
    ecx ecx xor 
LABEL .forward  
    ret 

LABEL qignore  
    $FFFFFFF0 dword [ ( edi * 4 ) - $04 ] test                  
    .forward jnz 
    edi pop 
    edi pop 
LABEL .forward  
    ret 

LABEL jump  
    edx pop 
    _TOS_ edx add 
    [ edx + ( _TOS_ * 4 ) + $05 ] edx lea                       
    [ edx - $04 ] edx add                                       
    _DROP_ 
    edx jmp 

\ convert block start address to cell address, add the RELOCATED colorForth system base
LABEL blockToCellAddress                                        \ ( blk -- a' )   \ add the RELOCATED offset and convert to cell address
    [ v_offset ] _TOS_ add                                      \ add the RELOCATED block number offset
    $08 _TOS_ shl                                               \ convert to cell address
    ret 

LABEL cellAddressToBlock                                        \ ( a -- blk )  \ convert cell address to block number and subtract the RELOCATED block number offset
    $08 _TOS_ shr                                               \  convert cell address to block number
    [ v_offset ] _TOS_ sub                                      \ subtract the block number of block 0
    ret 

LABEL _load_  \ ( blk -- )    \ load the given block number
    blockToCellAddress call                                     \ add the RELOCATED block number offset and convert to cell address
    edi push 
    _TOS_ edi mov 
    _DROP_ 
LABEL interpret  
    [ ( edi * 4 ) + $00 ] edx mov                               
    edi inc 
    byte $0F edx and                                            
    [ ( edx * 4 ) + tokenActions ] call                         
    short interpret jmp                                         

    db 0 4 align                                                \ fill the gap with 0's

\ : r@ qdup $8B 1, $C7 1, \     \ mov _TOS_, edi  also db $89, 0xF8
\ : nload r@ $0100 / #2 + load \ 
\ : +load ( n -- ) r@ $0100 / + load \ 
LABEL nload  \ ( -- )    \ load the next source block following the one currently being loaded
    cblk call 
    $02 _TOS_ add  
    _load_ jmp 

LABEL plusLoad  \ ( n -- )    \ load the n'th source block following the one currently being loaded
_TOS_  constant mov                                             \ save the required offset
    _DROP_ 
    cblk call 
    _SCRATCH_ _TOS_ add                                         
    _load_ jmp 

\ : THRU ( f l -- )   1+ SWAP DO  I LOAD  LOOP \ 
LABEL thru_  \ ( first last -- )   \ load from the first to the last block
    $02 _TOS_ add  
    _TOS_ _SCRATCH_ mov                                         
    _DROP_                                                      \ TOS = first, SCRATCH = last
    _SCRATCH_ ecx mov                                           
    _TOS_ ecx sub                                               \ ecx = count
    .end jz                                                     \ exit if count is zero
    .end jc                                                     \ exit if count is negative
    1 ecx shr                                                   \ divide by 2, as we skip 2 blocks each time round the loop
LABEL .back  
    _DUP_ 
    ecx push 
    _load_ call 
    ecx pop 
    $02 _TOS_ add  
    .back loop 
LABEL .end  
    _DROP_ 
    ret 

LABEL v_temp  
    18         ,   

LABEL plusThru_  \ ( first+ last+ -- )   \ load from the first to the last block relative to the current block being loaded
    cblk call 
    _TOS_ [ v_temp ] mov                                        
    _DROP_ 
    [ v_temp ] _SCRATCH_ mov                                    
    _SCRATCH_ 0 [ESI] add                                       \ add current block to second on stack
    _SCRATCH_ _TOS_ add                                         \ add current block to top of stack
    thru_ call 
    ret 

LABEL cblk  \ ( -- n )  \ return the currently compiling block number - only valid while compiling
    _DUP_ 
    edi _TOS_ mov                                               \ edi  contains the cell address in the block currently being compiled
    cellAddressToBlock call                                     \ convert to block number relative to block 0
    ret 

LABEL v_numberOfMagentas  
    0          ,   

LABEL v_curs_to_source                                          \ ( -- a32 )   \ return the cell address of the current cursor position in the current block being edited
    _DUP_ 
    [ v_blk ] _TOS_ mov                                         \ get the currently edited block number
    blockToCellAddress call                                     
    dword [ v_curs ] _TOS_ add                                  \ add the cursor position (cell address) in the block
    ret 

LABEL v_curs_to_token  \ ( -- tok )   \ return the token at the current cursor position in the current block being edited
    v_curs_to_source call                                       
    $02 _TOS_ shl                                               \ convert cell address to byte address
    [ _TOS_ ] _TOS_ mov                                         \ fetch the token
    ret 

\ : ?f $C021 2, \ 
\  f:
\    db $21, 0xC0   \  and _TOS_, _TOS_
\    ret

\ *****************************************************************************
\ *****************************************************************************

LABEL top_  \ ( -- )   \ set the cursor to the left margin horizontally and 3 pixels down from the top vertically
    [ v_leftMargin ] ecx mov                                    
    $10 ecx shl  
    byte $03 ecx add                                            
    ecx [ v_xy ] mov                                            
    \ mov [ xycr], ecx
    ret 

LABEL qcr  \ ( -- )   \ ?cr  do a  CR  if the cursor has gone past the right margin
    [ v_x ] cx mov 
    [ v_rightMargin ] cx cmp                                    
    cr_forward js 
LABEL cr_  \ ( -- )
    [ v_leftMargin ] ecx mov                                    
    $10 ecx shl  
    [ v_xy ] cx mov                                             
    [ v_iconh ] cx add                                          
    ecx [ v_xy ] mov                                            
LABEL cr_forward  
    ret 

LABEL green  \ ( -- )
    _DUP_ 
    colour_green _TOS_ mov                                      
    color jmp 

LABEL yellow  \ ( -- )
    _DUP_ 
    colour_yellow _TOS_ mov                                     
    color jmp 

LABEL red  \ ( -- )
    _DUP_ 
    colour_red _TOS_ mov                                        
    color jmp 

LABEL white  \ ( -- )
    _DUP_ 
    colour_white _TOS_ mov                                      
LABEL color  \ ( rgb16 -- )
    _TOS_ [ v_foregroundColour ] mov                            
    _DROP_ 
    ret 

LABEL rgb  \ ( rgb32 -- rgb16 )    \  convert from 32 bit ( 8:8:8:8 _RGB ) colour to 16 bit ( 5:6:5 RGB ) colour value
    8 _TOS_ ror 
    2 ax shr 
    6 _TOS_ ror 
    3 al shr 
    ( 6 + 5 ) _TOS_ rol                                         
    $0000FFFF _TOS_ and                                         
    ret 

LABEL bye_  \ ( -- )   \ exit colorForth
    setRealModeAPI call                                         
#16 BitsMode ! 
    $19 int     \ reboot the computer
    \ should never get past this point.... but in case we do...
    cli                                                         \ BIOS might have left interrupts enabled
    setProtectedModeAPI call                                    \ called from 16 bit code, returns to 32 bit code
#32 BitsMode ! 
    ret 

0 [IF]  
LABEL pci  
    $0CF8 edx mov  
    _TOS_ dx out 
    [ edx + $04 ] edx lea                                       
    dx _TOS_ in 
    ret 

LABEL device  
    ( $93a - ( $ - $$ ) ) nop times                             \ fill with nops to find_display ???

LABEL find_display                                              \ called by warm
    $3000000 _TOS_ mov                                          \ PCI class code 3 = display controller
    device call                                                 \ returns header address
    [ _TOS_ + $10 ] _TOS_ lea                                   \ point to Base Address #0 (BAR0)
    $06 cl mov  
LABEL .next  
    _DUP_ 
    pci call 
    $FB al and  
    $08 al xor  
    .forward jz 
    _DROP_ 
    [ _TOS_ + $04 ] _TOS_ lea                                   
    .next loop 
    [ _TOS_ - $18 ] _TOS_ lea                                   
    _DUP_ 
    pci call 
    $F0 al and  
LABEL .forward  
    _TOS_ [ v_frameBuffer ] mov                                 \ set framebuffer address
    _DROP_ 
    ret 

LABEL fifo  
    _DROP_ 
    ret 

[THEN]  

LABEL graphAction  
    ret 

\ *****************************************************************************
\ *****************************************************************************
\ grapics mode dependent code
\ *****************************************************************************
\ *****************************************************************************

\ *****************************************************************************
\ 1024x768 display
\ *****************************************************************************

1024  constant scrnw1                                           \ screen width in pixels
768  constant scrnh1                                            \ screen height in pixels
( 16 + 4 )  constant iconw1                                     \ icon width
( 24 + 4 )  constant iconh1                                     \ icon height for 768 pixel high screen

4   constant keypadY1                                           \ location of keyboard display vertically in lines from the bottom

LABEL initIconSize1  
    iconw1 dword [ v_iconw ] mov                                
    ( iconw1 * 9 ) dword [ v_nine_iconw ] mov                   
    ( iconw1 * ( 13 + 9 ) ) dword [ v_twentytwo_iconw ] mov     
    ( iconw1 * $10000 ) dword [ v_10000_iconw ] mov             
    iconh1 dword [ v_iconh ] mov                                
    keypadY1 * iconh1 dword [ v_keypadY_iconh ] mov             
    ret 

LABEL switch1  \ copy our created image to the real display buffer
    esi push 
    edi push 
    dword [ frame ] esi mov                                     \ frame  points to where we create our image
    [ vesa_PhysBasePtr ] edi mov                                \ VESA frame buffer, saved by VESA BIOS call, the address in RAM that is displayed by the hardware
    ( ( scrnw1 * scrnh1 ) / 4 ) * BYTES_PER_PIXEL ecx mov       \ the / 4 is because we are moving doubles = 4 bytes each
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    edi pop 
    esi pop 
    ret 

LABEL clip1  
    [ v_xy ] edi mov                                            
    edi ecx mov  
    cx cx test 
    .forward jns 
    ecx ecx xor  
LABEL .forward  
    $0000FFFF ecx and                                           
    ecx [ v_yc ] mov                                            
    ( scrnw1 * BYTES_PER_PIXEL ) ecx imul                       
    16 edi sar  
    .forward2 jns 
    edi edi xor  
LABEL .forward2  
    edi [ v_xc ] mov                                            
    [ edi * BYTES_PER_PIXEL + ecx ] edi lea                     
    [ frame ] edi add                                           
    ret 

LABEL bit16                                                     \ write a 16 x 24 glyph to the graphic screen
    lodsw                                                       \ load the 16 bit value pointed to by SI into  ax
    ah al xchg                                                  \ eax_TOS_
LABEL .back  
    $01 ax shl                                                  \ eax_TOS_
    .forward jnc 
    dx [ edi ] mov                                              \ dx is either  dx  or  edx  depending on BITS_PER_PIXEL  = 16 or 32
    .forward2 jmp 
LABEL .forward  
    $10 edx ror                                                 \ use the background colour, in the high 16 bits
\    mov [ edi ], dx        \  dx is either  dx  or  edx  depending on BITS_PER_PIXEL  = 16 or 32
    $10 edx ror                                                 \ return to the foreground colour, in the low 16 bits
LABEL .forward2  
    byte BYTES_PER_PIXEL edi add                                
    .back loop 
    ret 

\ write the background after the glyph
LABEL bit16Background                                           \ number of pixels to write in  ecx ,  screen address in  edi , colours in edx
    $10 edx ror                                                 \ use the background colour, in the high 16 bits
LABEL .back  
\    mov [ edi ], dx        \  dx is either  dx  or  edx  depending on BITS_PER_PIXEL  = 16 or 32
    byte BYTES_PER_PIXEL edi add                                
    .back loop 
    $10 edx ror                                                 \ return to the foreground colour, in the low 16 bits
    ret 

LABEL bit32                                                     \ write a 32 x 48 double size glyph to the graphic screen
    lodsw                                                       \ load the 16 bit value pointed to by SI into  ax
    ah al xchg                                                  \ eax_TOS_
    $10 ecx mov  
LABEL .back  
    1 _TOS_ shl                                                 \ eax_TOS_
    .forward jnc 
    \ dx is either  dx  or  edx  depending on BITS_PER_PIXEL  = 16 or 32
    dx [ edi ] mov                                              
    dx [ edi + BYTES_PER_PIXEL ] mov                            

    0 byte [ displayMode ] cmp                                  
    .width2 jnz 
    dx [ edi + ( scrnw1 * BYTES_PER_PIXEL ) ] mov               
    dx [ edi + ( scrnw1 * BYTES_PER_PIXEL ) + BYTES_PER_PIXEL ] mov 
    .widthEnd jmp 
LABEL .width2  
    dx [ edi + ( scrnw2 * BYTES_PER_PIXEL ) ] mov               
    dx [ edi + ( scrnw2 * BYTES_PER_PIXEL ) + BYTES_PER_PIXEL ] mov 
LABEL .widthEnd  
LABEL .forward  
    byte ( BYTES_PER_PIXEL * 2 ) edi add                        
    .back loop 
    ret 

LABEL emit1  \ ( c -- )   \ display a single width and height character
    qcr call 
    esi push 
    edi push 
    edx push 
    byte 16*24/8 _TOS_ imul                                     
    [ _TOS_ + font16x24 ] esi lea                               
    clip1 call 
    [ v_foregroundColour ] edx mov                              
    $18 ecx mov                                                 \ 24 lines
LABEL .back  
    ecx push 
    $10 ecx mov  
    bit16 call 
    $04 ecx mov  
    edi push 
    bit16Background call                                        
    edi pop 
    ecx pop 
    ( scrnw1 - 16 ) * BYTES_PER_PIXEL edi add                   \ address of the next line of the glyph
    .back loop                                                  \ next horizontal line

    $04 ecx mov                                                 \ 4 background lines
LABEL .back2  
    ecx push 
    $10 ecx mov  
    bit16Background call                                        
    $04 ecx mov  
    edi push 
    bit16Background call                                        
    edi pop 
    ecx pop 
    ( scrnw1 - 16 ) * BYTES_PER_PIXEL edi add                   \ address of the next line of the glyph
    .back2 loop                                                 \ next horizontal line

    edx pop 
    edi pop 
    esi pop 
    _DROP_ 
LABEL space1  
    iconw1 * $10000 dword [ v_xy ] add                          \ 22 horizontal pixels
    ret 

LABEL two_emit1  \ double width and height character
    esi push 
    edi push 
    edx push 
    byte 16*24/8 _TOS_ imul                                     
    [ _TOS_ + font16x24 ] esi lea                               
    clip1 call 
    [ v_foregroundColour ] edx mov                              
    24 ecx mov 
LABEL .back  
    ecx push 
    bit32 call 
    (2*scrnw1-16*2)*BYTES_PER_PIXEL edi add                     
    ecx pop 
    .back loop 
    edx pop 
    edi pop 
    esi pop 
    iconw1 * 2 * $10000 dword [ v_xy ] add                      \ 44 horizontal pixels
    _DROP_ 
    ret 

LABEL setupText__1  \ setup for full screen text window display
    white call 
    $03 dword [ v_leftMargin ] mov                              
    ( scrnw1 - iconw1 ) dword [ v_rightMargin ] mov             
    dword top_ jmp 

LABEL box1  \ ( width height -- )
    clip1 call 
    scrnh1+1 _TOS_ cmp                                          
    .forward js 
    scrnh1 _TOS_ mov                                            
LABEL .forward  
    _TOS_ ecx mov 
    [ v_yc ] ecx sub                                            
    .forward3 jng 
    scrnw1+1 dword [esi] cmp                                    
    .forward2 js 
    scrnw1 dword [esi] mov                                      
LABEL .forward2  
    [ v_xc ] _TOS_ mov                                          
    _TOS_ [esi] sub                                             
    .forward3 jng 
    scrnw1 edx mov 
    [esi] edx sub 
    PIXEL_SHIFT edx shl                                         
    [ v_foregroundColour ] _TOS_ mov                            
LABEL .back  
    ecx push 
    [esi] ecx mov 
    stosw rep   \ stosw depends on BYTES_PER_PIXEL, either stosw or stosd
    edx edi add 
    ecx pop 
    .back loop 
LABEL .forward3  
    _DROP_ 
    _DROP_ 
    ret 

LABEL wash1  \ ( colour -- )   \ fill the full screeen with the given colour
    color call 
    _DUP_ 

    _TOS_ _TOS_ xor                                             \ x,y = 0,0 top left corner
    _TOS_ [ v_xy ] mov                                          

    scrnw1 _TOS_ mov                                            
    _DUP_ 
    scrnh1 _TOS_ mov                                            
    dword box_ jmp 

\ *****************************************************************************
\ 800x600 screen
\ *****************************************************************************

800  constant scrnw2                                            \ screen width in pixels
600  constant scrnh2                                            \ screen height in pixels
( 16 + 1 )  constant iconw2                                     \ icon width
( 24 - 1 )  constant iconh2                                     \ icon height for NC10 600 pixel high screen

4   constant keypadY2                                           \ location of keyboard display vertically in lines from the bottom

LABEL initIconSize2  
    iconw2 dword [ v_iconw ] mov                                
    ( iconw2 * 9 ) dword [ v_nine_iconw ] mov                   
    ( iconw2 * ( 13 + 9 ) ) dword [ v_twentytwo_iconw ] mov     
    ( iconw2 * $10000 ) dword [ v_10000_iconw ] mov             
    iconh2 dword [ v_iconh ] mov                                
    keypadY2 * iconh2 dword [ v_keypadY_iconh ] mov             
    ret 

LABEL switch2  \ copy our created image to the real display buffer
    esi push 
    edi push 
    dword [ frame ] esi mov                                     \ frame  points to where we create our image
    [ vesa_PhysBasePtr ] edi mov                                \ VESA frame buffer, saved by VESA BIOS call, the address in RAM that is displayed by the hardware
    ( ( scrnw2 * scrnh2 ) / 4 ) * BYTES_PER_PIXEL ecx mov       \ the / 4 is because we are moving doubles = 4 bytes each
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    edi pop 
    esi pop 
    ret 

LABEL clip2  
    [ v_xy ] edi mov                                            
    edi ecx mov  
    cx cx test 
    .forward jns 
    ecx ecx xor  
LABEL .forward  
    $0000FFFF ecx and                                           
    ecx [ v_yc ] mov                                            
    ( scrnw2 * BYTES_PER_PIXEL ) ecx imul                       
    16 edi sar  
    .forward2 jns 
    edi edi xor  
LABEL .forward2  
    edi [ v_xc ] mov                                            
    [ edi * BYTES_PER_PIXEL + ecx ] edi lea                     
    [ frame ] edi add                                           
    ret 

LABEL emit2  \ ( c -- )   \ display a single width and height character
    qcr call 
    esi push 
    edi push 
    edx push 
    byte 16*24/8 _TOS_ imul                                     
    [ _TOS_ + font16x24 ] esi lea                               
    clip2 call 
    [ v_foregroundColour ] edx mov                              
    $18 ecx mov                                                 \ 24 lines
LABEL .back  
    ecx push 
    $10 ecx mov  
    bit16 call 
    $04 ecx mov  
    edi push 
    bit16Background call                                        
    edi pop 
    ecx pop 
    ( scrnw2 - 16 ) * BYTES_PER_PIXEL edi add                   \ address of the next line of the glyph
    .back loop                                                  \ next horizontal line

    $04 ecx mov                                                 \ 4 background lines
LABEL .back2  
    ecx push 
    $10 ecx mov  
    bit16Background call                                        
    $04 ecx mov  
    edi push 
    bit16Background call                                        
    edi pop 
    ecx pop 
    ( scrnw2 - 16 ) * BYTES_PER_PIXEL edi add                   \ address of the next line of the glyph
    .back2 loop                                                 \ next horizontal line

    edx pop 
    edi pop 
    esi pop 
    _DROP_ 
LABEL space2  
    iconw2 * $10000 dword [ v_xy ] add                          \ 22 horizontal pixels
    ret 

LABEL two_emit2  \ double width and height character
    esi push 
    edi push 
    edx push 
    byte 16*24/8 _TOS_ imul                                     
    [ _TOS_ + font16x24 ] esi lea                               
    clip2 call 
    [ v_foregroundColour ] edx mov                              
    24 ecx mov 
LABEL .back  
    ecx push 
    bit32 call 
    (2*scrnw2-16*2)*BYTES_PER_PIXEL edi add                     
    ecx pop 
    .back loop 
    edx pop 
    edi pop 
    esi pop 
    iconw2 * 2 * $10000 dword [ v_xy ] add                      \ 44 horizontal pixels
    _DROP_ 
    ret 

LABEL setupText__2  \ setup for full screen text window display
    white call 
    $03 dword [ v_leftMargin ] mov                              
    ( scrnw2 - iconw2 ) dword [ v_rightMargin ] mov             
    dword top_ jmp 

LABEL box2  \ ( width height -- )
    clip2 call 
    scrnh2+1 _TOS_ cmp                                          
    .forward js 
    scrnh2 _TOS_ mov                                            
LABEL .forward  
    _TOS_ ecx mov 
    [ v_yc ] ecx sub                                            
    .forward3 jng 
    scrnw2+1 dword [esi] cmp                                    
    .forward2 js 
    scrnw2 dword [esi] mov                                      
LABEL .forward2  
    [ v_xc ] _TOS_ mov                                          
    _TOS_ [esi] sub                                             
    .forward3 jng 
    scrnw2 edx mov 
    [esi] edx sub 
    PIXEL_SHIFT edx shl                                         
    [ v_foregroundColour ] _TOS_ mov                            
LABEL .back  
    ecx push 
    [esi] ecx mov 
    stosw rep   \ stosw depends on BYTES_PER_PIXEL, either stosw or stosd
    edx edi add 
    ecx pop 
    .back loop 
LABEL .forward3  
    _DROP_ 
    _DROP_ 
    ret 

LABEL wash2  \ ( colour -- )   \ fill the full screeen with the given colour
    color call 
    _DUP_ 

    _TOS_ _TOS_ xor                                             \ x,y = 0,0 top left corner
    _TOS_ [ v_xy ] mov                                          

    scrnw2 _TOS_ mov                                            
    _DUP_ 
    scrnh2 _TOS_ mov                                            
    dword box_ jmp 

\ *****************************************************************************
\ select which display mode code to use
\ *****************************************************************************

LABEL displayMode  
    1          ,   \ 0 = 1024x768x16, 1 = 80$600x16

LABEL initIconSize  
    0 byte [ displayMode ] cmp                                  
    initIconSize1 jz                                            
    initIconSize2 jmp                                           

LABEL switch  
    0 byte [ displayMode ] cmp                                  
    switch1 jz 
    switch2 jmp 

LABEL clip  
    0 byte [ displayMode ] cmp                                  
    clip1 jz 
    clip2 jmp 

LABEL emit_  
    0 byte [ displayMode ] cmp                                  
    emit1 jz 
    emit2 jmp 

LABEL space_  
    0 byte [ displayMode ] cmp                                  
    space1 jz 
    space2 jmp 

LABEL two_emit  
    0 byte [ displayMode ] cmp                                  
    two_emit1 jz 
    two_emit2 jmp 

LABEL setupText_  \ setup for full screen text window display
    0 byte [ displayMode ] cmp                                  
    setupText__1 jz 
    setupText__2 jmp                                            

LABEL line_  \ ( startX length -- )   \ draw a horizontal line in the current colour, from startX relative to current clip window, of given length in pixels
    0 byte [ displayMode ] cmp                                  
    .forward jnz 
    clip1 call 
    .common jmp 
LABEL .forward  
    clip2 call 
LABEL .common  
    [esi] ecx mov 
    PIXEL_SHIFT ecx shl                                         
    ecx edi sub 
    _TOS_ ecx mov 
    [ v_foregroundColour ] _TOS_ mov                            
    stosw rep   \ stosw is either  stosw  or  stosd  depending on BITS_PER_PIXEL = 16 or 32
    dword [ v_xy ] inc                                          
    _DROP_ 
    _DROP_ 
    ret 

LABEL box_  
    0 byte [ displayMode ] cmp                                  
    box1 jz 
    box2 jmp 

LABEL page_  \ ( -- )    \ fill the full screen with the current background colour
    _DUP_ 
    colour_background _TOS_ mov                                 \
    wash_ jmp 

LABEL screen_  \ ( -- )   \ fill the full screen with the current foreground colour
    _DUP_ 
    [ v_foregroundColour ] _TOS_ mov                            \     \  select the foreground colour in the low 16 bits
\    jmp wash_                      \  fall through to wash1

LABEL wash_  \ ( colour -- )   \ fill the full screeen with the given colour
    _TOS_ [ v_washColour ] mov                                  
    0 byte [ displayMode ] cmp                                  
    wash1 jz 
    wash2 jmp 

\ *****************************************************************************
\ *****************************************************************************
\ *****************************************************************************

LABEL setCyan  
    _DUP_ 
    colour_cyan _TOS_ mov                                       
    dword color jmp 

LABEL setMagenta  
    _DUP_ 
    colour_magenta _TOS_ mov                                    
    dword color jmp 

LABEL setBlue  
    _DUP_ 
    colour_blue _TOS_ mov                                       
    dword color jmp 

LABEL setRed  
    _DUP_ 
    colour_red _TOS_ mov                                        
    dword color jmp 

LABEL setGreen  
    _DUP_ 
    colour_green _TOS_ mov                                      
    dword color jmp 

LABEL setSilver  
    _DUP_ 
    colour_silver _TOS_ mov                                     
    dword color jmp 

LABEL history  
    11 db 0 times 

LABEL echo_  
    esi push 
    11-1 ecx mov 
    [ history ] edi lea                                         
    [edi+1] esi lea                                             
    movsb rep 
    esi pop 
    al byte [ history+11-1 ] mov                                
    _DROP_ 
    ret 

LABEL right  
    _DUP_ 
    11 ecx mov 
    [history] edi lea                                           
    _TOS_ _TOS_ xor                                             
    stosb rep 
    _DROP_ 
    ret 

LABEL down  
    _DUP_ 
    edx edx xor 
    [ v_iconh ] ecx mov                                         
    ecx div 
    edx _TOS_ mov 
    [ v_iconh ] edx sub                                         
    ( 3 * $10000 )+ $8000 + 3 edx add                           
    edx [ v_xy ] mov                                            
\ zero:
    _TOS_ _TOS_ test                                            
    0 _TOS_ mov 
    .dw jnz 
    _TOS_ inc 
LABEL .dw  
    ret 

LABEL lm  \ ( leftMargin -- )
    _TOS_ [ v_leftMargin ] mov                                  
    _DROP_ 
    ret 

LABEL rm  \ ( rightMargin -- )
    _TOS_ [ v_rightMargin ] mov                                 
    _DROP_ 
    ret 

LABEL _at  \ ( y x -- )
    ax word [ v_y ] mov                                         
    _DROP_ 
    ax word [ v_x ] mov                                         
    _DROP_ 
    ret 

LABEL plus_at  \ ( y x -- )
    ax word [ v_y ] add                                         
    _DROP_ 
    ax word [ v_x ] add                                         
    _DROP_ 
    ret 

0 [IF]  
\ the various pieces of code used by a! and +! in colorForth blocks 22 and 24
LABEL  plusStore  \ ( n a -- )
    \  : a! ?lit if $BA 1, , \  then $D08B 2, drop ;
    $12345678 dword edx mov                                     \ db $BA, 0x78, 0x56, 0x34, 0x12
    _TOS_ edx mov                                               \ db $8B, $D0 == db 0x89, 0xC2
    \  : +! ?lit if ?lit if $0581 2, swap a, , \  then $0501 2, a, drop ; then a! $0201 2, drop ;
    _TOS_ [ dword $12345678 ] add                               \ db $01, 0x05, 0x78, 0x56, 0x34, 0x12
    $98765432 dword [ dword $12345678 ] add                     \ db 0x81, 0x05, 0x78, 0x56, 0x34, 0x12, 0x32, 0x54, 0x76, 0x98
    _TOS_ [ edx ] add                                           \ db $01, $02
    ret 
[THEN]  

LABEL octant  
    _DUP_ 
    $43 _TOS_ mov  
    [ esi + $04 ] edx mov                                       
    edx edx test 
    .forward jns 
    edx neg 
    edx [ esi + $04 ] mov                                       
    $01 al xor  
LABEL .forward  
    0 [ESI] edx cmp                                             
    .forward2 jns 
    $04 al xor  
LABEL .forward2  
    ret 

LABEL hicon  
    $18 c,  $19 c,  $1A c,  $1B c,  $1C c,  $1D c,  $1E c,  $1F c,  
    $20 c,  $21 c,  $05 c,  $13 c,  $0A c,  $10 c,  $04 c,  $0E c,  

LABEL edig1  
    _DUP_ 
LABEL digit  
    ecx push 
    [ _TOS_ + hicon ] al mov                                    
    emit_ call 
    ecx pop 
    ret 

LABEL odig  
    $04 _TOS_ rol  
    _DUP_ 
    byte $0F _TOS_ and                                          
    ret 

LABEL h_dot_n  
    _TOS_ edx mov 
    _TOS_ neg 
    [ ( _TOS_ * 4 ) + $20 ] ecx lea                             
    _DROP_ 
    cl _TOS_ rol 
    edx ecx mov 
    short h_dot2 jmp                                            

LABEL dotHex8  \ ( u -- )   \ display a hexadecimal number with leading zeros, 8 .hex
    $08 ecx mov  
LABEL h_dot2  
    odig call 
    digit call 
    h_dot2 loop 
    _DROP_ 
    ret 

LABEL dotHex  \ ( u -- )   \ display a hexadecimal number
    $6F EMIT_IMM  \ '$'
    $07 ecx mov  
LABEL .back  
    odig call 
    .forward jnz 
    _DROP_ 
    .back loop 
    ecx inc 
LABEL .back2  
    odig call 
LABEL .back3  
    digit call 
    .back2 loop 
    space_ call 
    _DROP_ 
    ret 
LABEL .forward  
    ecx inc 
    short .back3 jmp                                            

LABEL qdot  \ ( u -- )   \ display a decimal or hexadecimal number, depending on base
    byte 10 dword [ base ] cmp                                  
    dotHex jnz 
LABEL dotDecimal                                                \ display a decimal number
\    EMIT_IMM $64 \  '#'
    _TOS_ edx mov 
    edx edx test 
    .forward jns 
    constant neg                                                \ display a negative sign if required
    _DUP_ 
    $23 _TOS_ mov                                               \ '-'
    emit_ call 
LABEL .forward  
    $08 ecx mov  
LABEL .back  
    edx _TOS_ mov 
    edx edx xor 
    dword [ ecx * 4 + tens ] div                                
    _TOS_ _TOS_ test                                            
    .forward2 jnz 
    ecx dec 
    .back jns 
    short .forward3 jmp                                         
LABEL .back2  
    edx _TOS_ mov 
    edx edx xor 
    dword [ ecx * 4 + tens ] div                                
LABEL .forward2  
    edig1 call 
    ecx dec 
    .back2 jns 
LABEL .forward3  
    edx _TOS_ mov 
    edig1 call 
    space_ call 
    _DROP_ 
    ret 

LABEL eight  
    byte $0C edi add                                            
    four call 
    space_ call 
    byte $10 edi sub                                            
LABEL four  
    $04 ecx mov  
LABEL four1   constant  \ set ecx to the required number of characters to display, so 'four'1 is a misnomer...
    ecx push 
    _DUP_ 
    _TOS_ _TOS_ xor                                             
    [edi+$04] al mov                                            
    edi inc 
    emit_ call 
    ecx pop 
    four1 loop 
    ret 

LABEL displayTheStack  \ display the stack
    ( DATA_STACK_0 - 4 ) edi mov                                \ save empty stack pointer, plus one ( stack grows downwards )
LABEL .back  
    [ God ] edx mov                                             \ copy the current stack pointer
    edi [edx] cmp 
    .forward jnc                                                \ test for empty stack, meaning done
    _DUP_ 
    [edi] _TOS_ mov                                             \ fetch the value of the current stack item
    byte $04 edi sub                                            \
    qdot call                                                   \ display one stack item
    short .back jmp                                             \ next stack item
LABEL .forward  
    ret 

3   constant yShift 

LABEL displayBlockNumber                                        \  ( -- )     \  in the top right corner of the screen
    _DUP_ 
    [ v_foregroundColour ] _TOS_ mov                            
    _DUP_ 
    [ vesa_XResolution ] _TOS_ mov                              \ was this  : mov _TOS_, ( scrnw )
    $FFFF _TOS_ and                                             
    [ v_nine_iconw ] _TOS_ sub                                  
    _TOS_ _SCRATCH_ mov                                         \ save for later
    _TOS_ [ v_leftMargin ] mov                                  
    ax [ word v_y ] mov                                         
    [ v_nine_iconw ] _TOS_ add                                  
    _TOS_ [ v_rightMargin ] mov                                 
    _SCRATCH_ _TOS_ mov                                         
    16 _TOS_ shl 
    yShift _TOS_ add                                            
    _TOS_ [ v_xy ] mov                                          
    _DUP_ 
    [ v_washColour ] _TOS_ mov                                  \ so we do not see the number yet, just measure its width
\    mov _TOS_, colour_blockNumber
\    shr _TOS_, 16                 \  select the background colour in the high 16 bits
    color call 
    _DUP_ 
    [ v_blk ] _TOS_ mov                                         
    qdot call 
    [ v_xy ] _SCRATCH_ mov                                      \ current x,y coordinate, x in high 16 bits
    16 _SCRATCH_ shr                                            
    [ v_leftMargin ] _SCRATCH_ sub                              \ _SCRATCH_ is now the width of number string, in pixels
    [ v_iconw ] _SCRATCH_ sub                                   \ correction...
    16 _SCRATCH_ shl                                            
    [ vesa_XResolution ] _TOS_ mov                              \ screen width in pixels
    \ and _TOS_, $FFFF   \  not needed because of the  shl  below
    16 _TOS_ shl 
    yShift _TOS_ add                                            
    _SCRATCH_ _TOS_ sub                                         
    _TOS_ [ v_xy ] mov                                          

    _DUP_ 
    colour_blockNumber _TOS_ mov                                
    16 _TOS_ ror 
    color call 
    _DUP_ 
    [ v_iconw ] _TOS_ mov                                       
    _TOS_ _TOS_ add                                             
    _DUP_ 
    [ v_iconh ] _TOS_ mov                                       
    box_ call 
    _TOS_ [ v_xy ] mov                                          

    colour_blockNumber _TOS_ mov                                
    _DUP_ 
    color call 
    _DUP_ 
    [ v_blk ] _TOS_ mov                                         
\    mov _TOS_, [ v_numberOfMagentas ]

    qdot call 
    _DROP_ 
    _TOS_ [ v_foregroundColour ] mov                            
    _DROP_ 
    ret 

\ *****************************************************************************
\ keyboard displays
\ *****************************************************************************

LABEL showEditBox  \ v_at set up for start coordinate of box, width and height on stack
    $000C0004 dword [ v_xy ] sub                                \ move the start position left  and up   by $XXXXYYYY
    [ v_foregroundColour ] dword _SCRATCH_ mov                  
    colour_orange dword [ v_foregroundColour ] mov              
    2 ecx mov 
LABEL .loop  
    ecx push 
    _DUP_ 
    0 _TOS_ mov                                                 \ SOS = x start position in pixels, relative to current clip "window"
    _DUP_ 
    [ v_iconw ] _TOS_ mov                                       
    3 _TOS_ shl                                                 \ multiply by 8
    [ v_iconw ] _TOS_ add                                       \ multiply by 9
    [ v_iconw ] _TOS_ add                                       \ multiply by 10
    \ TOS = length of horizontal line in pixels
    line_ call 
    [ v_iconh ] ecx mov                                         
    2 ecx shl                                                   \ multiply by 4
    4 ecx add                                                   \ draw the lower line below the text
    ecx dword [ v_xy ] add                                      \ move the start position down by 4 character heights
    ecx pop 
    .loop loop 

    _SCRATCH_ dword [ v_foregroundColour ] mov                  
    ret 

LABEL displayTheKeyboard                                        \ the keyboard is the mnemonic at the bottom right of the display, showing the actions of each of the 27 keys used
    setupText_ call 
    [ dword currentKeyboardIcons ] edi mov                      
    _DUP_ 
    [ keyboard_colour ] _TOS_ mov                               
    color call 
    [ vesa_XResolution ] _TOS_ mov                              \ was this  : mov _TOS_, ( scrnw )
    $FFFF _TOS_ and                                             
    [ v_nine_iconw ] _TOS_ sub                                  
    16 _TOS_ sub 
    _TOS_ [ v_leftMargin ] mov                                  \ x coordinate of left margin of keyboard display
    _TOS_ edx mov                                               \
    [ v_nine_iconw ] edx add                                    \ x coordinate of right margin of keyboard display
    edx [ v_rightMargin ] mov                                   
    $10 _TOS_ shl  
    [ vesa_YResolution ] edx mov                                \ was this  : mov _TOS_, ( scrnw )
    $0000FFFF edx and                                           
    _SCRATCH_ push 
    [ v_keypadY_iconh ] _SCRATCH_ mov                           
    10 _SCRATCH_ add                                            
    _SCRATCH_ edx sub                                           \ ( ( keypadY * iconh ) + 10 )
    edx _TOS_ add 

    _TOS_ [ v_xy ] mov                                          

    $FF byte [ v_acceptMode ] test                              
    .forward jz 
    pusha 
    showEditBox call                                            
    popa 
    _TOS_ [ v_xy ] mov                                          
LABEL .forward  

    _SCRATCH_ pop 
    eight call 
    eight call 
    eight call 
    cr_ call 
\    add dword [ v_xy ],  ( 4 * iconw * $10000 )        \  shift horizontal pixels to the right
    [ v_iconw ] _SCRATCH_ mov                                   
    ( 2 + 16 ) _SCRATCH_ shl                                    \ ( 4 * iconw * $10000 )        \  shift horizontal pixels to the right
    _SCRATCH_ dword [ v_xy ] add                                
    [ shiftAction ] edi mov                                     
    byte $0C edi add                                            
    $03 ecx mov  
    four1 call 

    space_ call 
    _DUP_ 
    [ v_hintChar ] _TOS_ mov                                    
    emit_ call 

    $03 dword [ v_leftMargin ] mov                              
    $03 word [ v_x ] mov                                        
    displayTheStack call                                        
    [ vesa_XResolution ] _TOS_ mov                              \ was this  : mov _TOS_, ( scrnw )
    $FFFF _TOS_ and                                             
    [ v_twentytwo_iconw ] _TOS_ sub                             
    3 _TOS_ add 
    ax word [ v_x ] mov                                         
    [ ( history - 4 )] edi lea                                  \ the text entered so far
    $0B ecx mov  
    dword four1 jmp 

\ *****************************************************************************

LABEL alphaKeyboard                                             \ the 'alpha' character keyboard icons, the start screen for key entry
    $0D c,  $0A c,  $01 c,  $0C c,                              \ g c r l
    $14 c,  $02 c,  $06 c,  $08 c,                              \ h t n s
    $13 c,  $09 c,  $0F c,  $11 c,                              \ b m w v
    $12 c,  $0B c,  $0E c,  $07 c,                              \ p y f i
    $05 c,  $03 c,  $04 c,  $16 c,                              \ a o e u
    $17 c,  $24 c,  $15 c,  $10 c,                              \ q k x d

LABEL graphicsKeyboard                                          \ the 'graphics' character keyboard icons (Note: not numbers, just characters)
    $19 c,  $1A c,  $1B c,     0 c,                             \ 1 2 3 _
    $1C c,  $1D c,  $1E c,  $18 c,                              \ 4 5 6 0
    $1F c,  $20 c,  $21 c,  $2F c,                              \ 7 8 9 ?
    $29 c,  $28 c,  $2A c,  $2C c,                              \ : \  ! @
    $26 c,  $22 c,  $25 c,  $2E c,                              \ z j . ,W
    $2D c,  $27 c,  $2B c,  $23 c,                              \ * / + -

LABEL decimalKeyboard                                           \ the decimal number entry keyboard icons
    $19 c,  $1A c,  $1B c,     0 c,                             \ 1 2 3 _
    $1C c,  $1D c,  $1E c,  $18 c,                              \ 4 5 6 0
    $1F c,  $20 c,  $21 c,     0 c,                             \ 7 8 9 _
       0 c,     0 c,     0 c,     0 c,                          \ _ _ _ _
       0 c,     0 c,     0 c,     0 c,                          \ _ _ _ _
       0 c,     0 c,     0 c,     0 c,                          \ _ _ _ _

LABEL hexadecimalKeyboard                                       \ the hexadecimal number entry keyboard icons
    $19 c,  $1A c,  $1B c,     0 c,                             \ 1 2 3 _
    $1C c,  $1D c,  $1E c,  $18 c,                              \ 4 5 6 0
    $1F c,  $20 c,  $21 c,     0 c,                             \ 7 8 9 _
       0 c,  $05 c,  $13 c,  $0A c,                             \ _ a b c
       0 c,  $10 c,  $04 c,  $0E c,                             \ _ d e f
       0 c,     0 c,     0 c,     0 c,                          \ _ _ _ _

\ *****************************************************************************
\ get keyboard keys
\ *****************************************************************************

LABEL letter  
    $04 al cmp  
    .forward js 
    [ currentKeyboardIcons ] edx mov                            
    [ _TOS_ + edx ] al mov                                      
LABEL .forward  
    ret 

LABEL key_map_table  \ map 8042 scan type 1 keycode to colorForth character values
    16 c,  17 c,  18 c,  19 c,   0 c,   0 c,   4 c,   5 c,      \ $10 - $17
     6 c,   7 c,   0 c,   0 c,   0 c,   0 c,  20 c,  21 c,      \ $18 - $1F
    22 c,  23 c,   0 c,   0 c,   8 c,   9 c,  10 c,  11 c,      \ $20 - $27
     0 c,   0 c,   0 c,   0 c,  24 c,  25 c,  26 c,  27 c,      \ $28 - $2F
     0 c,   1 c,  12 c,  13 c,  14 c,  15 c,   0 c,   0 c,      \ $30 - $37 N
     3 c,   2 c,                                                \ $38 - $39 alt space

\ ToDo: add a timeout to the loop
LABEL WaitToReceiveKey                                          \ Wait until there is byte to receive from the keyboard controller
LABEL .back  
    $64 al in                                                   \ On-board controller status read
    1 al test                                                   \ OBF (Output Buffer Full)
    .forward jnz                                                \ exit when bit 0 = 1 the On-board controller has a new character for us
    _TOS_ _TOS_ xor                                             
    pause_ call    \ not ready yet, so let the other task(s) have a turn
    .back jmp                                                   \ jump back and try again
LABEL .forward  
\    call pause_    \  not ready yet, so let the other task(s) have a turn
    ret 

LABEL v_lineOffsetTablePtr                                      
    0          ,   \ times 16 dd 0

LABEL lineOffsetZero  
    $00 dword [ v_lineOffset ] mov                              
    ret 

LABEL lineOffsetPlus  
    $0C dword [ v_lineOffset ] add                              
    ret 

LABEL lineOffsetMinus  
    $0C dword [ v_lineOffset ] sub                              
    .forward jns 
    lineOffsetZero call                                         
LABEL .forward  
    ret 

LABEL help  
    $000000FF dword [ v_blk ] cmp                               
    .forward2 je 
    $000000FF dword [ v_blk ] mov                               
    .forward jmp 
LABEL .forward2  
    ( START_BLOCK_NUMBER + 1 ) dword [ v_blk ] mov              
LABEL .forward  
    ret 

LABEL e_plus  
\    call abort0
\    call c_     \  clear the stack
    colourBlindModeToggle call                                  \ ToDo: ...
    abort_e jmp 

LABEL abort_e  
    \ call abort
    c_ call 
LABEL abort_e2  
    RETURN_STACK_0 esp mov                                      
    e_ call 
    ret 

LABEL executeToken  \ ( Fkey -- )
    [ v_cad ] _TOS_ mov                                         
    1 _TOS_ sub                                                 \ step to before the token before the cursor
    2 _TOS_ shl                                                 \ convert cell address to byte address
    [ _TOS_ ] _TOS_ mov                                         
    _TOS_ _SCRATCH_ mov                                         
    $0F _SCRATCH_ and                                           \ check the token type = 3 == red
    $03 _SCRATCH_ cmp                                           
    .forward je 
    $0C _SCRATCH_ cmp                                           \ check the token type = 12 == magenta.  NOT WORKING YET ToDo: fix this
    .forward je 
    .forward2 jmp 
LABEL .forward  
    execute call 
LABEL .forward2  
    _DROP_ 
    ret 

59 constant FirstFkey                                           \ F1 = 59

LABEL FkeyTable  \ ( c -- a )   \ function key action table
\    dd nul              \  57
\    dd nul              \  58
    help       ,                                                \ 59 F1
    toggleBase0 ,                                               \ 60 F2 decimal/hex number display
    seeb       ,                                                \ 61 F3 show/hide blue words
    e_plus     ,                                                \ 62 F4 editor
    nul        ,                                                \ 63 F5
    nul        ,                                                \ 64 F6
    nul        ,                                                \ 65 F7
    nul        ,                                                \ 66 F8
    toggleBase ,                                                \ 67 F9
    c_         ,                                                \ 68 F10
    nul        ,                                                \ 69 Num Lock
    nul        ,                                                \ 70
    cursorHome ,                                                \ 71 Home
    cursorUp   ,                                                \ 72 Up arrow
    nextBlock  ,                                                \ 73 PgUp
    nul        ,                                                \ 74 -
    cursorLeft ,                                                \ 75 Left arrow
    otherBlock ,                                                \ 76
    cursorRight ,                                               \ 77 Right arrow
    nul        ,                                                \ 78 +
    cursorEnd  ,                                                \ 79 End
    cursorDown ,                                                \ 80 Down arrow
    previousBlock ,                                             \ 81 PgDn
    destack    ,                                                \ 82 Insert
    del        ,                                                \ 83 Delete
    nul        ,                                                \ 84
    nul        ,                                                \ 85
    nul        ,                                                \ 86
    toggleBase0 ,                                               \ 87 F11
    nul        ,                                                \ 88 F12
    executeToken ,                                              \ 89 really 121 Enter
    abort_e    ,                                                \ 90 really 123 Escape

LABEL processFkey  \ ( n -- )   \ process the given function key code
\    cmp _TOS_, 121
\    jne .forward1
\    sub _TOS_, ( 121 - 89 )
\  forward1:
    FirstFkey _TOS_ sub                                         \ convert Fn key value to index from 0
    $1F _TOS_ and  
    dword [ ( _TOS_ * 4 ) + FkeyTable ] call                    
\    _DROP_
    \ call e_
    ret 

LABEL get_key  \ ( -- c )   \ waits for and returns a character from the keyboard, assumes Scan Code Set 1, set up by the BIOS
    _DUP_ 
    _TOS_ _TOS_ xor                                             
LABEL .back  
    \ check if the key is a function key
    FirstFkey _TOS_ cmp                                         \ F1 key
    .forward4 js 
    FirstFkey + 32 _TOS_ cmp                                    \ Fxx key + 1
    .forward4 jns 
  \    \  jmp dword [ _TOS_ * 4 + qwertyActionTable - $200 ]
 \   xor dword [ current], ((setDecimalMode - $$) ^ (setHexMode - $$))
 \   xor byte [ numb0 + 18 ], ( $21 ^ 0x0E ) \   0x21 = '9' , 0x0E = 'f' toggle '9' and 'f' on keypad display line
 \   call [ current ]
 \  call toggleBase
    processFkey call                                            
LABEL .forward4  
    _DROP_ 
    get_qwerty_key call                                         
\    call WaitToReceiveKey  \  Wait until there is a byte to receive from the keyboard controller
\    in   al, $60       \  read the key value from the Keyboard data port
    [ v_scanCode ] al mov                                       
\    test al, $F0       \  we are only interested in certain key codes (?)
\    jz   .back
    $3A al cmp                                                  \ exclude keycodes greater than $39,  cmp  is like  sub  but only affects the flags
    .back jnc 
    [ key_map_table - $10 + EAX ] al mov                        \ convert to the colorForth value using the 'key_map_table' table
    ret 

\ *****************************************************************************
\ get qwerty keys
\ *****************************************************************************

    db 0 4 align \ fill the gap with 0's

LABEL qwerty_key_map_table                                      
\         0     1     2     3     4     5     6     7     8     9     A     B     C     D     E     F
    $0B c,  $18 c,  $02 c,  $19 c,  $03 c,  $1A c,  $04 c,  $1B c,  $05 c,  $1C c,  $06 c,  $1D c,  $07 c,  $1E c,  $08 c,  $1F c,  \ 0x00
    $09 c,  $20 c,  $0A c,  $21 c,  $1e c,  $05 c,  $30 c,  $13 c,  $2E c,  $0A c,  $20 c,  $10 c,  $12 c,  $04 c,  $21 c,  $0E c,  \ 0x10
    $22 c,  $0D c,  $23 c,  $14 c,  $17 c,  $07 c,  $24 c,  $22 c,  $25 c,  $24 c,  $26 c,  $0C c,  $32 c,  $09 c,  $31 c,  $06 c,  \ 0x20
    $18 c,  $03 c,  $19 c,  $12 c,  $10 c,  $17 c,  $13 c,  $01 c,  $1F c,  $08 c,  $14 c,  $02 c,  $16 c,  $16 c,  $2F c,  $11 c,  \ 0x30
    $11 c,  $0F c,  $2D c,  $15 c,  $15 c,  $0B c,  $2C c,  $26 c,  $0C c,  $23 c,  $34 c,  $25 c,  $35 c,  $27 c,  $27 c,  $28 c,  \ 0x40
    $28 c,  $29 c,  $82 c,  $2A c,  $8D c,  $2B c,  $83 c,  $2C c,  $89 c,  $2D c,  $33 c,  $2E c,  $B5 c,  $2F c,  $39 c,  $80 c,  \ 0x50
    $1C c,  $81 c,  $0E c,  $82 c,  $01 c,  $83 c,  $3B c,  $84 c,  $29 c,  $30 c,  

LABEL get_qwerty_key                                            \ get a qwerty key character
    _DUP_ 
LABEL .back  
    WaitToReceiveKey call                                       
    $60 al in  

    $1C _TOS_ cmp                                               \ the Enter key scan code
    .forward1 jne 
    \ add _TOS_, ( 89 - $1C ) \  convert the code for the Enter key to 89
    89 _TOS_ mov 
LABEL .forward1  

    $81 _TOS_ cmp                                               \ the Escape key scan code
    .forward2 jne 
    ( 90 - $81 ) _TOS_ add                                      \ convert the code for the Escp key to 90
LABEL .forward2  

    al [ v_scanCode ] mov                                       
    _TOS_ ecx mov                                               \ copy keycode into cl
    $7F cl and                                                  \ filter out key-up bit 7
    $2A cl cmp                                                  \ g?
    .got_c_or_g jz 
    $36 cl cmp                                                  \ c?
    .not_c_or_g jnz 
LABEL .got_c_or_g  
    $80 al and                                                  \ extract key-up bit
    $80 al xor                                                  \ complement it
    _TOS_ [ v_qwerty_key ] mov                                  
    short .back jmp 
LABEL .not_c_or_g  
    al al or                                                    \ check if key-up
    .back js                                                    \ if so, try again to get keydown event
    $7F al and                                                  \ filter out key-up bit
    [ v_qwerty_key ] _TOS_ or                                   
    qwerty_key_map_table edx mov                                
    $35 ecx mov  
LABEL .back2  
    al [edx] cmp 
    .forward jz 
    byte $02 edx add                                            
    .back2 loop 
    _TOS_ _TOS_ xor                                             
    ret 
LABEL .forward  
    [edx+$01] al mov                                            
    qwerty_key_map_table edx sub                                
    1 edx shr 
    edx [ v_digin ] mov                                         
    59 _TOS_ cmp   \ F1 key
\    jnz .forward4
\    \  jmp dword [ _TOS_ * 4 + qwertyActionTable - $200 ]
\    xor dword [ current], ((setDecimalMode - $$) ^ (setHexMode - $$))
\    call toggleBase
\  forward4:
    ret 

\ *****************************************************************************
\ keyboard jump tables
\ *****************************************************************************

LABEL graph0  
    nul0 ,   nul0 ,   nul0 ,   alph0 ,                          
    $00 c,  $00 c,  $05 c,  $00 c,                              \ _ _ a _

LABEL graph1  
    word0 ,   x ,   lj ,   alph ,                               
    $15 c,  $25 c,  $05 c,  $00 c,                              \ x . a _

LABEL alpha0  
    nul0 ,   nul0 ,   number ,   star0 ,                        
    $00 c,  $21 c,  $2D c,  $00 c,                              \ _ 9 * _

LABEL alpha1  
    word0 ,   x ,   lj ,   graph ,                              
    $15 c,  $25 c,  $2D c,  $00 c,                              \ x . * _

LABEL numb0  
    nul0 ,   minus ,   alphn ,   toggleBase ,                   
    $23 c,  $05 c,  $0E c,  $00 c,                              \ - a f _

LABEL numb1  
    number0 ,   xn ,   endn ,   number0 ,                       
    $15 c,  $25 c,  $00 c,  $00 c,                              \ x . _ _

\ *****************************************************************************
\ Shannon-Fano compression
\ *****************************************************************************

LABEL bits_  
    $1C        c,  

LABEL lj0  
    [ bits_ ] cl mov                                            
    $04 cl add  
    cl dword 0 [ESI] shl                                        
    ret 

LABEL lj  
    lj0 call 
    _DROP_ 
    ret 

LABEL full  
    lj0 call 
    dword [ v_words ] inc                                       
    $1C byte [ bits_ ] mov                                      
    ch [ bits_ ] sub                                            
    edx _TOS_ mov 
    _DUP_ 
    ret 

LABEL pack0  
    byte  $50 _TOS_ add                                         
    $07 cl mov  
    short pack1 jmp 

LABEL pack  
    $10 al cmp  
    pack0 jnc 
    $04 cl mov  
    $08 al test  
    pack1 jz 
    ecx inc 
    $18 al xor  
LABEL pack1  
    _TOS_ edx mov 
    cl ch mov 
LABEL .back  
    cl [ bits_ ] cmp                                            
    .forward jnc 
    1 al shr 
    full jc 
    cl dec 
    short .back jmp 
LABEL .forward  
    cl dword 0 [ESI] shl                                        
    _TOS_ 0 [ESI] xor                                           
    cl [ bits_ ] sub                                            
    ret 

LABEL x  
    right call 
    [ v_words ] _TOS_ mov                                       
    [esi+_TOS_*4] esi lea                                       
    _DROP_ 
    accept jmp 

LABEL word_  
    right call 
    $01 dword [ v_words ] mov                                   
    $01 dword [ chars ] mov                                     
    _DUP_ 
    $00 dword 0 [ESI] mov                                       
    $1C byte [ bits_ ] mov                                      
LABEL word1  
    letter call 
    .forward jns 
    [ shiftAction ] edx mov                                     
    dword [edx+_TOS_*4] jmp                                     
LABEL .forward  
    al al test 
    word0 jz 
    _DUP_ 
    echo_ call 
    pack call 
    dword [ chars ] inc                                         
LABEL word0  
    _DROP_ 
    get_key call 
    short word1 jmp 

\ *****************************************************************************
\ number display
\ *****************************************************************************

LABEL digitTable  
    14 c,  10 c,   0 c,   0 c,                                  
     0 c,   0 c,  12 c,   0 c,   0 c,   0 c,  15 c,   0 c,      
    13 c,   0 c,   0 c,  11 c,   0 c,   0 c,   0 c,   0 c,      
     0 c,   1 c,   2 c,   3 c,   4 c,   5 c,   6 c,   7 c,      
     8 c,   9 c,  

LABEL v_sign  
    $00        c,  

LABEL minus  
    al [v_sign ] mov                                            
    short number2 jmp                                           

LABEL number0  
    _DROP_ 
    short number3 jmp                                           

LABEL number  
    [ current ] call                                            
    $00 byte [ v_sign ] mov                                     
    _TOS_ _TOS_ xor                                             
LABEL number3  
    get_key call 
    letter call 
    .forward jns 
    [ shiftAction ] edx mov                                     
    dword [edx+_TOS_*4] jmp                                     

LABEL .forward  
    al al test 
    number0 jz 
    [ _TOS_ + digitTable - 4 ] al mov                           
    $1F byte [ v_sign ] test                                    
    .forward2 jz 
    _TOS_ neg 
LABEL .forward2  
    0 [ESI] edx mov                                             
    [ base ] edx imul                                           
    _TOS_ edx add 
    edx 0 [ESI] mov                                             
LABEL number2  
    _DROP_ 
    numb1 dword [ shiftAction ] mov                             
    short number3 jmp                                           

LABEL endn  
    _DROP_ 
    [ anumber] call 
    accept jmp 

LABEL setDecimalMode  
    $0A dword [ base ] mov                                      
    numb0 dword [ shiftAction ] mov                             
    ( decimalKeyboard - 4 ) dword [ currentKeyboardIcons] mov   
    ret 

LABEL setHexMode  
    $10 dword [ base ] mov                                      
    numb0 dword [ shiftAction ] mov                             
    ( hexadecimalKeyboard - 4 ) dword [ currentKeyboardIcons] mov 
    ret 

LABEL toggleBase0  
    ((setDecimalMode - $$) ^ (setHexMode - $$)) dword [ current] xor 
    ( $21 ^ $0E ) byte [ numb0 + 18 ] xor                       \  0x21 = '9' , 0x0E = 'f' toggle '9' and 'f' on keypad display line
    [ current ] call                                            
    ret 

LABEL toggleBase  
    toggleBase0 call                                            
    dword number0 jmp                                           

\ *****************************************************************************
\ text entry
\ *****************************************************************************

LABEL xn  
    _DROP_ 
    _DROP_ 
    accept jmp 

LABEL nul0  
    _DROP_ 
    short accept2 jmp                                           

LABEL clearHintChar  
    _TOS_ push 
    _TOS_ _TOS_ xor                                             
    $00 byte [ v_hintChar ] mov                                 \ clear the hint character
    _TOS_ pop 
    ret 

LABEL accept  \ get a word from keyboard
    alpha0 dword [ shiftAction ] mov                            
    [ alphaKeyboard - 4] edi lea                                
LABEL accept1  
    edi [ dword currentKeyboardIcons ] mov                      
LABEL accept2  
    $FFFFFFFF dword [ x_qwerty ] test                           
    .forward jz 
    dword [ x_qwerty ] jmp                                      \ jump to the address in x_qwerty if it is non-zero
LABEL .forward  
    get_key call                                                \ calls pause_ while waiting for a character
    $04 al cmp  
    .forward2 jns 
    [ shiftAction ] edx mov                                     
    dword [ edx + _TOS_ * 4 ] jmp                               
LABEL .forward2  
    byte +$14 dword [ shiftAction ] add                         
    word_ call 
    [ aword ] call 
    short accept jmp                                            \ endless loop

LABEL alphn  
    _DROP_ 

LABEL alph0  
    alpha0 dword [ shiftAction ] mov                            
    [ alphaKeyboard - 4 ] edi lea                               
    short Xstar0 jmp                                            

LABEL star0  
    graph0 dword [ shiftAction ] mov                            
    [ ( graphicsKeyboard - 4 ) ] edi lea                        
LABEL Xstar0  
    _DROP_ 
    short accept1 jmp                                           

LABEL alph  
    alpha1 dword [ shiftAction ] mov                            
    [ alphaKeyboard - 4] edi lea                                
    short Xgraph jmp                                            

LABEL graph  
    graph1 dword [ shiftAction ] mov                            
    [ ( graphicsKeyboard - 4 ) ] edi lea                        

LABEL Xgraph  
    edi [ currentKeyboardIcons ] mov                            
    dword word0 jmp 

\ *****************************************************************************
\ Shannon-Fano decompression and display
\ *****************************************************************************

LABEL unpack  \ ( token -- token' nextCharacter )
    _DUP_   \ copy TOS to our data stack SOS
    _TOS_ _TOS_ test                                            
    .forward js 
    $04 dword 0 [ESI] shl                                       
    $04 _TOS_ rol  
    byte $07 _TOS_ and                                          
    ret 
LABEL .forward  
    1 _TOS_ shl 
    .forward2 js 
    $05 dword 0 [ESI] shl                                       
    $04 _TOS_ rol  
    byte $07 _TOS_ and                                          
    $08 al xor  
    ret 
LABEL .forward2  
    $07 dword 0 [ESI] shl                                       
    $06 _TOS_ rol  
    byte $3F _TOS_ and                                          
    $10 al sub  
    ret 

LABEL qring  \ ( a cursor -- a' )  edx  contains pointer to current address to display
    _DUP_ 
    dword 0 [ESI] inc                                           
    edi [ v_curs ] cmp                                          
    .forward jnz                                                \ address to display = cursor address?
    _TOS_ [ v_curs ] mov                                        \ yes,
LABEL .forward  
    [ v_curs ] _TOS_ cmp                                        \ no
    .forward2 jz 
    .forward4 jns                                               \ time to draw the cursor?
    edi [ v_pcad ] mov                                          \ no, so exit
LABEL .forward4  
\     _DUP_
\     mov _TOS_, $0F
\     call doColourBlind             \  display the final colourblind punctuation, set up for next call of plusList
    _DROP_ 
    ret                                                         \ exit here
LABEL .forward2  
    edi [ v_cad ] mov                                           
    _SCRATCH_ push 
    [ v_10000_iconw ] _SCRATCH_ mov                             
    _SCRATCH_ dword [ v_xy ] sub                                \ move one icon's worth of horizontal pixels to the left
    _DUP_ 
    [ v_foregroundColour ] _SCRATCH_ mov                        \ save the current colour
    colour_PacMan _TOS_ mov                                     
    color call 
    $30 _TOS_ mov                                               \ display the "PacMan" cursor
    [ v_x ] cx mov 
    [ v_rightMargin ] cx cmp                                    
    .forward5 js 
    emit_ call 
    [ v_10000_iconw ] _SCRATCH_ mov                             
    _SCRATCH_ dword [ v_xy ] sub                                \ move one icon's worth of horizontal pixels to the left
    .forward6 jmp 
LABEL .forward5  
    emit_ call 
LABEL .forward6  
    _SCRATCH_ dword [ v_foregroundColour ] mov                  \ restore the current colour
    _SCRATCH_ pop 
    ret 

\ *****************************************************************************
\ Conventional Forth display (does not require colours)
\ *****************************************************************************

LABEL currentState  
    0          ,   

LABEL lastState  
    0          ,   

LABEL txt0  
    white call 
    $6D EMIT_IMM 
    space_ call 
    ret 

LABEL txt1  
    white call 
    $6E EMIT_IMM 
    space_ call 
    ret 

LABEL imm0  
    yellow call 
    $58 EMIT_IMM 
    space_ call 
    ret 

LABEL imm1  
    yellow call 
    $59 EMIT_IMM 
    space_ call 
    ret 

LABEL mvar0  
    yellow call 
    $58 EMIT_IMM  \ '['
    space_ call 
    $09 EMIT_IMM  \ 'm'
    $11 EMIT_IMM  \ 'v'
    $05 EMIT_IMM  \ 'a'
    $01 EMIT_IMM  \ 'r'
    space_ call 
    ret 

LABEL mvar1  
    yellow call 
    $59 EMIT_IMM  \ ']'
    space_ call 
    ret 

\ unfortunately we need to display the ':' after the CR, so must do this in  redWord , not here
\ colon0:
\     call red
\     EMIT_IMM $59
\     call space_
\     ret
\ 
\     dd nul, imm0, nul, colon0, nul, nul, nul, nul, nul, txt0, nul, nul, mvar0, nul, nul, nul

LABEL txts  
    0 c,  1 c,  1 c,  3 c,  4 c,  5 c,  6 c,  7 c,  1 c,  9 c,  9 c,  9 c,  12 c,  13 c,  14 c,  15 c,  

LABEL tx  \ ( c -- c )   \ return the value in the given offset in  txts
    $FF _TOS_ and  
    [ _TOS_ + txts ] _TOS_ mov                                  
    $FF _TOS_ and  
    ret 

LABEL newActions  
    nul ,   imm0 ,   nul ,   nul ,   nul ,   nul ,   nul ,   nul ,   nul ,   txt0 ,   nul ,   nul ,   mvar0 ,   nul ,   nul ,   nul ,   

LABEL dotNew  \ ( state -- )
    [ ( _TOS_ * 4 ) + newActions ] call                         
    ret 

LABEL oldActions  
    nul ,   imm1 ,   nul ,   nul ,   nul ,   nul ,   nul ,   nul ,   nul ,   txt1 ,   nul ,   nul ,   mvar1 ,   nul ,   nul ,   nul ,   

LABEL dotOld  \ ( state -- )
    [ ( _TOS_ * 4 ) + oldActions ] call                         
    ret 

LABEL colourBlindAction   constant                              \ ( state -- state )    \ perform the required actionon change of state
    _SCRATCH_ push 
    _DUP_ 
    tx call 
    $00 _TOS_ cmp  
    .end jz                                                     \ no action on extension tokens, value 0
    [ currentState ] _SCRATCH_ mov                              
    _TOS_ [ currentState ] mov                                  
    [ currentState ] _SCRATCH_ cmp                              \ compare the new state on TOS to the last one saved in currentState
    .end jz                                                     \ exit if there has been no change of state
    _DUP_ 
    _SCRATCH_ _TOS_ mov                                         
    dotOld call                                                 \
    [ currentState ] _TOS_ mov                                  
    dotNew call 
    _DROP_ 
    $0000 byte [ currentState ] cmp                             
    .end jz 
    [ currentState ] _SCRATCH_ mov                              
    _SCRATCH_ [ lastState ] mov                                 
LABEL  .end  
    _DROP_ 
    _SCRATCH_ pop 
    ret 

\ \ Block 70
\ ( Colourblind Editor Display )
\ #1 MagentaV currentState $01 MagentaV lastState
\ : +txt white $6D emit space \ 
\ : -txt white $6E emit space \ 
\ : +imm yellow $58 emit space \ 
\ : -imm yellow $59 emit space \ 
\ : +mvar yellow $09 emit $11 emit $05 emit $01 emit space \ 
\ : txts string $03010100 , $07060504 , $09090901 , $0F0E0D0C , ( \  )
\ : tx ( c-c ) $0F and txts + 1@ $0F and \ 
\ : .new currentState @ $0F and jump nul +imm nul nul nul nul nul nul nul +txt nul nul +mvar nul nul nul \ 
\ : .old lastState @ $0F and jump nul -imm nul nul nul nul nul nul nul -txt nul nul nul nul nul nul \ 
\ here
\ : cb ( n-n ) #0 + 0if \  then tx
\    currentState @ swap dup currentState ! - drop if .old .new
\    currentState @ #0 + if dup lastState ! then then \ 
\ : cbs ( -- here ) #0 + $00 + cblind ! \ 

\ colourBlind:    \  ( state -- state )    \ vectored colorForth to display colourBlind extra characters ( e.g. ':' for red words )
\    call dword [ x_colourBlind ]
\    ret

\ *****************************************************************************

LABEL lowercase  \ display a white text word in normal lower-case letters
    white call 
LABEL type_  \ ( -- )   \ display a Shanon-Fano encoded word pointed to by  edi  in the current colour
    _DUP_ 
    [ ( edi * 4 ) - $04 ] _TOS_ mov                             
LABEL showShannonFano                                           \ ( token -- ) \ display the given Shanon-Fano encoded word in the current colour
    byte -$10 _TOS_ and                                         
LABEL lowercasePrimitive                                        \ ( token -- ) \ display the given Shanon-Fano encoded word in the current colour
    unpack call 
    lowercasePrimitiveEnd jz                                    
    emit_ call 
    lowercasePrimitive jmp                                      
LABEL lowercasePrimitiveEnd                                     
    space_ call 
    _DROP_ 
    _DROP_ 
    ret 

LABEL typeNumber32tok                                           \ ( token -- ) \ display the given Shanon-Fano encoded word as a number in the current colour
    _DROP_ \ call dotHex8
    $FFFFFFF dword [ lastTokenWasLiteral ] mov                  
    ret 

LABEL typeNumber32  \ ( token -- ) \ display the given Shanon-Fano encoded word as a hex number in the current colour
    dotHex8 call 
    $00000000 dword [ lastTokenWasLiteral ] mov                 
    ret 

LABEL typeNumber27  \ ( token -- ) \ display the given Shanon-Fano encoded word as a 27 bit hex number in the current colour
    5 _TOS_ shr 
    dotHex call 
    ret 

LABEL lastTokenWasLiteral                                       
    $00        ,   

LABEL lastShannonFanoToken                                      
    $00        ,   

LABEL magentaPrimitive                                          \ ( token -- )
    lowercasePrimitive call                                     
    $FFFFFFF dword [ lastTokenWasLiteral ] mov                  
    ret 

LABEL displayOneShannonFanoActions                              \    * = number
    lowercasePrimitive ,                                        \ 0     extension token, remove space from previous word, do not change the colour
    lowercasePrimitive ,                                        \ 1     yellow "immediate" word
    typeNumber32tok ,                                           \ 2  *  yellow "immediate" 32 bit number in the following pre-parsed cell
    lowercasePrimitive ,                                        \ 3     red forth wordlist "colon" word
    lowercasePrimitive ,                                        \ 4     green compiled word
    typeNumber32tok ,                                           \ 5  *  green compiled 32 bit number in the following pre-parsed cell
    typeNumber27 ,                                              \ 6  *  green compiled 27 bit number in the high bits of the token
    lowercasePrimitive ,                                        \ 7     cyan macro wordlist "colon" word
    typeNumber27 ,                                              \ 8  *  yellow "immediate" 27 bit number in the high bits of the token
    lowercasePrimitive ,                                        \ 9     white lower-case comment
    camelcasePrimitive ,                                        \ A     first letter capital comment
    uppercasePrimitive ,                                        \ B     white upper-case comment
    magentaPrimitive ,                                          \ C     magenta variable
    lowercasePrimitive ,                                        \ D
    lowercasePrimitive ,                                        \ E     editor formatting commands
    lowercasePrimitive ,                                        \ F

LABEL dotsf  \ ( token -- )   \ display the given Shannon-Fano encoded word in the token's colour
    edi push 
    _TOS_ edx mov  
    $FFFFFFF0 _TOS_ and                                         
    _DUP_ 
    [ lastTokenWasLiteral ] edi mov                             
    $00000000 edi test                                          
    .forward3 jz 
    0 edx mov 
LABEL .forward3  
    byte $0F edx and                                            
    .forward jnz                                                \ do not change the colour if this is an extension token
    [ lastShannonFanoToken ] edx mov                            
    forward2 jmp 
LABEL     .forward  
    edx [ lastShannonFanoToken ] mov                            
LABEL forward2  
    _TOS_ push 
    [ ( edx * 4 ) + actionColourTable ] _TOS_ mov               
    color call 
    _TOS_ pop 
    [ ( edx * 4 ) + displayOneShannonFanoActions ] call         
    edi pop 
    ret 

LABEL redWord  \ display a red word
    [ v_x ] cx mov 
    [ v_leftMargin ] cx cmp                                     
    .forward jz                                                 \ do not do a  cr  if we are already at the left margin
    [ v_not_cr ] cl mov                                         
    0 cl cmp 
    .forward jnz                                                \ do not do a  cr  if it has been disabled by a blue  -cr  token
    cr_ call 
LABEL .forward  
    0 byte [ v_not_cr ] mov                                     
    setRed call 

    $00 byte [ v_colourBlindMode ] cmp                          
    .forward2 jz 
    $01 byte [ v_blk ] test                                     \ do not display colourblind characters in odd numbered shadow blocks
    .forward2 jnz 
    $29 EMIT_IMM                                                \ emit a ':' if in colourblind mode
    space_ call 
LABEL .forward2  
    type_ jmp 

LABEL greenWord  \ display a green word
    setGreen call 
    type_ jmp 

LABEL cyanWord  \ display a cyan word
    setCyan call 
    type_ jmp 

LABEL yellowWord  \ display a yellow word
    yellow call 
    type_ jmp 

LABEL camelcase  \ display a white word with the first letter Capitalised
    white call 
    _DUP_ 
    [ ( edi * 4 ) - $04 ] _TOS_ mov                             
    byte -$10 _TOS_ and                                         
LABEL camelcasePrimitive  
    unpack call 
    $30 al add                                                  \ make the first character upper case
    emit_ call                                                  \ display it
    lowercasePrimitive jmp                                      \ display the rest of the word

LABEL uppercase  \ display a white word with all letters CAPITALISED
    white call 
    _DUP_ 
    [ ( edi * 4 ) - $04 ] _TOS_ mov                             
    byte -$10 _TOS_ and                                         
LABEL uppercasePrimitive  
    unpack call 
    lowercasePrimitiveEnd jz                                    
    $30 al add  
    emit_ call 
    uppercasePrimitive jmp                                      

LABEL extension  \ display an extension token, do not change the colour
    [ v_10000_iconw ] _SCRATCH_ mov                             
    _SCRATCH_ dword [ v_xy ] sub                                \ move iconw horizontal pixels back, to remove the space at the end of the last word
    $FFFFFFF0 dword [ ( edi * 4 ) - $04 ] test                  
    type_ jnz 
    edi dec 
    edi [ v_lcad ] mov                                          
    space_ call 
    qring call 
    edx pop                                                     \ EXIT from calling word
    _DROP_                                                      \ the ret below will return to the word that called  extension
    ret                                                         \ so it looks like it never happened

LABEL greenShortNumber                                          \ display the green compiled 27 bit number in the high bits of the token
    [ ( edi * 4 ) - $04 ] edx mov                               
    $05 edx sar  
    short greenNumber1 jmp                                      

LABEL magentaVariable                                           \ display a magenta variable using the 32 bit number in the following pre-parsed cell
    dotDecimal dword [ x_numberDisplay ] mov                    
    setMagenta call 
    type_ call                                                  \ display the name of the variable
LABEL greenNumber                                               \ display the value of a hexadecimal/decimal number in green
    [ ( edi * 4 ) + $00 ] edx mov                               \ load the value of the variable from the pre-parsed source
    edi inc                                                     \ step over the variable value in the pre-parsed source
LABEL greenNumber1  
    green call 
    short displayNumber jmp                                     

LABEL yellowShortNumber  
    [ ( edi * 4 ) - $04 ] edx mov                               \ load the value of the number from the current token in the pre-parsed source
    $05 edx sar                                                 \ remove the token colour bits
    short yellowNumber1 jmp                                     

LABEL yellowNumber  \ ( -- ) display a number word, constant value following in the pre-parsed source
    [ ( edi * 4 ) + $00 ] edx mov                               \ load the value of the number from the pre-parsed source
    edi inc                                                     \ step over the number value in the pre-parsed source
LABEL yellowNumber1  \ ( -- ) display a yellow number word
    yellow call 
LABEL displayNumber  \ ( rgb -- )   display the number in e dx  with the given colour, using the base implied in  x_numberDisplay
    _DUP_ 
    edx _TOS_ mov 
    \ jmp qdot
    dword [ x_numberDisplay ] jmp                               

\ *****************************************************************************
\ Blue words - formatting the editor display
\ *****************************************************************************

LABEL get_x  \ ( -- c )  \ return the current x character position
    edx push 
    _DUP_ 
    _TOS_ _TOS_ xor                                             
    word [ v_x ] ax mov                                         
    edx edx xor                                                 \ clear high 32 bits of dividend
    dword [ v_iconw ] div                                       \ EDX:EAX divided by the icon width , EAX now contains the current character position, EDX the remainder
    edx pop 
    ret 

LABEL set_x  \ ( c -- )  \ set the current x character position
    edx push 
    edx edx xor 
    dword [ v_iconw ] mul                                       
    ax word [ v_x ] mov                                         
    edx pop 
    _DROP_ 
    ret 

24 constant TAB_SIZE                                            

LABEL tab  \ ( -- )    \ align to the next n character column
\    _DUP_
    pusha 
    get_x call 
    edx edx xor                                                 \ clear high 32 bits of dividend
    TAB_SIZE _SCRATCH_ mov                                      
    _SCRATCH_ div 
    _SCRATCH_ mul 
\    shr _TOS_, 5
\    shl _TOS_, 5
    TAB_SIZE _TOS_ add                                          
    set_x call 
    popa 
    ret 

\    add _TOS_, 3
\    and _TOS_, $FFFFFFFC
\    add _TOS_, 4

\    mov ecx, 4
\  back:
\    call space_
\    loop .back

LABEL not_cr  
    byte [ v_not_cr ] not                                       
    ret 

LABEL blueWord  \ ( -- )    \ format the editor display screen using certain blue tokens
    _DUP_ 
    [ v_seeb ] al mov                                           
    0 al cmp 
    .forward jz 
    setBlue call 
    type_ call 
LABEL .forward  
    [ ( edi * 4 ) - $04 ] _TOS_ mov                             
    $9080000E _TOS_ cmp                                         \ cr
    .skip1 jnz 
    cr_ call 
LABEL .skip1  
    $E64B8C0E _TOS_ cmp                                         \ -tab
    .skip2 jnz 
    not_cr call 
    tab call 
LABEL .skip2  
    $25C6000E _TOS_ cmp                                         \ tab
    .skip3 jnz 
    tab call 
LABEL .skip3  
    $C620000E _TOS_ cmp                                         \ br
    .skip4 jnz 
    cr_ call 
    cr_ call 
LABEL .skip4  
    $E721000E _TOS_ cmp                                         \ -cr
    .skip5 jnz 
    not_cr call 
LABEL .skip5  
    $90FB000E _TOS_ cmp                                         \ cr+   cr and 3 spaces
    .skip6 jnz 
    cr_ call 
    space_ call 
    space_ call 
    space_ call 
LABEL .skip6  
    $F0000000 _TOS_ and                                         
    $E0000000 _TOS_ cmp                                         \ . .. ...
    .skip7 jnz 
    space_ call 
LABEL .skip7  
    _DROP_ 
    ret 

\ silverWord:   \  ( -- )    ; ToDo: document this
\     mov edx, [ ( edi * 4 ) - $04 ] \  load the value of the action from the current token in the pre-parsed source
\     sar edx, $05                   \  remove the token colour bits
\     _DUP_
\     mov _TOS_, colour_white
\     cmp dword [ x_numberDisplay ], dotDecimal
\     jz .forward
\     mov _TOS_, colour_silver
\ .forward:
\    jmp short displayNumber
\    ret

LABEL silverWord  \ display a silver word
    setSilver call 
    type_ jmp 

LABEL displayShannonFanoActions                                 \    * = number
    extension  ,                                                \ 0     extension token, remove space from previous word, do not change the colour
    yellowWord ,                                                \ 1     yellow "immediate" word
    yellowNumber ,                                              \ 2  *  yellow "immediate" 32 bit number in the following pre-parsed cell
    redWord    ,                                                \ 3     red forth wordlist "colon" word
    greenWord  ,                                                \ 4     green compiled word
    greenNumber ,                                               \ 5  *  green compiled 32 bit number in the following pre-parsed cell
    greenShortNumber ,                                          \ 6  *  green compiled 27 bit number in the high bits of the token
    cyanWord   ,                                                \ 7     cyan macro wordlist "colon" word
    yellowShortNumber ,                                         \ 8  *  yellow "immediate" 27 bit number in the high bits of the token
    lowercase  ,                                                \ 9     white lower-case comment
    camelcase  ,                                                \ A     first letter capital comment
    uppercase  ,                                                \ B     white upper-case comment
    magentaVariable ,                                           \ C     magenta variable
    silverWord ,                                                \ D
    blueWord   ,                                                \ E     editor formatting commands
    nul        ,                                                \ F

LABEL v_lineOffset  
    1          ,   \ the top line of the display

LABEL doColourBlind  \ ( state -- )   \ add conventional Forth punctuation based on the newand last states
    $00 byte [ v_colourBlindMode ] cmp                          
    .forward3 jz 
    $01 byte [ v_blk ] test                                     \ do not display colourblind characters in odd numbered shadow blocks
    .forward3 jnz 
    dword colourBlindAction call                                \ pass the new state to colourBlind so that extra characters can be added to the display
LABEL     .forward3  
    _DROP_ 
    ret 

LABEL plusList  \ ( -- ) display the current colorForth block
    _DUP_ 
    _TOS_ _TOS_ xor                                             
    _TOS_ [ currentState ] mov                                  
    _TOS_ [ lastState ] mov                                     
    _DROP_ 

    setupText_ call                                             \ setup the clip window for this display
    _DUP_ 
    [ v_lcad ] _TOS_ mov                                        
    _TOS_ [ v_cad ] mov                                         
    [ v_blk ] _TOS_ mov                                         \ get the current block number to be edited
    blockToCellAddress call                                     \ add the RELOCATED block number offset and convert to cell address
    _TOS_ edi mov 
    _TOS_ _TOS_ xor                                             
    [ v_lineOffset ] edi add                                    
    edi [ v_pcad ] mov                                          
LABEL .back  
    dword [ ( edi * 4 ) + $00 ] edx mov                         \ edi is the display pointer and is a cell address
    qring call                                                  \ show one Shannon-Fano encoded word pointed to by edi
    edi inc 
    \ adjust the number base according to bit 5 of the token value, only used by number display words
    dotDecimal dword [ x_numberDisplay ] mov                    
    $10 dl test  
    .forward2 jz 
    dotHex dword [ x_numberDisplay ] mov                        
LABEL .forward2  
    byte $0F edx and                                            
    _DUP_ 
    edx _TOS_ mov 
    doColourBlind call                                          
    [ ( edx * 4 ) + displayShannonFanoActions ] call            
    short .back jmp 

LABEL refresh  
    show call 
    page_ call 
    displayBlockNumber call                                     
    plusList call 
    _DUP_ 
    $0F _TOS_ mov  
    doColourBlind call                                          \ display the final colourblind punctuation, set up for next call of plusList
    dword displayTheKeyboard jmp                                

    db 0 4 align \ fill the gap with 0's

LABEL actionColourTable                                         \    * = number
    colour_orange ,                                             \ 0     extension token, remove space from previous word, do not change the colour
    colour_yellow ,                                             \ 1     yellow "immediate" word
    colour_yellow ,                                             \ 2  *  yellow "immediate" 32 bit number in the following pre-parsed cell
    colour_red ,                                                \ 3     red forth wordlist "colon" word
    colour_green ,                                              \ 4     green compiled word
    colour_green ,                                              \ 5  *  green compiled 32 bit number in the following pre-parsed cell
    colour_green ,                                              \ 6  *  green compiled 27 bit number in the high bits of the token
    colour_cyan ,                                               \ 7     cyan macro wordlist "colon" word
    colour_yellow ,                                             \ 8  *  yellow "immediate" 27 bit number in the high bits of the token
    colour_white ,                                              \ 9     white lower-case comment
    colour_white ,                                              \ A     first letter capital comment
    colour_white ,                                              \ B     white upper-case comment
    colour_magenta ,                                            \ C     magenta variable
    colour_silver ,                                             \ D
    colour_blue ,                                               \ E     editor formatting commands
    colour_black ,                                              \ F

LABEL vector  
    0          ,   \ pointer to call table for keypad ( see keypd )

LABEL action  
    1          c,  

    db 0 4 align \ fill the gap with 0's

LABEL cursorLeft  \ ( -- )
    dword [ v_curs ] dec                                        
    cursorRight.forward jns                                     
LABEL cursorRight  
    dword [ v_curs ] inc                                        
LABEL .forward  
    ret 

LABEL cursorUp  \ ( -- )
    byte $10 dword [ v_curs ] sub                               
    .forward jns 
    $00 dword [ v_curs ] mov                                    
LABEL .forward  
    ret 

LABEL cursorDown  \ ( -- )
    _DUP_ 
\    xor _TOS_, _TOS_
\    mov dword [ v_numberOfMagentas ], _TOS_
    $00100 ecx mov                                              \ 256 x 4 byte cells = 1 block
LABEL .loop  
    dword [ v_curs ] inc                                        
    v_curs_to_token call                                        
    $00 _TOS_ cmp  
    .forward jz                                                 \ exit if the token value is 0 == end of block
    $0F _TOS_ and                                               \ look at the token type
    $03 _TOS_ cmp                                               \ red token
    .forward je 
    $0C _TOS_ cmp                                               \ magenta token
    .forward2 jne 
    dword [ v_curs ] dec                                        
\    inc dword [ v_numberOfMagentas ] \  correction for magenta variables
    .forward jmp 
LABEL .forward2  
    _DROP_ 
    .loop loop 
LABEL .forward                                                  \ found red or magenta token, so exit
    _DROP_ 
\    mov _TOS_, dword [ v_numberOfMagentas ]
\    sub dword [ v_curs ], _TOS_  \  the correction for magenta variables
    dword [ v_curs ] inc                                        
    _DROP_ 
    ret 

LABEL cursorEnd  \ ( -- )
    _DUP_ 
    _TOS_ _TOS_ xor                                             
    _TOS_ dword [ v_numberOfMagentas ] mov                      
    $00100 ecx mov                                              \ 256 x 4 byte cells = 1 block
LABEL .loop  
    dword [ v_curs ] inc                                        
    v_curs_to_token call                                        
    $00 _TOS_ cmp  
    .forward jz                                                 \ exit if the token value is 0 == end of block
    $0F _TOS_ and                                               \ look at the token type
\    cmp _TOS_, $03     \  red token
\    je .forward
    $0C _TOS_ cmp                                               \ magenta token
    .forward2 jne 
    dword [ v_numberOfMagentas ] inc                            \ correction for magenta variables
\    jmp .forward
LABEL .forward2  
    _DROP_ 
    .loop loop 
LABEL .forward                                                  \ found red or magenta token, so exit
    _DROP_ 
\    mov _TOS_, dword [ v_numberOfMagentas ]
\    sub dword [ v_curs ], _TOS_  \  the correction for magenta variables
\    dec dword [ v_curs ]
    _DROP_ 
    ret 

LABEL cursorHome  \ ( -- )
    _SCRATCH_ _SCRATCH_ xor                                     
    _SCRATCH_ dword [ v_numberOfMagentas ] mov                  
    _SCRATCH_ dword [ v_curs ] mov                              \ the graphics cursor for drawing the block
    _SCRATCH_ dword [ v_lineOffset ] mov                        \ the cursor position to start drawing the block
    _SCRATCH_ dword [ v_lineOffsetTablePtr ] mov                \ a pointer to the cursor for each line in the display
    _SCRATCH_ dword [ v_numberOfMagentas ] mov                  \ count of Magentat variable displyed so far in the edited block
    ret 

LABEL nextBlock  \ ( -- )
    byte $02 dword [ v_blk ] add                                
    lineOffsetZero call                                         
    ret 

LABEL previousBlock  
    byte ( START_BLOCK_NUMBER + 2 ) dword [ v_blk ] cmp         
    .forward js 
    byte $02 dword [ v_blk ] sub                                
LABEL .forward  
    lineOffsetZero call                                         
    ret 

LABEL otherBlock  
    [ v_blk ] ecx mov                                           
    [ v_otherBlock ] ecx xchg                                   
    ecx [ v_blk ] mov                                           
\    mov dword [ esi ], ecx            \  update the block number in SOS
    ret 

LABEL shadow  \ alternate between source and shadow blocks
    byte $01 dword [ v_blk ] xor                                
    ret 

LABEL insert0  
    [ v_lcad ] ecx mov                                          
    [ v_words ] ecx add                                         
    [ v_lcad ] ecx xor                                          
    $FFFFFF00 ecx and                                           
    insert1 jz 
    [ v_words ] ecx mov                                         
LABEL .back  
    _DROP_ 
    .back loop 
    ret 

LABEL insert1  
    esi push 
    [ v_lcad ] esi mov                                          
    esi ecx mov 
    esi dec 
    esi edi mov 
    [ v_words ] edi add                                         
    $02 edi shl  
    [ v_cad ] ecx sub                                           
    .forward js 
    $02 esi shl  
    std 
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    cld 
LABEL .forward  
    esi pop 
    $02 edi shr  
    edi inc 
    edi [ v_curs ] mov                                          
    [ v_words ] ecx mov                                         
LABEL .back  
    edi dec 
    _TOS_ [ ( edi * 4 ) + $00 ] mov                             
    _DROP_ 
    .back loop 
    ret 

LABEL insert  
    insert0 call 
    [ action ] cl mov                                           
    cl [ edi * 4 + $00 ] xor                                    
    $03 cl cmp  
    .forward jnz 
    $04 byte [ action ] mov                                     
    colour_green dword [ keyboard_colour ] mov                  
LABEL     .forward  
    ret 

LABEL _word1  
    dword [ aword ] pop                                         
    ex1 dword [ aword ] mov                                     
    ret 

LABEL _word  
    _word1 dword [ aword ] mov                                  
    dword accept jmp                                            

LABEL tokenAction_1  
    _DUP_ 
    $01 _TOS_ mov  
    $04 byte [ action ] cmp                                     
    .forward2 jz 
    $03 al mov  
LABEL .forward2  
    byte $0A dword [ base ] cmp                                 
    .forward jz 
    $10 al xor  
LABEL .forward  
    _SWAP_ 
    $02 dword [ v_words ] mov                                   
    short insert jmp                                            

LABEL tokenAction  
    $0A byte [ action ] test                                    
    .forward jnz 
    _TOS_ edx mov 
    $FC000000 edx and                                           
    .forward2 jz 
    $FC000000 edx cmp                                           
    tokenAction_1 jnz                                           
LABEL .forward2  
    $05 _TOS_ shl  
    $02 al xor  
    $04 byte [ action ] cmp                                     
    .forwardBack jz 
    $0B al xor  
LABEL .forwardBack  
    byte $0A dword [ base ] cmp                                 
    .forward4 jz 
    $10 al xor  
LABEL .forward4  
    $01 dword [ v_words ] mov                                   
    insert jmp 
LABEL .forward  
    $09 byte [ action ] cmp                                     
    .forward3 jnz 
    _TOS_ edx mov 
    $05 edx shl  
    $05 edx sar  
    _TOS_ edx cmp 
    .forward5 jz 
LABEL .forward3  
    _DROP_ 
    ret 
LABEL .forward5  
    $05 _TOS_ shl  
    $06 al xor  
    short .forwardBack jmp                                      

LABEL enstack  
    _DUP_ 
    [ v_cad ] _TOS_ mov                                         
    [ v_pcad ] _TOS_ sub                                        
    .forward jz 
    _TOS_ ecx mov 
    edx _TOS_ xchg 
    esi push 
    [ v_cad ] esi mov                                           
    [esi*4-$04] esi lea                                         
    [ v_trash ] edi mov                                         
LABEL .back  
    std 
    _DROP_ 
    cld 
    stosd 
    .back loop 
    edx _TOS_ xchg 
    stosd 
    edi [ v_trash] mov                                          
    esi pop 
LABEL .forward  
    _DROP_ 
    ret 

LABEL del  
    enstack call 
    [ v_pcad ] edi mov                                          
    [ v_lcad ] ecx mov                                          
    edi ecx sub 
    $02 edi shl  
    esi push 
    [ v_cad ] esi mov                                           
    $02 esi shl  
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    esi pop 
    dword cursorLeft jmp                                        

LABEL act0  
    enstack call 
    dword cursorLeft jmp                                        

LABEL act1  
    $01 al mov  
    short actt jmp 

LABEL act3  
    $03 al mov  
    short actt jmp 

LABEL act4  
    $04 al mov  
    short actt jmp 

LABEL act9  
    $09 al mov  
    short actt jmp 

LABEL act10  
    $0A al mov  
    short actt jmp 

LABEL act11  
    $0B al mov  
    short actt jmp 

LABEL act13  
    $0D al mov  
    short actt jmp 

LABEL act14  
    $0E al mov  
    short actt jmp 

LABEL act7  
    $07 al mov  

LABEL actt  \ ( action -- )
    al [ action ] mov                                           
    insert dword [ aword ] mov                                  
    [ ( _TOS_ * 4 ) + actionColourTable ] _TOS_ mov             
LABEL actn  
    _TOS_ [ keyboard_colour ] mov                               
    _TOS_ pop 
    _DROP_ 
    dword accept jmp                                            

LABEL actv  \ magenta variable action
    $0C byte [ action ] mov                                     
    colour_magenta _TOS_ mov                                    
    .act1 dword [ aword ] mov                                   
    short actn jmp 

LABEL .act1  
    _DUP_ 
    _TOS_ _TOS_ xor                                             
    dword [ v_words ] inc                                       
    dword insert jmp                                            

LABEL eout  \ ( -- )   \ leave the editor
    _TOS_ pop 
    _DROP_ 
\    _DROP_      \  discard the screen number  v_blk  value on the top of the stack
    ex1 dword [ aword ] mov                                     
    nul dword [ anumber ] mov                                   
    $00 byte [ alpha0 + ( 4 * 4 ) ] mov                         
    nul0 dword [ alpha0 + 4 ] mov                               
    colour_yellow dword [ keyboard_colour ] mov                 
    $00 byte [ v_acceptMode ] mov                               
    $00 byte [ v_hintChar ] mov                                 
    dword accept jmp                                            

LABEL destack  
    [ v_trash ] edx mov                                         
    TRASH_BUFFER edx cmp                                        
    .forward jnz 
    ret 
LABEL .forward  
    byte $08 edx sub                                            
    [edx+$04] ecx mov                                           
    ecx [ v_words ] mov                                         
LABEL .back  
    _DUP_ 
    [edx] _TOS_ mov                                             
    byte $04 edx sub                                            
    .back loop 
    byte $04 edx add                                            
    edx [ v_trash ] mov                                         
    dword insert0 jmp                                           

LABEL editorKeyTable  
    nul            ,   del           ,   eout           ,   destack ,   \
    act1           ,   act3          ,   act4           ,   shadow ,   \ y r g *
    cursorLeft     ,   cursorUp      ,   cursorDown     ,   cursorRight ,   \ l u d r
    previousBlock  ,   actv          ,   act7           ,   nextBlock ,   \ - m c +
    nul            ,   act11         ,   act10          ,   act9 ,   \ _ S C t
    nul            ,   nul           ,   nul            ,   otherBlock ,   \ _ _ _ j
LABEL ekbd0  
    act13          ,   act14         ,   nul            ,   act0 ,   \ a b _ _
    $15           c,  $25          c,  $07           c,  $00 c,  \ four characters to display on the bottom line of the keyboard

LABEL editorKeyTableHintChars                                   \ display the current edit colour and mode in the bottom right hand corner of the keyboard
    $00 c,  $00 c,  $00 c,  $00 c,                              \
    $0B c,  $01 c,  $0D c,  $00 c,                              \ y r g _
    $00 c,  $00 c,  $00 c,  $00 c,                              \ l u d r
    $00 c,  $09 c,  $0A c,  $00 c,                              \ - m c +
    $00 c,  $38 c,  $3A c,  $02 c,                              \ _ S C t
    $00 c,  $00 c,  $00 c,  $00 c,                              \ _ _ _ j
    $05 c,  $13 c,  $00 c,  $00 c,                              \ a b _ _

\ Editor keypad display
\ _ S C t  y r g *
\ c d f j  l u d r
\ a b _ k  - m c +
\     x . i

LABEL editorKeyboard                                            \ the main editor keyboard icons
    $0B c,  $01 c,  $0D c,  $2D c,                              \ y r g *
    $0C c,  $16 c,  $10 c,  $01 c,                              \ l u d r
    $23 c,  $09 c,  $0A c,  $2B c,                              \ - m c +
    $00 c,  $38 c,  $3A c,  $02 c,                              \ _ S C t
    $00 c,  $00 c,  $00 c,  $22 c,                              \ _ _ _ j    \    db 0x00, 0x10, 0x0E, 0x22  ; _ d f j
    $05 c,  $13 c,  $00 c,  $00 c,                              \ a b _ _

LABEL set_e_main  
    ekbd0 dword [ shiftAction ] mov                             
    ( editorKeyboard - 4 ) dword [ currentKeyboardIcons ] mov   
    colour_yellow dword [ keyboard_colour ] mov                 
    ret 

LABEL e0  
    _DROP_ 
    short plus_e2 jmp                                           

LABEL edit  \ ( n -- )   \ edit block n
    [ v_blk ] ecx mov                                           
    ecx [ v_otherBlock ] mov                                    
    _TOS_ [ v_blk ] mov                                         \ move TOS to blk variable
    _DROP_ 
LABEL e_  
    $FF byte [ v_acceptMode ] mov                               
    refresh call 
LABEL plus_e  
    tokenAction dword [ anumber ] mov                           
    $25 byte [ alpha0+4*4 ] mov                                 
    e0 dword [ alpha0 + 4 ] mov                                 
LABEL plus_e2  
    set_e_main call 
LABEL .back  
    clearHintChar call                                          
    get_key call 
    _TOS_ push 
    editorKeyTableHintChars _SCRATCH_ mov                       
    [ _SCRATCH_ + _TOS_ ] byte al mov                           
    _TOS_ [ v_hintChar ] mov                                    
    _TOS_ pop 
    [ ( _TOS_ * 4 ) + editorKeyTable ] call                     
\    mov byte [ v_hintChar ], $00               \  clear the hint character when we have finished editing text
    _DROP_ 
    short .back jmp 

LABEL convertAddress                                            \ ( a32 -- )    set up the block at the given 32 bit cell address, including the cursor position
    _TOS_ _SCRATCH_ mov                                         
    $00FF _SCRATCH_ and                                         
    _SCRATCH_ [ v_curs ] mov                                    \ cell offset in block
    cellAddressToBlock call                                     
    _TOS_ [ v_blk ] mov                                         
    _DROP_ 
    ret 

LABEL editAddress  \ ( a32 -- )    edit the block at the given 32 bit cell address, including the cursor position
    convertAddress call                                         
    abort_e2 call                                               \ abort and show the editor display
    ret 

LABEL keypd  \ display the keypad vectors and display characters at the address on top of the return stack
    edx pop                                                     \ keypd is followed by call table then keymap
    edx [ vector ] mov                                          \ edx points to the next colorForth word to be executed
    ( 28 * 5 ) edx add                                          \ 28 keys, 5 bytes per compiled call
    edx [ currentKeyboardIcons ] mov                            
    byte +16 edx sub                                            
    edx [ shiftAction ] mov                                     
LABEL .back  
    get_key call \ Pause
    [ vector ] edx mov                                          
    _TOS_ edx add 
    [ ( _TOS_ * 4 ) + edx + $05 ] edx lea                       
    [ edx - $04 ] edx add                                       
    _DROP_ 
LABEL pad1  
    edx call 
    short keypd.back jmp                                        

\ *****************************************************************************
\ QWERTY support
\ *****************************************************************************

LABEL qwertyKeyboard  
    0          ,   
    0          ,   
    0          ,   
    $01040f17  ,                                                \ 'qwer'
    0          ,   
    0          ,   

LABEL qwertToggleBase  
    ((setDecimalMode - $$) ^ (setHexMode - $$)) dword [ current ] xor 
    $2F byte [ ( numb0 + 12 ) ] xor                             
LABEL qwertToggleBase1  
\    call [ current ]
\    mov dword [ qwertyKeyboard ], $00           \  '' => decimal
\    cmp dword [ base ], byte +0x10
\    jnz .forward
\    mov dword [ qwertyKeyboard ], $00150414     \  'hex'
\ .forward:
\    mov dword [ currentKeyboardIcons ], pad1
\    mov dword [ shiftAction ], qwertyKeyboard
    ret 

LABEL qwertyAction4  
    qwertToggleBase call                                        
    qwertyAction3 jmp                                           

LABEL qwertyActionTable  
    endn ,   endn ,   xn ,   qwertyAction3 ,   qwertyAction4 ,   

LABEL qwertFunction1  
    right call 
    $C7        c,  
    ( qwertyKeyboard + 4 ) _TOS_ add                            
    es push 
    ss push 
    _TOS_ [_TOS_] or                                            
    qwertToggleBase1 call                                       
    $00 byte [ v_sign ] mov                                     
    [ v_digin ] _TOS_ mov                                       
LABEL qwertyAction5  
    get_qwerty_key call                                         
    .forward4 jz 
    dword [ _TOS_ * 4 + qwertyActionTable - $200 ] jmp          
LABEL .forward4  
    _TOS_ _TOS_ test                                            
    qwertyAction3 jng                                           
    $23 al cmp  
    .forward3 jz 
    [ v_digin ] _TOS_ mov                                       
    [ base ] _TOS_ cmp                                          
    .forward2 jns 
    $FF byte [ v_sign ] test                                    
    .forward jz 
    _TOS_ neg 
LABEL .forward  
    0 [ESI] edx mov                                             
    [ base] edx imul                                            
    _TOS_ edx add 
    edx 0 [ESI] mov                                             
LABEL .forward2  
    short qwertyAction3 jmp                                     
LABEL .forward3  
    _TOS_ [ v_sign ] xor                                        
    dword 0 [ESI] neg                                           

LABEL qwertyAction3  
    _DROP_ 
    short qwertyAction5 jmp                                     

LABEL qwertToggleBaseTable2                                     
    lj ,   lj ,   x ,   

LABEL qwertyFunction2  
    $02150402 dword [ ( qwertyKeyboard + 4 ) ] mov              \ 'text'
    right call 
    $01 dword [ v_words ] mov                                   
    $01 dword [ chars] mov                                      
    _DUP_ 
    $00 dword 0 [ESI] mov                                       
    $1C byte [ bits_ ] mov                                      
LABEL .back  
    .forward jz 
    $83 _TOS_ cmp  
    .forward jns 
    dword [ _TOS_*4 + qwertToggleBaseTable2 - $200 ] jmp        
LABEL .forward  
    _TOS_ _TOS_ test                                            
    .forward2 jng 
    $30 _TOS_ cmp  
    .forward2 jns 
    _DUP_ 
    echo_ call 
    pack call 
    dword [ chars] inc                                          
LABEL .forward2  
    _DROP_ 
    get_qwerty_key call                                         
    short .back jmp 

LABEL qwertyAction2  
    qwertToggleBase call                                        
    dword nul0 jmp 

LABEL qwertyAction1  
    dword [ alpha0 + 4 ] jmp                                    

LABEL qwertyTable1  
    nul0       ,   
    nul0       ,   
    nul0       ,   
    qwertyAction1 ,                                             
    qwertyAction2 ,                                             

LABEL qwertyDoAction  
    $00 dword [ ( qwertyKeyboard + 4 ) ] mov                    \ clear the 'text' string
    qwertyKeyboard dword [ shiftAction ] mov                    
    pad1 dword [ currentKeyboardIcons ] mov                     

LABEL .back2  
    get_qwerty_key call                                         
    .forward jz 
    dword [ ( _TOS_ * 4 ) + qwertyTable1 - $0200 ] jmp          

LABEL .forward  
    $30 al cmp  
    .back jnz 
    $02150402 dword [ ( qwertyKeyboard + 4 ) ] mov              \ 'text'
    _DROP_ 
    short .back2 jmp                                            
LABEL .back  
    _TOS_ _TOS_ test                                            
    .forward3 jng 
    $FFFFFFFF dword [ ( qwertyKeyboard + 4 ) ] test             
    .forward2 jnz 
    $0A byte [ v_digin ] cmp                                    
    qwertFunction1 js                                           
LABEL .forward2  
    $30 _TOS_ cmp  
    .forward3 jns 
    qwertyFunction2 call                                        
    [ aword ] call 
    _DUP_ 
LABEL .forward3  
    _DROP_ 
    dword accept jmp                                            

LABEL qwert  \ selects QWERTY keyboard entry
    qwertyDoAction dword [ x_qwerty ] mov                       
    ret 

\ *****************************************************************************

LABEL abort_action  
    ( RELOCATED / 4 ) edi cmp                                   \ if we are compiling a block, show the location of the error
    \ edi is a cell address, so divide by 4
    .forward jc 
    _DUP_ 
    [ v_blk ] _TOS_ mov                                         
    _TOS_ [ v_otherBlock ] mov                                  \ save the last block to be edited
    edi _TOS_ mov 
    convertAddress call                                         
LABEL .forward  
    RETURN_STACK_0 esp mov                                      
    ( DATA_STACK_0 + 4 ) esi cmp                                
    .forward2 jc 
    ( DATA_STACK_0 + 4 ) esi mov                                
LABEL .forward2  
    forthd dword [ tokenActions + ( 3 * 4 ) ] mov               
    qcompile dword [ tokenActions + ( 4 * 4 ) ] mov             
    cnum dword [ tokenActions + ( 5 * 4 ) ] mov                 
    cshort dword [ tokenActions + ( 6 * 4 ) ] mov               
    $2F _TOS_ mov                                               \ '?' character to follow the display of the unknown word
    echo_ call 
\    jmp abort_e2
    dword accept jmp                                            

\ *****************************************************************************

LABEL rquery  \ r?
    _DUP_ 
    RETURN_STACK_0 _TOS_ mov                                    
    esp _TOS_ sub 
    1 _TOS_ shr 
    1 _TOS_ shr 
    ret 

LABEL boot  
    \ see http://wiki.osdev.org/PS2_Keyboard#CPU_Reset
    $FE al mov  
    al $64 out  
    short $ jmp                                                 \ we should never get here, because the processor will be rebooted... stop here just in case

LABEL wipe  \ ( -- )    \ wipe the currently edited block
    _DUP_ 
    [ v_blk ] _TOS_ mov                                         
    $40 ecx mov  
LABEL wipe2  
    edi push 
    blockToCellAddress call                                     \ add the RELOCATED block number offset and convert to cell address
    2 _TOS_ shl                                                 \ convert to byte address
    _TOS_ edi mov 
    _TOS_ _TOS_ xor                                             
    stosd rep 
    edi pop 
    _DROP_ 
    ret 

LABEL wipes  \ ( startblock# #blocks -- )    \ wipes #blocks starting from block startblock#   ( was erase )
    _TOS_ ecx mov 
    $06 ecx shl                                                 \  convert blocks to cells, multiply by 64
    _DROP_ 
    wipe2 jmp 

LABEL copy_  \ ( blk -- )    \ copy the given block (and shadow) to the currently displayed block (and shadow)
    byte $0C _TOS_ cmp                                          \ below block 12 is machine code
    abort jc 
    edi push 
    esi push 
    ecx push 
    blockToCellAddress call                                     \ source block
    $02 _TOS_ shl                                               \ convert cell address to byte address
    _TOS_ esi mov 
    [ v_blk ] _TOS_ mov                                         
    blockToCellAddress call                                     \ destination block
    $02 _TOS_ shl                                               \ convert cell address to byte address
    _TOS_ edi mov 
    $0200 ecx mov  
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    ecx pop 
    esi pop 
    edi pop 
    _DROP_ 
    ret 

LABEL debug  
    $302B5 dword [ v_xy ] mov                                   
    _DUP_ 
    [ God ] _TOS_ mov                                           
    dword [_TOS_] push                                          
    dotHex call 
    _DUP_ 
    _TOS_ pop 
    dotHex call 
    _DUP_ 
    [ main ] _TOS_ mov                                          
    dotHex call 
    _DUP_ 
    esi _TOS_ mov 
    dword dotHex jmp                                            

\ *****************************************************************************

LABEL tic0  
    dword [ v_words ] dec                                       
    .forward jz 
    _DROP_ 
    short tic0 jmp 
LABEL .forward  
    ret 

LABEL tic_  \ ( -- a )   \ return the byte address of the next word entered
    _word call                                                  \ allow user to enter the word to  search for
    tic0 call                                                   \ remove the entered word from the stack
    find_ call                                                  \ find the word in the dictionary, return its index in  ecx
    abort jnz 
    [ ( ecx * 4 ) + ForthJumpTable ] _TOS_ mov                  \ return the word's address from the jump table
    ret 

LABEL itick  
    $FFFFFFF0 _TOS_ and                                         
    find_ call 
    [ ( ecx * 4 ) + ForthJumpTable ] _TOS_ mov                  
    ret 

\ *****************************************************************************

\ ToDo: fix this!!!
LABEL showWords_  \ ( -- )   \ show all words in the Forth wordlist
    show call 
    edi push 
    setRed call 
    [ ForthNames - 4 ] edi lea                                  \ set edi to the bottom of the Forth name table
    [ v_ForthWordCount ] ecx mov                                \ count of Forth wordlist words
LABEL .loop  
    type_ call                                                  \ show one Shannon-Fano encoded word
    space_ call 
    edi inc 
    .loop loop 
    edi pop 
    ret 

LABEL words_  
    show call 
LABEL .back  
    page_ call 
    displayBlockNumber call                                     
    plusList call 
    .back jmp 

\ *****************************************************************************

    db 0 4 align                                                \ fill the gap with 0's

LABEL dma_ready  
    1          c,                                               \ set to zero once DMA is initialized
    0          c,  
LABEL command  
    0          c,  
    0          c,  
LABEL cylinder  
    0          c,  
    0          c,                                               \ head
    1          c,                                               \ starting sector
    2          c,                                               \ 512 bytes/sector
    18         c,                                               \ sectors to transfer
    27         c,                                               \ standard gap 3 for DD floppy
    $ff        c,                                               \ data length, should be $ff unless sector size = 0
    0          c,  

LABEL ready  
    $3F4 dx mov                                                 \ diskette data rate select register
LABEL .back  
    dx al in 
    al DEBUGGER out                                             
    1 al shl                                                    \ shift RQM bit into carry flag, set when ready for data transfer
    .back jnc 
    edx inc                                                     \ set DX to data register $3F5
    al al test                                                  \ set sign bit (was bit 6, DIO, before shl) if FDC has data for CPU
    ret 

\ floppy disk commands, see http://www.brokenthorn.com/Resources/OSDev20.html
LABEL cmd  
    [ command] edx lea                                          
    al [edx] mov                                                \ store command byte

LABEL cmd0  
    esi push 
    edx esi mov 

LABEL cmd1  
    ready call                                                  \ side effect: sets DX to data register
    .nodata jns 
    dx al in                                                    \ fetch byte of data from FDC
    al DEBUGGER out                                             
    short cmd1 jmp 
LABEL .nodata                                                   \ FDC had nothing for us
    lodsb                                                       \ next byte of command string
    $1E ah mov                                                  \ for time delay
    al dx out 
LABEL .back  
    al DEBUGGER out                                             
    ah dec 
    .back jnz 
    cmd1 loop 
    esi pop 
    ret 

LABEL _sense  
    $08 al mov  
    $01 cl mov                                                  \ number of bytes in command
    cmd call 
    ready call 
    dx al in 
    al DEBUGGER out                                             
    $80 al cmp  
    ret 

LABEL seek  
    al $0B out  
LABEL .back  
    _sense call 
    .back jnz 
    $0F al mov  
    $03 cl mov                                                  \ number of bytes in command

LABEL cmd_wait  
    cmd call 
LABEL .back  
    _sense call 
    .back jz 
    ret 

LABEL calibrate  
    $1c al mov                                                  \ motor on
\    call toggle_motor
    75000000 ecx mov                                            
    $ loop                                                      \ busy-wait
    $07 al mov                                                  \ calibrate command
    $02 cl mov                                                  \ number of bytes in command
    short cmd_wait jmp                                          

LABEL dma_setup                                                 \ set up DMA channel 2
    \ see  http://bochs.sourceforge.net/techspec/PORTS.LST
    al $05 out                                                  \ word count byte 0
    ah al mov 
    al $05 out                                                  \ word count byte 1
    SECTOR_BUFFER _TOS_ mov                                     
    al $04 out                                                  \ address byte 0
    ah al mov 
    al $04 out                                                  \ address byte 1
    16 _TOS_ shr 
    al $81 out                                                  \ address byte 2
    $b al mov  
    al $f out                                                   \ DMA write mask register
    $2A1 word [ command + 1] mov                                \ step rate=10, head unload time=1
    \ head load time=1, non-DMA mode (low bit 0)
    \ this command returns nothing, that's why DMA can be off
    $03 al mov                                                  \ timing (SPECIFY command)
    $03 cl mov                                                  \ number of bytes in command
    cmd call 
    $00 word [ command + 1 ] mov                                
    $00 byte [ dma_ready ] mov                                  
    ret 

LABEL transfer  
    $09 cl mov                                                  \ number of bytes in command
    cmd call 
    byte [ cylinder] inc                                        \ cylinder number

LABEL .back  
    ready call 
    .back jns                                                   \ keep calling until ready
    ret 

LABEL _read                                                     \ close to CM's original READ from sources at colorforth.com
    $16 al mov  
    seek call 
    $E6 al mov                                                  \ read normal data
    \ M=1, multitrack (both heads)\  F=1 (double density), S=1 (skip deleted)
    transfer call 
    esi push 
    SECTOR_BUFFER esi mov                                       
    ( CYLINDER / CELL ) ecx mov                                 \ $1200
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    esi pop 
    ret 

LABEL write_cylinder                                            \ copy to buffer, then DMA write
    SECTOR_BUFFER edi mov                                       
    ( CYLINDER / CELL ) ecx mov                                 \ $1200
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    $1A al mov                                                  \ write DMA 2
    seek call 
    $C5 al mov                                                  \ write data
    short transfer jmp                                          

LABEL ready_floppy                                              \ called by both read and write
    al [ cylinder ] mov                                         \ cylinder number, assumed to be no larger than 255
    _DUP_ 
    $03F2 dx mov                                                \ diskette DOR (digital output register)
    dx al in 
    al DEBUGGER out 
    $10 al test                                                 \ FD_MS_BUSY set?
    .return jnz 
    dword calibrate jmp                                         

LABEL .return  
    ret 

LABEL setup_dma                                                 \ called by both read and write
    ecx ecx xor 
    $ff byte [ dma_ready] test                                  \ check if DMA already set up
    .done jz                                                    \ already set up, return
    _DUP_ 
    CYLINDER - 1 _TOS_ mov                                      \ bytes to transfer less one
    dma_setup call 
    _DROP_ 
LABEL .done  
    ret 

LABEL read  \ ( c a -- c' a')
    setup_dma call 
    ready_floppy call                                           \ turn on and calibrate, side effect: dup #cylinder
    edi push 
    [ esi + $04 ] edi mov                                       
    $02 edi shl                                                 \ get byte address from cell address
    _read call 
    edi pop 
LABEL finish_read_write                                         \ updates cylinder and cell address
    _DROP_                                                      \ drop, cylinder back into EAX
    _TOS_ inc                                                   \ increment cylinder number
    CYLINDER / CELL dword 0 [ESI] add                           \ $1200 cells, normal floppy geometry
    ret 

LABEL write  \ (c a - c' a')
    setup_dma call 
    ready_floppy call                                           \ turn on and calibrate, side effect: dup #cylinder
    esi push 
    [ esi + $04 ] esi mov                                       
    $02 esi shl                                                 \ convert word address to byte address
    write_cylinder call                                         
    esi pop 
    short finish_read_write jmp                                 

\ *****************************************************************************

\ int $13 Disk Address Packet (DAP) pointed to by  si  :
: o_Int13_DAP_size ( -- )   $00 ;                               \ 2  $0010
: o_Int13_DAP_num_sectors ( -- )   $02 ;                        \ 2  $0001
: o_Int13_DAP_address ( -- )   $04 ;                            \ 2  $2000
: o_Int13_DAP_segment ( -- )   $06 ;                            \ 2  $0000
: o_Int13_DAP_LBA_32 ( -- )   $08 ;                             \ 4  $00000028
: o_Int13_DAP_spare_32 ( -- )   $0C ;                           \ 4  $00000000

LABEL fetchDX_  \ ( -- w )   \ return the value of dx from the BIOS when the program is first run. dl = drive number
    edi push 
    _DUP_ 
    eax eax xor 
    (data_area - $$ + BOOTOFFSET) edi mov                       \ setup the data index pointer
    [ byte di + ( driveinfo_Drive_DX - data_area ) ] al mov     \ restore the boot drive into dl
    edi pop 
    ret 

LABEL storeDAP_  \ ( sector -- )   \ setup the DAP to the given LBA sector number

    edi push 
    DAP_BUFFER edi mov                                          

    eax dword [ edi + o_Int13_DAP_LBA_32 ] mov                  

    eax eax xor 
    eax dword [ edi + o_Int13_DAP_spare_32 ] mov                

    \ now set the configuration buffer values from the registers
    $0010 eax mov  
    ax word [ edi + o_Int13_DAP_size ] mov                      \ setup buffer size
    $0001 eax mov  
    ax word [ edi + o_Int13_DAP_num_sectors ] mov               
    SECTOR_BUFFER word [ edi + o_Int13_DAP_address ] mov        
    eax eax xor 
    ax word [ edi + o_Int13_DAP_segment ] mov                   
    _DROP_ 
    edi pop 
    ret 

LABEL v_saved_regs  
    0          ,   

#16 BitsMode ! 
LABEL storeBefore  \ ( -- )   \ store registers to the V_REGS array
    ax word [ V_REGS + $00 ] mov                                
    bx word [ V_REGS + $04 ] mov                                
    cx word [ V_REGS + $08 ] mov                                
    dx word [ V_REGS + $0C ] mov                                
    si word [ V_REGS + $10 ] mov                                
    di word [ V_REGS + $14 ] mov                                
    bp word [ V_REGS + $18 ] mov                                
    ax push                                                     \ save eax
    pushfd                                                      \ push the 32 bit eflags register onto the stack
    ax pop                                                      \ and pop it off into eax
    ax word [ V_REGS + $1C ] mov                                \ eflags
    ax pop 
    ax word [ V_REGS + $1E ] mov                                \ eflags top 16 bits
    ax pop                                                      \ restore eax
    ret 

LABEL storeAfter  \ ( -- )   \ store registers to the V_REGS array
    ax word [ V_REGS + $20 ] mov                                
    bx word [ V_REGS + $24 ] mov                                
    cx word [ V_REGS + $28 ] mov                                
    dx word [ V_REGS + $2C ] mov                                
    si word [ V_REGS + $30 ] mov                                
    di word [ V_REGS + $34 ] mov                                
    bp word [ V_REGS + $38 ] mov                                
    ax push                                                     \ save eax
    pushfd                                                      \ push the 32 bit eflags register onto the stack
    ax pop                                                      \ and pop it off into eax
    ax word [ V_REGS + $3C ] mov                                \ eflags
    ax pop 
    ax word [ V_REGS + $3E ] mov                                \ eflags top 16 bits
    ax pop                                                      \ restore eax
    ret 
#32 BitsMode ! 

LABEL bios_rusb  \ aka  "stop_"
    cli                                                         \ disable interrupts
    pushad  \ Pushes all general purpose registers onto the stack in the following order:
            \ EAX, ECX, EDX, EBX, ESP, EBP, ESI, EDI. The value of ESP is the value before the actual push of ESP
            \  7    6    5    4    3    2    1    0   offset in cells from ESP

    edx push 
    ecx push 
    ebx push 
    esi push 
    edi push 

\    call storeBefore

    BIOS_thunk call 

\    call storeAfter

    edi pop 
    esi pop 
    ebx pop 
    ecx pop 
    edx pop 

    popad   \ restore the stack values pushed by  pushad

    ret 

LABEL BIOS_thunk  \ ( -- )   \ call the BIOS - registers will have previously been setup
    setRealModeAPI call                                         
#16 BitsMode ! 
    ax push 
    es push                                                     \ this operation messes with ES
    di push                                                     \ and DI
    storeBefore call                                            
    $13 int 
    $ jc                                                        \ stop here on error
    storeAfter call 
    di pop 
    es pop 
    ax pop 
    cli                                                         \ BIOS might have left interrupts enabled
    setProtectedModeAPI call                                    \ called from 16 bit code, returns to 32 bit code
#32 BitsMode ! 
    ret 

LABEL th_   \ ( ax bx cx dx si di es -- w )   \ th ( thunk to BIOS Int $13 )
            \ eax = $ DH DL AH AL , returns in same order
    cli                                                         \ disable interrupts
    pushad  \ Pushes all general purpose registers onto the stack in the following order:
            \ EAX, ECX, EDX, EBX, ESP, EBP, ESI, EDI. The value of ESP is the value before the actual push of ESP
            \  7    6    5    4    3    2    1    0   offset in cells from ESP

\    call storeDAP_

    edi push 
    (data_area - $$ + BOOTOFFSET) di mov                        \ setup the data index pointer
    [ word di + ( driveinfo_Drive_DX - data_area) ] dx mov      \ restore the boot drive from dx (and head? )

    SECTOR_BUFFER ebx mov                                       
    ( $0200 + ( ( SECTOR_BUFFER_SIZE / 512 ) & $FF ) ) eax mov  \ read n sectors to fill the buffer
    $0201 ecx mov                                               \ cylinder | sector

    BIOS_thunk call 

    edi pop 
    popad   \ restore the stack values pushed by  pushad
    ret 

LABEL rsect_  \ ( sector -- ax )   \
    pushad  \ Pushes all general purpose registers onto the stack
    edi push 

\    call sector_chs                 \  store th sector number into the Disk Address Packet

    (data_area - $$ + BOOTOFFSET) di mov                        \ setup the data index pointer
    [ word di + ( driveinfo_Drive_DX - data_area) ] dx mov      \ restore the boot drive from dx (and head? )

    cli                                                         \ disable interrupts
\    call fetchDX_
\    mov esi, DAP_BUFFER
\    _DUP_
    $0201 eax mov                                               \ BIOS read, one sector
    SECTOR_BUFFER bx mov                                        
    BIOS_thunk call 

    edi pop 
    popad   \ restore the stack values pushed by  pushad
    ret 

LABEL stop_  
    storeDAP_ call                                              \ store the sector number into the Disk Address Packet
    cli                                                         \ disable interrupts
    pushad  \ Pushes all general purpose registers onto the stack
    fetchDX_ call 
    DAP_BUFFER esi mov                                          
    _DUP_ 
    $4200 eax mov                                               \ BIOS read

    BIOS_thunk call 

    edi pop 
    popad   \ restore the stack values pushed by  pushad
    ret 

\ *****************************************************************************

LABEL dma_write  
    $1A al mov                                                  \ write DMA 2
    seek call 
    esi edx mov 
    $06 cl mov  
    cmd0 call 
LABEL .back  
    ready call 
    .back jns 
    ret 

\ *****************************************************************************
\ *****************************************************************************

: FORTH_INITIAL_WORD_COUNT ( -- )   ForthJumpTableROM_end ForthJumpTableROM -  4 / ;  \ in cells
: MACRO_INITIAL_WORD_COUNT ( -- )   MacroJumpTableROM_end MacroJumpTableROM -  4 / ;  \ in cells

LABEL warm  \ warm start
    STACK_MEMORY_START _SCRATCH_ mov                            \ start of stack memory area
    ( TOTAL_STACK_SIZE >> 2 ) ecx mov                           \ number of 32 bit cells to fill with the pattern
LABEL .back  
    $55555555 dword [ _SCRATCH_ ] mov                           \ fill with this pattern
    $04 _SCRATCH_ add                                           
    .back loop 

    ecx ecx xor                                                 \ assumed by initshow to have been previously zeroed

    initshow call                                               \ sets up do-nothing "show" process
\    call initserve                      \  sets up do-nothing "serve" process
\    call stop_                           \  turn off floppy motor and point trash to floppy buffer
\    mov byte [ dma_ready ], $01        \  not ready
    FORTH_INITIAL_WORD_COUNT dword [ v_ForthWordCount ] mov     \ initial #words
    MACRO_INITIAL_WORD_COUNT dword [ v_MacroWordCount ] mov     \ initial #macros
    TRASH_BUFFER dword [ v_trash] mov                           
    esi push 
\  orth wordlist
    [ ForthNamesROM ] esi lea                                   
    ForthNames edi mov                                          
    [ v_ForthWordCount ] ecx mov                                
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    [ ForthJumpTableROM ] esi lea                               
    ForthJumpTable edi mov                                      
    [ v_ForthWordCount ] ecx mov                                
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
\ Macro wordlist
    [ MacroNamesROM ] esi lea                                   
    MacroNames edi mov                                          
    [ v_MacroWordCount ] ecx mov                                
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi
    [ MacroJumpTableROM ] esi lea                               
    MacroJumpTable edi mov                                      
    [ v_MacroWordCount ] ecx mov                                
    movsd rep                                                   \ copy ecx 32 bit words from ds:esi to es:edi

    esi pop 
    H0 dword [ v_H ] mov                                        
    $00 dword [ x_qwerty ] mov                                  \ select non-qwerty mode
    ( RELOCATED >> ( 2 + 8 ) ) dword [ v_offset ] mov           \ $10000 >> 2 >> 8, offset of RELOCATED block 0 as 1024 byte block number

    \ setup  v_bytesPerLine
    [ vesa_XResolution ] _TOS_ mov                              
    $FFFF _TOS_ and                                             
    BYTES_PER_PIXEL _TOS_ imul                                  
    _TOS_ [ v_bytesPerLine + RELOCATED ] mov                    

    \ set up  fov
    [ vesa_YResolution ] _TOS_ mov                              
    $0000FFFF _TOS_ and                                         
    _TOS_ _SCRATCH_ mov                                         
    1 _SCRATCH_ shl                                             
    1 _TOS_ shr 
    _SCRATCH_ _TOS_ add                                         
    10 _TOS_ imul 
    _TOS_ [ v_fov + RELOCATED ] mov                             

    \ select which code to use, depending on the display mode
    0 byte [ displayMode ] mov                                  
    scrnw1 word [ vesa_XResolution ] cmp                        
    .forward jz 
    1 byte [ displayMode ] mov                                  
LABEL .forward  

    initIconSize call                                           

    cursorHome call                                             \ setup the initial display

    c_ call                                                     \ clear the stack
    _DUP_                                                       \
    START_BLOCK_NUMBER _TOS_ mov                                \ load the colorForth source at the given block
    _load_ call 
    dword accept jmp                                            

\ *****************************************************************************
\ *****************************************************************************

    db 0 4 align   \ must be on dword boundary for variables

LABEL hsvv_  
    0          ,   
    $28 db 0 times 

\ see http://karig.macminicolo.net/os/cf/u.htm for docs on the following
LABEL nc  \ ( -- a )
LABEL tsim  \ ( -- a )
    _DUP_ 
    v_nc LOAD_RELATIVE_ADDRESS                                  
    ret 

LABEL xy  \ ( -- a )
    _DUP_ 
    v_xy LOAD_RELATIVE_ADDRESS                                  
    ret 

LABEL fov_  \ ( -- a )
    _DUP_ 
    v_fov LOAD_RELATIVE_ADDRESS                                 
    ret 

LABEL tokenActions_  \ ( -- a )
    _DUP_ 
    tokenActions LOAD_RELATIVE_ADDRESS                          
    ret 

LABEL last_  \ ( -- a )
    _DUP_ 
    last LOAD_RELATIVE_ADDRESS                                  
    ret 

LABEL version_  \ ( -- a )
    _DUP_ 
    version LOAD_RELATIVE_ADDRESS                               
    ret 

LABEL vframe  \ ( -- a )  \ the video frame address, where we create the image to be displayed
    _DUP_ 
    [ frame ] _TOS_ mov                                         
    ret 

LABEL vars_  
    _DUP_ 
    vars LOAD_RELATIVE_ADDRESS                                  
    ret 

LABEL base_  
    _DUP_ 
    base LOAD_RELATIVE_ADDRESS                                  
    ret 

LABEL hex_  
    16 byte [ base ] mov                                        
    ret 

LABEL decimal_  
    10 byte [ base ] mov                                        
    ret 

LABEL block_  \ ( block -- address )   \ : block ( n -- n ) $400 * \  address is in bytes
    $0A _TOS_ shl  
    RELOCATED _TOS_ add                                         
    ret 

LABEL scrnw_  \ ( -- n )   screen width ( number of horizontal pixels )
    _DUP_ 
    _TOS_ _TOS_ xor                                             
    [ vesa_XResolution ] word ax mov                            
    ret 

LABEL scrnh_  \ ( -- n )    screen height ( number of vertical pixels )
    _DUP_ 
    _TOS_ _TOS_ xor                                             
    [ vesa_YResolution ] word ax mov                            \ v_scrnh
    ret 

LABEL bpp_  \ ( -- n )    bits per pixel
    _DUP_ 
    _TOS_ _TOS_ xor                                             
    [ vesa_BitsPerPixel ] byte al mov                           \ v_bitsPerPixel
    ret 

LABEL iconw_  \ ( -- n )    icon width ( number of pixels between characters, fixed font width )
    _DUP_ 
    [ v_iconw ] _TOS_ mov                                       
    ret 

LABEL iconh_  \ ( -- n )    icon height ( number of pixels between lines )
    _DUP_ 
    [ v_iconh ] _TOS_ mov                                       
    ret 

LABEL font_  \ ( -- n )    font16x24 font address
    _DUP_ 
    font16x24 LOAD_RELATIVE_ADDRESS                             
    ret 

LABEL last  \ ( -- a )
    _DUP_ 
    v_last LOAD_RELATIVE_ADDRESS                                
    ret 

LABEL blk  \ ( -- a )
    _DUP_ 
    v_blk LOAD_RELATIVE_ADDRESS                                 
    ret 

LABEL seeb  \ ( -- )    \ toggle the display of blue words in the editor
    byte [ v_seeb ] not                                         
    ret 

LABEL colourBlindModeToggle                                     \ ( -- )    \ toggle the editor display colorForth / ANS style
    byte [ v_colourBlindMode ] not                              
    ret 

LABEL curs  \ ( -- a )
    _DUP_ 
    v_curs LOAD_RELATIVE_ADDRESS                                
    ret 

0 [IF]  
LABEL stacks_  \  ( -- a )   \ return the address of the stack memory information ( see v_stack_info for details )
\  ETURN_STACK_SIZE
\  ATA_STACK_SIZE
\  TACK_MEMORY_START       \ bottom of stack memory
\  OTAL_STACK_SIZE
    _DUP_ 
    RETURN_STACK_0 - $3C _TOS_ mov                              \ top of task 0 return stack
    _DUP_ 
    DATA_STACK_0   - $3C _TOS_ mov                              \ top of task 0 data stack
    _DUP_ 
    RETURN_STACK_1 - $3C _TOS_ mov                              \ top of task 1 return stack
    _DUP_ 
    DATA_STACK_1   - $3C _TOS_ mov                              \ top of task 1 data stack
    _DUP_ 
    RETURN_STACK_2 - $3C _TOS_ mov                              \ top of task 2 return stack
\    _DUP_
\    mov _TOS_, DATA_STACK_2   - $3C           \  top of task 2 data stack
\    LOAD_RELATIVE_ADDRESS v_stack_info
    ret 
[THEN]  

LABEL ekt  \ ( -- a ) \  editor key table - variable containing vectors for editor keys beginning with null
    \ and the shift keys.  Then follows right hand top, middle, bottom rows,
    \ and left hand top, middle, bottom rows. (from ColorForth2.0a.doc)
    _DUP_ 
    editorKeyTable LOAD_RELATIVE_ADDRESS                        
    ret 

LABEL vword_  \ ( -- a )
    _DUP_ 
    v_words LOAD_RELATIVE_ADDRESS                               
    ret 

LABEL vregs_  \ ( -- a )
    _DUP_ 
    V_REGS eax mov 
    ret 

LABEL dap_  \ ( -- a )
    _DUP_ 
    DAP_BUFFER eax mov                                          
    ret 

LABEL sect_  \ ( -- a )
    _DUP_ 
    SECTOR_BUFFER eax mov                                       
    ret 

LABEL digin  \ ( -- a )
    _DUP_ 
    v_digin LOAD_RELATIVE_ADDRESS                               
    ret 

LABEL actc  \ ( -- a )
    _DUP_ 
    actionColourTable LOAD_RELATIVE_ADDRESS                     
    ret 

LABEL tickh  \ ( -- a )    HERE variable address
    _DUP_ 
    v_H LOAD_RELATIVE_ADDRESS                                   
    ret 

LABEL forths_  \ ( -- a )
    _DUP_ 
    v_ForthWordCount LOAD_RELATIVE_ADDRESS                      
    ret 

LABEL macros_  \ ( -- a )
    _DUP_ 
    v_MacroWordCount LOAD_RELATIVE_ADDRESS                      
    ret 

LABEL offset_  \ ( -- a )
    _DUP_ 
    v_offset LOAD_RELATIVE_ADDRESS                              
    ret 

LABEL aper  \ ( -- a )
    _DUP_ 
    frame LOAD_RELATIVE_ADDRESS                                 
    ret 

LABEL vesa  \ ( -- a )
    _DUP_ 
    ( VESA_BUFFER ) >> 2 _TOS_ mov                              
    ret 

LABEL trash_  \ ( -- a )
    _DUP_ 
    v_trash LOAD_RELATIVE_ADDRESS                               
    ret 

LABEL buffer_  \ ( -- a )
    _DUP_ 
    SECTOR_BUFFER _TOS_ mov                                     \  x25300
    ret 

LABEL cad  \ ( -- a )   , the address of the cursor as an offset from the start of the currently displayed block
    _DUP_ 
    v_cad LOAD_RELATIVE_ADDRESS                                 
    ret 

LABEL pcad  \ ( -- a )
    _DUP_ 
    v_pcad LOAD_RELATIVE_ADDRESS                                
    ret 

LABEL hsvv  \ ( -- a )
    _DUP_ 
    hsvv_ LOAD_RELATIVE_ADDRESS                                 
    ret 

LABEL displ  \ ( -- a )
    _DUP_ 
    displayShannonFanoActions LOAD_RELATIVE_ADDRESS             
    ret 

LABEL cBlindAddr_  \ ( -- a )
    _DUP_ 
    x_colourBlind LOAD_RELATIVE_ADDRESS                         
    ret 

    nop 4 align 

LABEL tens  
    10         ,   
    100        ,   
    1000       ,   
    10000      ,   
    100000     ,   
    1000000    ,   
    10000000   ,   
    100000000  ,   
    1000000000 ,   

LABEL x_numberDisplay  
    dotDecimal ,   

LABEL v_blk     \ the currently edited block
    START_BLOCK_NUMBER ,                                        \ the default edited block

LABEL v_otherBlock  \ the previously edited block
    START_BLOCK_NUMBER + 1 ,                                    \ the default other block is the shadow of the default edited block

LABEL v_curs  
    0          ,   

LABEL v_cad  
    0          ,   

LABEL v_pcad  
    0          ,   

LABEL v_lcad  
    0          ,   

LABEL v_trash                                                   \ pointer to "trash" buffer, saves words deleted while editing
    TRASH_BUFFER ,   

LABEL v_offset  
    0          ,   

LABEL v_nc  
    $10        ,   \ number of 18 Kbyte cylinders to save when saving the system = 288 K

LABEL v_bitsPerPixel  
    16         ,   \ default, set using VESA info

LABEL v_iconw  
    0          ,   \ iconw

LABEL v_iconh  
    0          ,   \ iconh

LABEL v_keypadY_iconh  
    0          ,   \ keypadY * iconh

LABEL v_nine_iconw  
    0          ,   

LABEL v_twentytwo_iconw                                         \ width of 12 history characters, 1 space and 9 keypad characters
    0          ,                                                \ to calculate the start of the history display, subtracted from the right edge of the screen

LABEL v_10000_iconw  
    0          ,   \ iconw*$10000

LABEL x_qwerty                                                  \ selects non-QWERTY if set to 0, else jumps to the address
    $FFFFFFFF  ,                                                \

LABEL x_abort  
    abort_action ,   

LABEL x_colourBlind  \ ( state -- state )
    colourBlindAction ,                                         

\ byte variables
LABEL v_seeb                                                    \ if = 255, show blue words in editor
    0          c,                                               \ 255 enable, 0 disable

LABEL v_colourBlindMode                                         \ if = 255, select ANS style editor display
    0          c,                                               \ 255 enable, 0 disable

LABEL v_not_cr  \ true to disable the  cr  before a red word is displayed in the editor
    0          c,  

LABEL v_acceptMode                                              \ if non zero, the keypad is in Edit mode, else  TIB mode
    0          c,                                               \ 255 enable, 0 disable

LABEL v_hintChar  \ the character to display in the bottom right hand corner of the keyboard as a hint to the colour being used
    0          ,   

    4 align 

LABEL currentKeyboardIcons                                      
    ( alphaKeyboard - 4 ) ,                                     

LABEL shiftAction  
    alpha0     ,   

LABEL vars  \ colorForth system variables start here
LABEL base  
    10         ,   

LABEL current  
    setDecimalMode ,                                            

LABEL keyboard_colour  
    colour_yellow ,                                             \ current key colour for displaying key presses

LABEL chars  
    1          ,   

LABEL aword  
    ex1        ,   

LABEL anumber  
    nul        ,   

LABEL v_words  
    1          ,   

LABEL v_regs  \ register values before and after Int $13 BIOS call
\ inputs
    0          ,   \ eax
    0          ,   \ ebx
    0          ,   \ ecx
    0          ,   \ edx
    0          ,   \ esi
    0          ,   \ edi
    0          ,   \ flags
    0          ,   \ spare
\ outputs
    0          ,   \ eax
    0          ,   \ ebx
    0          ,   \ ecx
    0          ,   \ edx
    0          ,   \ esi
    0          ,   \ edi
    0          ,   \ flags
    0          ,   \ spare

LABEL v_qwerty_key  
    0          ,   

LABEL v_digin  
    0          ,   

LABEL lit  
    adup       ,   

LABEL v_washColour  
    0          ,   

LABEL mark_MacroWordCount                                       
    MACRO_INITIAL_WORD_COUNT ,                                  \ initial #macros
    \ number of Macro words, saved by  mark , empty restores to this value
LABEL mark_v_ForthWordCount                                     
    FORTH_INITIAL_WORD_COUNT ,                                  \ initial #words
    \ number of Forth words, saved by  mark , empty restores to this value

LABEL mark_H  
    H0         ,   \ $100000     \  top of dictionary pointer H , saved by  mark , empty restores to this value

LABEL v_H  
    H0         ,   \ $40000*4    \  variable H , dictionary pointer HERE, where new definitions go

LABEL v_last  
    0          ,   

LABEL class  
    0          ,   

LABEL list  
    0          ,   
\ ( list + 4 )
    0          ,   

LABEL v_ForthWordCount  
    FORTH_INITIAL_WORD_COUNT ,                                  \ initial #words  \  number of words in the Forth wordlist, empty resets this value

LABEL v_MacroWordCount  
    MACRO_INITIAL_WORD_COUNT ,                                  \ initial #macros  \  number of words in the Macro wordlist, empty resets this value

LABEL tokenActions                                              \
    qignore    ,                                                \ 0  extension token
    execute_lit ,                                               \ 1
    num        ,                                                \ 2
LABEL adefine  \ where definitions go, either in the Macro Dictionary or Forth Dictionary
    forthd     ,                                                \ 3
    qcompile   ,                                                \ 4
    cnum       ,                                                \ 5
    cshort     ,                                                \ 6
    compile    ,                                                \ 7
    short_     ,                                                \ 8
    nul        ,                                                \ 9
    nul        ,                                                \ A
    nul        ,                                                \ B
    m_variable ,                                                \ C magenta variable
    nul        ,                                                \ D
    nul        ,                                                \ E
    nul        ,                                                \ F

LABEL v_xy  \ variable that holds the XY position for drawing characters, ( 0, 0 ) is top left
LABEL v_y  
    $0003      w,  
LABEL v_x  
    $0003      w,  

LABEL v_leftMargin  
    $00000003  ,   \ left margin

LABEL v_rightMargin  
    0          ,   \ right margin

\ xycr:
    \ dd 0

LABEL v_fov  \ abstract display scale
    0          ,   \ 10 * ( 2 * scrnh + scrnh / 2 )

LABEL frame  \ pointer to display frame buffer where we create our image, down from top of 32 Mbytes RAM ( $2000000 )
    $2000000 - ( MAX_SCREEN_WIDTH * MAX_SCREEN_HEIGHT * BYTES_PER_PIXEL ) ,   

LABEL v_frameBuffer                                             \ framebuffer address
    $00000000  ,                                                \

LABEL v_foregroundColour  
    $00000000  ,                                                \ the display foreground colour, set by  color

LABEL v_xc  
    $00000000  ,                                                \
LABEL v_yc  
    $00000000  ,                                                \

LABEL MacroNamesROM  
    $F0000000  ,                                                \ semicolon "\
    $C19B1000  ,                                                \ dup
    $CF833620  ,                                                \ qdup
\    dd $FF833620       \  ?dup
    $C0278800  ,                                                \ drop
    $2C88C000  ,                                                \ then
    $C6957600  ,                                                \ begin
\ MacroNamesROM_end:

LABEL MacroJumpTableROM                                         \ jump table for the macro wordlist
    semicolon  ,                                                \ \
    cdup       ,                                                \ compile dup
    qdup       ,                                                \ qdup
\    dd qdup             \  ?dup
    cdrop      ,                                                \ compile drop
    then       ,                                                \
    begin      ,                                                \
LABEL MacroJumpTableROM_end                                     

LABEL ForthNamesROM                                             \ displayed using cf2ansi
    $C6664000  ,                                                \ boot
    $BA8C4000  ,                                                \ warm
    $C4B9A080  ,                                                \ pause
    $8AC84C00  ,                                                \ macro
    $B1896400  ,                                                \ forth
    $90000000  ,                                                \ c
    $811E2000  ,                                                \ stop
    $2C800000  ,                                                \ th        th_ ( thunk to BIOS Int $13 )
    $145C0000  ,                                                \ read
    $B8B92000  ,                                                \ write
    $18248800  ,                                                \ rsect
    $F9832800  ,                                                \ @dx
    $F5817100  ,                                                \ !dap
    $69000000  ,                                                \ nc
    $91C62AC0  ,                                                \ comman(d)
    $82274000  ,                                                \ seek
    $145C1300  ,                                                \ ready
    $59100000  ,                                                \ act(tivate)
    $8643B800  ,                                                \ show
    $A1AE0000  ,                                                \ load
    $6A1AE000  ,                                                \ nload
    $F7435C00  ,                                                \ +load
    $2C839800  ,                                                \ thru
    $F6590730  ,                                                \ +thru
    $963A7400  ,                                                \ cblk   return the block number currently being compiled, calculated from  edi
    $C8828000  ,                                                \ here
    $FF472000  ,                                                \ ?lit
    $D7F80000  ,                                                \ 3,
    $D5F80000  ,                                                \ 2,
    $D3F80000  ,                                                \ 1,
    $FC000000  ,                                                \ ,
    $A2420000  ,                                                \ less
    $E59A3880  ,                                                \ jump
    $59493110  ,                                                \ accept
    $C4B80000  ,                                                \ pad
    $E893C580  ,                                                \ keypd
    $BBE24000  ,                                                \ wipe
    $BBE24800  ,                                                \ wipes     was erase
    $91E29800  ,                                                \ copy
    $8A8F4000  ,                                                \ mark
    $48E22000  ,                                                \ empt
    $48B90000  ,                                                \ emit
    $C0F57200  ,                                                \ digit
    $D4917200  ,                                                \ 2emit
    $EA000000  ,                                                \ .
    $C9D40000  ,                                                \ h.
    $C9D58000  ,                                                \ h.n
    $90800000  ,                                                \ cr
    $86259200  ,                                                \ space
    $C0776000  ,                                                \ down
    $4C0E4000  ,                                                \ edit
    $40000000  ,                                                \ e
    $A4400000  ,                                                \ lm
    $18800000  ,                                                \ rm
    $A8AE2C80  ,                                                \ graph
    $24CA4000  ,                                                \ text
    $E893C660  ,                                                \ keybo(ard)
    $C098F300  ,                                                \ debu(g)
    $52000000  ,                                                \ at
    $F6A40000  ,                                                \ +at
    $CB300000  ,                                                \ xy
    $C4B54000  ,                                                \ page
    $84851180  ,                                                \ screen
    $B1E10000  ,                                                \ fov
\    dd $B3D8C000   \  fifo
    $C6794000  ,                                                \ box
    $A3B20000  ,                                                \ line
    $91D0C400  ,                                                \ color
    $3912B100  ,                                                \ octant
    $86200000  ,                                                \ sp
    $A2C08000  ,                                                \ last
    $CCD89640  ,                                                \ unpac(k)
    $C4B2E800  ,                                                \ pack
    $C74E8000  ,                                                \ blk
    $8485AE00  ,                                                \ scrnw  screen width in pixels
    $8485B200  ,                                                \ scrnh  screen height in pixels
    $C78B1000  ,                                                \ bpp    bits per pixel
    $B1B10000  ,                                                \ font   address of font16x24
    $791B5C00  ,                                                \ iconw  icon width in pixels
    $791B6400  ,                                                \ iconh  icon height in pixels
    $C2820000  ,                                                \ ver
    $96618000  ,                                                \ curs
    $C7439740  ,                                                \ block
    $C36158A0  ,                                                \ vframe  video frame address, where we create the image to be displayed
    $C2A30000  ,                                                \ vars
\ new words
    $82263000  ,                                                \ seeb      ( see blue words, toggle )
\    dd $812CBA40   \  stacks_
    $C0650B00  ,                                                \ dotsf type a ShannonFano token
    $1AE30000  ,                                                \ rgb
    $C7340000  ,                                                \ bye
    $B98E0000  ,                                                \ word
    $4E840000  ,                                                \ ekt
    $5C662400  ,                                                \ abort
    $27974C80  ,                                                \ tickh  HERE variable address
    $C79AD640  ,                                                \ buffe(r)  buffer_
    $3B5A0840  ,                                                \ offset
    $27900000  ,                                                \ tic   tic_
    $5C482000  ,                                                \ aper
    $C2905000  ,                                                \ vesa
    $21586400  ,                                                \ trash trash_
    $C90C3840  ,                                                \ hsvv
    $283C4000  ,                                                \ tsim
    $C3731C00  ,                                                \ vword
    $C2295800  ,                                                \ vregs
    $C0B88000  ,                                                \ dap
    $82488000  ,                                                \ sect
    $B98E0800  ,                                                \ words
    $E8930000  ,                                                \ key
    $CFD12600  ,                                                \ qkey
    $C0F57600  ,                                                \ digin
    $CF741200  ,                                                \ qwert
    $1FE00000  ,                                                \ r?
    $6CD40000  ,                                                \ nul
    $92E00000  ,                                                \ cad
    $C525C000  ,                                                \ pcad
    $C0F0C540  ,                                                \ displ(ay)
    $59148000  ,                                                \ actc
    $F7478100  ,                                                \ +list
    $72797400  ,                                                \ itick
    $A3C00000  ,                                                \ lis
    $F6800000  ,                                                \ +e
    $820E1400  ,                                                \ serve
    $4C0E4A00  ,                                                \ edita     editAddress
    $963A3B60  ,                                                \ cblind
\    dd $B18C5480   \  format
\    dd $C5270000   \  pci
\    dd $68248000   \  nsec      was devic(e)
    $85DCA590  ,                                                \ switch
    $B0A27640  ,                                                \ freeze
    $23C40000  ,                                                \ top
\    dd $B1896480   \  forths
\    dd $8AC84E00   \  macros
    0          ,   

LABEL ForthJumpTableROM                                         \ jumptable:
    boot       ,                                                \
    warm       ,                                                \
    pause_     ,                                                \ pause
    macro      ,                                                \
    forth      ,                                                \
    c_         ,                                                \ c
    bios_rusb  - $$ + BOOTOFFSET ,                              \ jmp bios_rusb   'stop_'
    th_        - $$ + BOOTOFFSET ,                              \ jmp th_ ( thunk to BIOS Int $13 )
    bios_read  - $$ + BOOTOFFSET ,                              \ jmp bios_read   'read'
    bios_write - $$ + BOOTOFFSET ,                              \ jmp bios_write  'write'
    rsect_     - $$ + BOOTOFFSET ,                              \ jmp rsect_  'rsect'
    fetchDX_   ,                                                \ @dx
    storeDAP_  ,                                                \ !dap
    nc         ,                                                \
    0          ,                                                \ comman(d) \  cmdf
    seek       ,                                                \ seek      \  seekf
    0          ,                                                \ ready     \  readyf
    activate   ,                                                \ act
    show       ,                                                \
    _load_     ,                                                \
    nload      ,                                                \ nload
    plusLoad   ,                                                \ +load
    thru_      ,                                                \ thru
    plusThru_  ,                                                \ +thru
    cblk       ,                                                \           return the block number currently being compiled, calculated from  edi
    here       ,                                                \
    qlit       ,                                                \
    comma3     ,                                                \
    comma2     ,                                                \
    comma1     ,                                                \
    comma      ,                                                \
    less       ,                                                \
    jump       ,                                                \
    accept     ,                                                \
    keypd      ,                                                \
    keypd      ,                                                \
    wipe       ,                                                \
    wipes      ,                                                \
    copy_      ,                                                \ copy
    mark       ,                                                \
    empty      ,                                                \
    emit_      ,                                                \ emit
    digit      ,                                                \
    two_emit   ,                                                \ 2emit
    dotDecimal ,                                                \ .
    dotHex8    ,                                                \ h.
    h_dot_n    ,                                                \ h.n
    cr_        ,                                                \ cr
    space_     ,                                                \ space
    down       ,                                                \
    edit       ,                                                \
    e_         ,                                                \
    lm         ,                                                \
    rm         ,                                                \
    graphAction ,                                               \ graph
    setupText_ ,                                                \ text
    displayTheKeyboard ,                                        \
    debug      ,                                                \
    _at        ,                                                \ at
    plus_at    ,                                                \ +at
    xy         ,                                                \
    page_      ,                                                \ page
    screen_    ,                                                \ screen
    fov_       ,                                                \
\    dd fifo         \ 
    box_       ,                                                \ box
    line_      ,                                                \ line
    color      ,                                                \
    octant     ,                                                \
    tokenActions_ ,                                             \ tokenActions table
    last       ,                                                \
    unpack     ,                                                \
    pack       ,                                                \
    blk        ,                                                \
    scrnw_     ,                                                \ scrnw  screen width in pixels
    scrnh_     ,                                                \ scrnh  screen height in pixels
    bpp_       ,                                                \ bpp    bits per pixel
    font_      ,                                                \ font   address of font16x24
    iconw_     ,                                                \ iconw  icon width in pixels
    iconh_     ,                                                \ iconh  icon height in pixels
    version_   ,                                                \ ver
    curs       ,                                                \ curs
    block_     ,                                                \ block
    vframe     ,                                                \ vframe
    vars_      ,                                                \ vars
\ new words
    seeb       ,                                                \ seeb
\    dd stacks_      \ 
    dotsf      ,                                                \ dotsf
    rgb        ,                                                \ rgb
    bye_       ,                                                \ bye
    _word      ,                                                \
    ekt        ,                                                \
    abort      ,                                                \
    tickh      ,                                                \
    buffer_    ,                                                \ buffe(r)
    offset_    ,                                                \
    tic_       ,                                                \ tic
    aper       ,                                                \
    vesa       ,                                                \
    trash_     ,                                                \ trash
    hsvv       ,                                                \
    tsim       ,                                                \ nc
    vword_     ,                                                \ ('%s')", DB_NAME,
    vregs_     ,                                                \ vregs
    dap_       ,                                                \ dap
    sect_      ,                                                \ sect
    words_     ,                                                \ words
    get_key    ,                                                \ key
    get_qwerty_key ,                                            \ qkey
    digin      ,                                                \
    qwert      ,                                                \
    rquery     ,                                                \ r?
    nul        ,                                                \
    cad        ,                                                \
    pcad       ,                                                \
    displ      ,                                                \
    actc       ,                                                \
    plusList   ,                                                \ +list
    itick      ,                                                \
    refresh    ,                                                \ lis
    plus_e     ,                                                \ +e
    serve      ,                                                \
    editAddress ,                                               \ edita
    cBlindAddr_ ,                                               \ cblind
\    dd format       \ 
\    dd pci          \ 
\    dd device       \ 
    switch     ,                                                \
    freeze     ,                                                \
    top_       ,                                                \
\    dd forths_      \ 
\    dd macros_      \ 

LABEL ForthJumpTableROM_end                                     

\ times ( 2048 - 32 ) db 0

    ( ( START_BLOCK_NUMBER - SIZE_OF_FONT_IN_BLOCKS ) * $400 ) - $40 - ($ - $$) db 0 times \ fill with no-ops to 55AA at end of boot sector

LABEL version  
    'cf2018 1v0 2018Jun10 Chuck Moore'  c,  $00 c,              
    ' Howerd Oakford inventio.co.uk'  c,   $00 c,               

\ the above produces a 12K boot image, we then add :
LABEL font16x24  
\ incbin "cf2018icons.fnt"            \  append the font : ASCII order, 128 x 2 byte wide * 24 high  = 6144 bytes
\ incbin "smallIcons.fnt"           \  append the font : ASCII order, 128 x 2 byte wide * 24 high  = 6144 bytes small characters
\ colorForth:                         \  the colorForth source blocks
    OFFSET_OF_FONT, ( 512 * 1024 ) "cf2018Ref.img" incbin       \ 18432        \  append the font and colorForth source blocks from the reference image, skip the 12K of kernel code
\ incbin "cf2018Ref.img",18432,(147456-18432) \  append the colorForth source blocks from the reference image, skip the 18K of kernel code
\ incbin "Oldcf2018Ref.img",147456 \  append the colorForth source blocks from the reference image, skip the 18K of kernel code

\ end of file

PREVIOUS FORTH DEFINITIONS
